@@ -1,6 +1,7 @@
 /****************************************************************
  *								*
- *	Copyright 2003, 2014 Fidelity Information Services, Inc	*
+ * Copyright (c) 2003-2016 Fidelity National Information	*
+ * Services, Inc. and/or its subsidiaries. All rights reserved.	*
  *								*
  *	This source code contains the intellectual property	*
  *	of its copyright holder(s), and is made available	*
@@ -13,7 +14,6 @@
 
 #include "gtm_string.h"
 #include "gtm_stat.h"
-#if defined(UNIX)
 #include "gtm_fcntl.h"
 #include "gtm_unistd.h"
 #include "eintr_wrappers.h"
@@ -21,12 +21,6 @@
 #if defined(__MVS__)
 #include "gtm_zos_io.h"
 #endif
-#elif defined(VMS)
-#include <rms.h>
-#include <iodef.h>
-#include <efndef.h>
-#include "iosb_disk.h"
-#endif
 
 #include "gtm_file_stat.h"
 #include "gtm_rename.h"
@@ -90,11 +84,6 @@ if (SYSCALL_ERROR(info->status) || SYSCALL_ERROR(info->status2))	\
 	return EXIT_ERR;						\
 }
 
-#if defined(VMS)
-#define ZERO_SIZE_IN_BLOCKS	127 /* 127 is RMS maximum blocks / write */
-#define ZERO_SIZE		(ZERO_SIZE_IN_BLOCKS * DISK_BLOCK_SIZE)
-#endif
-
 GBLREF 	jnl_gbls_t		jgbl;
 GBLREF	boolean_t		mupip_jnl_recover;
 GBLREF	jnl_process_vector	*prc_vec;
@@ -120,7 +109,7 @@ uint4	cre_jnl_file(jnl_create_info *info)
 	boolean_t	no_rename;
 
 	assert(0 != jgbl.gbl_jrec_time);
-	if (!info->no_rename)	/* ***MAY*** be rename is required */
+	if (!info->no_rename)	/* ***MAYBE*** rename is required */
 	{
 		no_rename = FALSE;
 		if (SS_NORMAL != (info->status = prepare_unique_name((char *)info->jnl, info->jnl_len, "", "",
@@ -177,13 +166,6 @@ uint4 cre_jnl_file_common(jnl_create_info *info, char *rename_fn, int rename_fn_
 	fd_type			channel;
 	char            	*jrecbuf, *jrecbuf_base;
 	gd_id			jnlfile_id;
-#	ifdef VMS
-	struct FAB      	fab;
-	struct NAM      	nam;
-	char            	es_buffer[MAX_FN_LEN], name_buffer[MAX_FN_LEN];
-	uint4			blk, block, zero_size;
-	io_status_block_disk	iosb;
-#	else
 	struct stat		stat_buf;
 	int			fstat_res;
 	ZOS_ONLY(int		realfiletag;)
@@ -192,14 +174,14 @@ uint4 cre_jnl_file_common(jnl_create_info *info, char *rename_fn, int rename_fn_
 	int			group_id;
 	struct stat		sb;
 	int			perm;
-	struct perm_diag_data	pdd;
-#	endif
 	int			idx;
 	trans_num		db_tn;
 	uint4			temp_offset, temp_checksum, pfin_offset, eof_offset;
 	uint4			jnl_fs_block_size;
+	sgmnt_addrs		*csa;
 
 	jrecbuf = NULL;
+	csa = info->csa;
 	if (info->no_rename)
 	{	/* The only cases where no-renaming is possible are as follows
     		 * (i) MUPIP SET JOURNAL where the new journal file name is different from the current journal file name
@@ -214,8 +196,8 @@ uint4 cre_jnl_file_common(jnl_create_info *info, char *rename_fn, int rename_fn_
 		assert(0 == create_fn[create_fn_len]);
 	} else
 	{
-		if (NULL != info->csa)
-			cre_jnl_file_intrpt_rename(info->csa);	/* deal with *_new.mjl files */
+		if (NULL != csa)
+			cre_jnl_file_intrpt_rename(csa);	/* deal with *_new.mjl files */
 		create_fn = &fn_buff[0];
 		if (SS_NORMAL != (info->status = prepare_unique_name((char *)info->jnl, (int)info->jnl_len, "", EXT_NEW,
 								     (char *)create_fn, &create_fn_len, 0, &info->status2)))
@@ -224,8 +206,7 @@ uint4 cre_jnl_file_common(jnl_create_info *info, char *rename_fn, int rename_fn_
 			return EXIT_ERR;
 		}
 	}
-#	ifdef UNIX
-	OPENFILE3((char *)create_fn, O_CREAT | O_EXCL | O_RDWR, 0600, channel);
+	OPENFILE3_CLOEXEC((char *)create_fn, O_CREAT | O_EXCL | O_RDWR, 0600, channel);
 	if (-1 == channel)
 	{
 		info->status = errno;
@@ -253,26 +234,11 @@ uint4 cre_jnl_file_common(jnl_create_info *info, char *rename_fn, int rename_fn_
 		F_CLOSE(channel, status);
 		return EXIT_ERR;
 	}
-	/* setup new group and permissions if indicated by the security rules.
-	 */
-	if (gtm_permissions(&sb, &user_id, &group_id, &perm, PERM_FILE, &pdd) < 0)
-	{
-		send_msg_csa(CSA_ARG(info->csa) VARLSTCNT(6+PERMGENDIAG_ARG_COUNT)
-			ERR_PERMGENFAIL, 4, RTS_ERROR_STRING("journal file"), RTS_ERROR_STRING(info->fn),
-			PERMGENDIAG_ARGS(pdd));
-		if (IS_GTM_IMAGE)
-			rts_error_csa(CSA_ARG(info->csa) VARLSTCNT(6+PERMGENDIAG_ARG_COUNT)
-				ERR_PERMGENFAIL, 4, RTS_ERROR_STRING("journal file"), RTS_ERROR_STRING(info->fn),
-				PERMGENDIAG_ARGS(pdd));
-		else
-			gtm_putmsg_csa(CSA_ARG(info->csa) VARLSTCNT(6+PERMGENDIAG_ARG_COUNT)
-				ERR_PERMGENFAIL, 4, RTS_ERROR_STRING("journal file"), RTS_ERROR_STRING(info->fn),
-				PERMGENDIAG_ARGS(pdd));
-		F_CLOSE(channel, status);
-		return EXIT_ERR;
-	}
+	/* Setup new group and permissions if indicated by the security rules */
+	gtm_permissions(&sb, &user_id, &group_id, &perm, PERM_FILE);
 	/* if group not the same then change group of temporary file */
-	if ((((-1 != user_id) && (user_id != stat_buf.st_uid)) || ((-1 != group_id) && (group_id != stat_buf.st_gid)))
+	if ((((INVALID_UID != user_id) && (user_id != stat_buf.st_uid))
+			|| ((INVALID_GID != group_id) && (group_id != stat_buf.st_gid)))
 		&& (-1 == fchown(channel, user_id, group_id)))
 	{
 		info->status = errno;
@@ -300,54 +266,7 @@ uint4 cre_jnl_file_common(jnl_create_info *info, char *rename_fn, int rename_fn_
 	jrecbuf = (char *)ROUND_UP2((uintszofptr_t)jrecbuf_base, jnl_fs_block_size);
 	memset(jrecbuf, 0, jnl_fs_block_size);
 	set_gdid_from_stat(&jnlfile_id, &stat_buf);
-#	else
-	nam = cc$rms_nam;
-	nam.nam$l_rsa = name_buffer;
-	nam.nam$b_rss = SIZEOF(name_buffer);
-	nam.nam$l_esa = es_buffer;
-	nam.nam$b_ess = SIZEOF(es_buffer);
-	nam.nam$b_nop = NAM$M_NOCONCEAL;
-	fab = cc$rms_fab;
-	fab.fab$l_nam = &nam;
-	fab.fab$b_org = FAB$C_SEQ;
-	fab.fab$b_rfm = FAB$C_FIX;
-	fab.fab$l_fop = FAB$M_UFO | FAB$M_MXV | FAB$M_CBT;
-	fab.fab$b_fac = FAB$M_GET | FAB$M_PUT | FAB$M_BIO;
-	fab.fab$b_shr = FAB$M_SHRGET | FAB$M_SHRPUT | FAB$M_UPI | FAB$M_NIL;
-	fab.fab$w_mrs = DISK_BLOCK_SIZE;
-	fab.fab$l_alq = info->alloc;
-	fab.fab$w_deq = info->extend;
-	fab.fab$l_fna = create_fn;
-	fab.fab$b_fns = create_fn_len;
-	info->status = sys$create(&fab);
-	if (0 == (info->status & 1))
-	{
-		info->status2 = fab.fab$l_stv;	/* store secondary status information */
-		STATUS_MSG(info);
-		return EXIT_ERR;
-	}
-	channel = fab.fab$l_stv;
-	jrecbufbase_size = ZERO_SIZE;
-	jrecbuf_base = jrecbuf = malloc(ZERO_SIZE);
-	memset(jrecbuf, 0, ZERO_SIZE);
-	block = (JNL_HDR_LEN >> LOG2_DISK_BLOCK_SIZE);
-	assert(block * DISK_BLOCK_SIZE == JNL_HDR_LEN);
-	for (blk = block;  blk < info->alloc;  blk += ZERO_SIZE_IN_BLOCKS)
-	{
-		zero_size = (blk + ZERO_SIZE_IN_BLOCKS <= info->alloc) ?
-			ZERO_SIZE : (info->alloc - blk) * DISK_BLOCK_SIZE;
-		JNL_DO_FILE_WRITE(NULL, NULL, channel, blk * DISK_BLOCK_SIZE, jrecbuf, zero_size, info->status, info->status2);
-		STATUS_MSG(info);
-		RETURN_ON_ERROR(info);
-	}
-	memcpy(jnlfile_id.dvi, &nam.nam$t_dvi, SIZEOF(jnlfile_id.dvi));
-	memcpy(jnlfile_id.did, &nam.nam$w_did, SIZEOF(jnlfile_id.did));
-	memcpy(jnlfile_id.fid, &nam.nam$w_fid, SIZEOF(jnlfile_id.fid));
-	jnl_fs_block_size = get_fs_block_size(channel);
-#	endif
-	info->checksum = compute_checksum(INIT_CHECKSUM_SEED, (uint4 *)&jnlfile_id, SIZEOF(gd_id));
-	/* Journal file header size relies on this assert */
-	assert(256 == GTMCRYPT_RESERVED_HASH_LEN);
+	info->checksum = compute_checksum(INIT_CHECKSUM_SEED, (unsigned char *)&jnlfile_id, SIZEOF(gd_id));
 	header = (jnl_file_header *)(ROUND_UP2((uintszofptr_t)hdr_base, jnl_fs_block_size));
 	/* We have already saved previous journal file name in info */
 	jfh_from_jnl_info(info, header);
@@ -362,7 +281,7 @@ uint4 cre_jnl_file_common(jnl_create_info *info, char *rename_fn, int rename_fn_
 	 * write is not so much since it is only once per journal file at creation time. All future writes of the
 	 * file header write only the real file header and not the 0-padding.
 	 */
-	JNL_DO_FILE_WRITE(info->csa, create_fn, channel, 0, header, JNL_HDR_LEN, info->status, info->status2);
+	JNL_DO_FILE_WRITE(csa, create_fn, channel, 0, header, JNL_HDR_LEN, info->status, info->status2);
 	STATUS_MSG(info);
 	RETURN_ON_ERROR(info);
 	assert(DISK_BLOCK_SIZE >= EPOCH_RECLEN + EOF_RECLEN + PFIN_RECLEN + PINI_RECLEN);
@@ -379,11 +298,11 @@ uint4 cre_jnl_file_common(jnl_create_info *info, char *rename_fn, int rename_fn_
 	/* Already process_vector[ORIG_JPV] is memset 0 */
 	pini_record->filler = 0;
 	pini_record->prefix.checksum = INIT_CHECKSUM_SEED;
-	temp_checksum = compute_checksum(INIT_CHECKSUM_SEED, (uint4 *)pini_record, SIZEOF(struct_jrec_pini));
+	temp_checksum = compute_checksum(INIT_CHECKSUM_SEED, (unsigned char *)pini_record, SIZEOF(struct_jrec_pini));
 	temp_offset = JNL_HDR_LEN;
 	ADJUST_CHECKSUM(temp_checksum, temp_offset, temp_checksum);
 	ADJUST_CHECKSUM(temp_checksum, info->checksum, pini_record->prefix.checksum);
-	/* EPOCHs are written unconditionally in Unix while they are written only for BEFORE_IMAGE in VMS */
+	/* EPOCHs are written unconditionally in Unix */
 	if (JNL_HAS_EPOCH(info))
 	{
 		epoch_record = (struct_jrec_epoch *)&jrecbuf[PINI_RECLEN];
@@ -398,22 +317,15 @@ uint4 cre_jnl_file_common(jnl_create_info *info, char *rename_fn, int rename_fn_
 		epoch_record->fully_upgraded = info->csd->fully_upgraded;
 		epoch_record->suffix.suffix_code = JNL_REC_SUFFIX_CODE;
 		epoch_record->jnl_seqno = info->reg_seqno;
-		UNIX_ONLY(
-			for (idx = 0; idx < MAX_SUPPL_STRMS; idx++)
-				epoch_record->strm_seqno[idx] = info->csd->strm_reg_seqno[idx];
-			if (jgbl.forw_phase_recovery)
-			{	/* If MUPIP JOURNAL -ROLLBACK, might need some adjustment. See macro definition for comments */
-				MUR_ADJUST_STRM_REG_SEQNO_IF_NEEDED(info->csd, epoch_record->strm_seqno);
-			}
-		)
-		VMS_ONLY(
-			for (idx = 0; idx < MAX_SUPPL_STRMS; idx++)
-				assert(0 == epoch_record->strm_seqno[idx]); /* should have been zeroed already by above memset */
-		)
+		for (idx = 0; idx < MAX_SUPPL_STRMS; idx++)
+			epoch_record->strm_seqno[idx] = info->csd->strm_reg_seqno[idx];
+		if (jgbl.forw_phase_recovery)
+		{	/* If MUPIP JOURNAL -ROLLBACK, might need some adjustment. See macro definition for comments */
+			MUR_ADJUST_STRM_REG_SEQNO_IF_NEEDED(info->csd, epoch_record->strm_seqno);
+		}
 		epoch_record->filler = 0;
 		epoch_record->prefix.checksum = INIT_CHECKSUM_SEED;
-		temp_checksum = compute_checksum(INIT_CHECKSUM_SEED,
-							(uint4 *)epoch_record, SIZEOF(struct_jrec_epoch));
+		temp_checksum = compute_checksum(INIT_CHECKSUM_SEED, (unsigned char *)epoch_record, SIZEOF(struct_jrec_epoch));
 		temp_offset = JNL_HDR_LEN + PINI_RECLEN;
 		ADJUST_CHECKSUM(temp_checksum, temp_offset, temp_checksum);
 		ADJUST_CHECKSUM(temp_checksum, info->checksum, epoch_record->prefix.checksum);
@@ -438,7 +350,7 @@ uint4 cre_jnl_file_common(jnl_create_info *info, char *rename_fn, int rename_fn_
 	pfin_record->suffix.suffix_code = JNL_REC_SUFFIX_CODE;
 	pfin_record->filler = 0;
 	pfin_record->prefix.checksum = INIT_CHECKSUM_SEED;
-	temp_checksum = compute_checksum(INIT_CHECKSUM_SEED, (uint4 *)pfin_record, SIZEOF(struct_jrec_pfin));
+	temp_checksum = compute_checksum(INIT_CHECKSUM_SEED, (unsigned char *)pfin_record, SIZEOF(struct_jrec_pfin));
 	ADJUST_CHECKSUM(temp_checksum, pfin_offset, temp_checksum);
 	ADJUST_CHECKSUM(temp_checksum, info->checksum, pfin_record->prefix.checksum);
 	eof_record->prefix.jrec_type = JRT_EOF;
@@ -450,7 +362,7 @@ uint4 cre_jnl_file_common(jnl_create_info *info, char *rename_fn, int rename_fn_
 	eof_record->suffix.suffix_code = JNL_REC_SUFFIX_CODE;
 	eof_record->filler = 0;
 	eof_record->prefix.checksum = INIT_CHECKSUM_SEED;
-	temp_checksum = compute_checksum(INIT_CHECKSUM_SEED, (uint4 *)eof_record, SIZEOF(struct_jrec_eof));
+	temp_checksum = compute_checksum(INIT_CHECKSUM_SEED, (unsigned char *)eof_record, SIZEOF(struct_jrec_eof));
 	ADJUST_CHECKSUM(temp_checksum, eof_offset, temp_checksum);
 	ADJUST_CHECKSUM(temp_checksum, info->checksum, eof_record->prefix.checksum);
 	/* Assert that the journal file header and journal records are all in sync with respect to the db tn. */
@@ -469,11 +381,20 @@ uint4 cre_jnl_file_common(jnl_create_info *info, char *rename_fn, int rename_fn_
 	 */
 	assert(ROUND_UP2(header->virtual_size, jnl_fs_block_size/DISK_BLOCK_SIZE)
 			> DIVIDE_ROUND_UP(JNL_HDR_LEN + write_size, DISK_BLOCK_SIZE));
-	JNL_DO_FILE_WRITE(info->csa, create_fn, channel, JNL_HDR_LEN, jrecbuf, write_size, info->status, info->status2);
+	JNL_DO_FILE_WRITE(csa, create_fn, channel, JNL_HDR_LEN, jrecbuf, write_size, info->status, info->status2);
 	STATUS_MSG(info);
 	RETURN_ON_ERROR(info);
-	UNIX_ONLY(GTM_JNL_FSYNC(info->csa, channel, status);)
+	UNIX_ONLY(GTM_JNL_FSYNC(csa, channel, status);)
 	F_CLOSE(channel, status);	/* resets "channel" to FD_INVALID */
+	/* Now that EOF record has been written, keep csa->jnl->jnl_buff->prev_jrec_time up to date.
+	 * One exception is if journaling is not yet turned on but is being turned on by the current caller.
+	 * In that case, dont worry about maintaining jb->prev_jrec_time. It will be maintained when this newly
+	 * created journal file is first opened.
+	 */
+	if ((NULL != csa) && (NULL != csa->nl))	/* this means database shared memory is accessible (i.e. region is open) */
+	{	/* Keep jb->prev_jrec_time up to date */
+		SET_JNLBUFF_PREV_JREC_TIME(csa->jnl->jnl_buff, eof_record->prefix.time, DO_GBL_JREC_TIME_CHECK_TRUE);
+	}
 	free(jrecbuf_base);
 	jrecbuf_base = NULL;
 #	ifdef DEBUG
@@ -487,15 +408,15 @@ uint4 cre_jnl_file_common(jnl_create_info *info, char *rename_fn, int rename_fn_
 	 * Following does rename of a.mjl to a.mjl_timestamp.
 	 * So system will have a.mjl_timestamp and a.mjl_new for a crash after this call
 	 */
-	WAIT_FOR_REPL_INST_UNFREEZE_SAFE(info->csa);	/* wait for instance freeze before journal file renames */
+	WAIT_FOR_REPL_INST_UNFREEZE_SAFE(csa);	/* wait for instance freeze before journal file renames */
 	if (SS_NORMAL != (info->status = gtm_rename((char *)info->jnl, (int)info->jnl_len,
 						    (char *)rename_fn, rename_fn_len, &info->status2)))
 	{
 		if (IS_GTM_IMAGE)
-			send_msg_csa(CSA_ARG(info->csa) VARLSTCNT(6) ERR_RENAMEFAIL, 4, info->jnl_len, info->jnl, rename_fn_len,
+			send_msg_csa(CSA_ARG(csa) VARLSTCNT(6) ERR_RENAMEFAIL, 4, info->jnl_len, info->jnl, rename_fn_len,
 					rename_fn);
 		else
-			gtm_putmsg_csa(CSA_ARG(info->csa) VARLSTCNT(6) ERR_RENAMEFAIL, 4, info->jnl_len, info->jnl, rename_fn_len,
+			gtm_putmsg_csa(CSA_ARG(csa) VARLSTCNT(6) ERR_RENAMEFAIL, 4, info->jnl_len, info->jnl, rename_fn_len,
 					rename_fn);
 		STATUS_MSG(info);
 		return EXIT_ERR;
@@ -503,24 +424,24 @@ uint4 cre_jnl_file_common(jnl_create_info *info, char *rename_fn, int rename_fn_
 	/* Following does rename of a.mjl_new to a.mjl.
 	 * So system will have a.mjl_timestamp as previous generation and a.mjl as new/current journal file
 	 */
-	WAIT_FOR_REPL_INST_UNFREEZE_SAFE(info->csa);	/* wait for instance freeze before journal file renames */
+	WAIT_FOR_REPL_INST_UNFREEZE_SAFE(csa);	/* wait for instance freeze before journal file renames */
 	if (SS_NORMAL !=  (info->status = gtm_rename((char *)create_fn, create_fn_len,
 						     (char *)info->jnl, (int)info->jnl_len, &info->status2)))
 	{
 		if (IS_GTM_IMAGE)
-			send_msg_csa(CSA_ARG(info->csa) VARLSTCNT(6) ERR_RENAMEFAIL, 4, info->jnl_len, info->jnl, rename_fn_len,
+			send_msg_csa(CSA_ARG(csa) VARLSTCNT(6) ERR_RENAMEFAIL, 4, info->jnl_len, info->jnl, rename_fn_len,
 					rename_fn);
 		else
-			gtm_putmsg_csa(CSA_ARG(info->csa) VARLSTCNT(6) ERR_RENAMEFAIL, 4, info->jnl_len, info->jnl, rename_fn_len,
+			gtm_putmsg_csa(CSA_ARG(csa) VARLSTCNT(6) ERR_RENAMEFAIL, 4, info->jnl_len, info->jnl, rename_fn_len,
 					rename_fn);
 		STATUS_MSG(info);
 		return EXIT_ERR;
 	}
 	if (IS_GTM_IMAGE)
-		send_msg_csa(CSA_ARG(info->csa) VARLSTCNT (6) ERR_FILERENAME, 4, info->jnl_len, info->jnl, rename_fn_len,
+		send_msg_csa(CSA_ARG(csa) VARLSTCNT (6) ERR_FILERENAME, 4, info->jnl_len, info->jnl, rename_fn_len,
 				rename_fn);
 	else
-		gtm_putmsg_csa(CSA_ARG(info->csa) VARLSTCNT (6) ERR_FILERENAME, 4, info->jnl_len, info->jnl, rename_fn_len,
+		gtm_putmsg_csa(CSA_ARG(csa) VARLSTCNT (6) ERR_FILERENAME, 4, info->jnl_len, info->jnl, rename_fn_len,
 				rename_fn);
 #		ifdef DEBUG
 		if (gtm_white_box_test_case_enabled && (WBTEST_JNL_CREATE_INTERRUPT == gtm_white_box_test_case_number))
