@@ -1,6 +1,7 @@
 /****************************************************************
  *								*
- *	Copyright 2001, 2013 Fidelity Information Services, Inc	*
+ * Copyright (c) 2001-2016 Fidelity National Information	*
+ * Services, Inc. and/or its subsidiaries. All rights reserved.	*
  *								*
  *	This source code contains the intellectual property	*
  *	of its copyright holder(s), and is made available	*
@@ -48,9 +49,9 @@ void db_auto_upgrade(gd_region *reg)
 	if (NULL == csd)
 		return;
 
-	if (0 == csd->mutex_spin_parms.mutex_hard_spin_count)
+	if (0 > csd->mutex_spin_parms.mutex_hard_spin_count)
 		csd->mutex_spin_parms.mutex_hard_spin_count = MUTEX_HARD_SPIN_COUNT;
-	if (0 == csd->mutex_spin_parms.mutex_sleep_spin_count)
+	if (0 > csd->mutex_spin_parms.mutex_sleep_spin_count)
 		csd->mutex_spin_parms.mutex_sleep_spin_count = MUTEX_SLEEP_SPIN_COUNT;
 	/* zero is a legitimate value for csd->mutex_spin_parms.mutex_spin_sleep_mask; so can't detect if need re-initialization */
 	INIT_NUM_CRIT_ENTRY_IF_NEEDED(csd);
@@ -103,14 +104,22 @@ void db_auto_upgrade(gd_region *reg)
 		 * d) Add a new case with the new minor version
 		 * e) Add assert(FALSE) and break (like it was before)
 		 */
-		switch(csd->minor_dbver)
+		switch (csd->minor_dbver)
 		{	/* Note that handling for any fields introduced in a version will not go in the "switch-case" block
 			 * of code introduced for the new version but will go in the PREVIOUS "switch-case" block.
 			 */
+			case GDSMV50000:
+			case GDSMV51000ALT:
 			case GDSMV51000:		/* Multi-site replication available */
 			case GDSMV52000:		/* Unicode */
 			case GDSMV53000:		/* M-Itanium release */
 				gvstats_rec_upgrade(csa); /* Move GVSTATS information to new place in file header */
+			case GDSMV53003:		/* ZSHOW "G" release */
+				 csd->is_encrypted = FALSE;
+				 memset(csd->encryption_hash, 0, GTMCRYPT_RESERVED_HASH_LEN);
+			case GDSMV53004:		/* New encryption fields */
+				csd->db_trigger_cycle = 0;
+			case GDSMV54000:		/* First trigger version */
 			case GDSMV54002:
 				/* GT.M V54002B introduced jnl_eov_tn for backward recovery */
 				csd->jnl_eovtn = csd->trans_hist.curr_tn;
@@ -132,11 +141,33 @@ void db_auto_upgrade(gd_region *reg)
 			case GDSMV60001:
 				/* GT.M V60002 introduced mutex_spin_parms.mutex_que_entry_space_size */
 				NUM_CRIT_ENTRY(csd) = DEFAULT_NUM_CRIT_ENTRY;
-				break;
 			case GDSMV60002:
+				/* GT.M V62001 introduced ^#t upgrade. Record this pending event in filehdr. */
+				UNIX_ONLY(csd->hasht_upgrade_needed = TRUE;)
+			case GDSMV62001:
+				/* GT.M V62002 introduced database file preallocation. */
+				csd->defer_allocate = TRUE;
+				/* GT.M V62002 incremented ^#t label. Record this pending event in filehdr. */
+				UNIX_ONLY(csd->hasht_upgrade_needed = TRUE;)
+				/* GT.M V62002 introduced epoch taper */
+				csd->epoch_taper = TRUE;
+		        	csd->epoch_taper_time_pct = EPOCH_TAPER_TIME_PCT_DEFAULT;
+		        	csd->epoch_taper_jnl_pct = EPOCH_TAPER_JNL_PCT_DEFAULT;
+			case GDSMV62002:
+				/* GT.M V63000 introduced non-null IV encryption and encryption on-the-fly. */
+				 csd->non_null_iv = FALSE;
+				 csd->encryption_hash_cutoff = UNSTARTED;
+				 csd->encryption_hash2_start_tn = 0;
+				 memset(csd->encryption_hash2, 0, GTMCRYPT_RESERVED_HASH_LEN);
+				 SPIN_SLEEP_MASK(csd) = 0;	/* previously unused, but was 7FF and it should now default to 0 */
+				break;
+			case GDSMV63000:
 				/* Nothing to do for this version since it is GDSMVCURR for now. */
 				assert(FALSE);		/* When this assert fails, it means a new GDSMV* was created, */
 				break;			/* 	so a new "case" needs to be added BEFORE the assert. */
+			default:
+				/* Unrecognized version in the header */
+				assertpro(FALSE && csd->minor_dbver);
 		}
 		csd->minor_dbver = GDSMVCURR;
 		if (0 == csd->wcs_phase2_commit_wait_spincnt)
@@ -149,7 +180,7 @@ void db_auto_upgrade(gd_region *reg)
                  */
 		assert(!dse_running);
 		csd->db_got_to_v5_once = TRUE; /* fix it in PRO */
-		send_msg(VARLSTCNT(6) ERR_DBBADUPGRDSTATE, 4, REG_LEN_STR(reg), DB_LEN_STR(reg));
+		send_msg_csa(CSA_ARG(csa) VARLSTCNT(6) ERR_DBBADUPGRDSTATE, 4, REG_LEN_STR(reg), DB_LEN_STR(reg));
 	}
 	return;
 }
