@@ -1,6 +1,7 @@
 /****************************************************************
  *								*
- *	Copyright 2010, 2014 Fidelity Information Services, Inc	*
+ * Copyright (c) 2010-2016 Fidelity National Information	*
+ * Services, Inc. and/or its subsidiaries. All rights reserved.	*
  *								*
  *	This source code contains the intellectual property	*
  *	of its copyright holder(s), and is made available	*
@@ -29,19 +30,21 @@
 #include "gtm_socket.h"
 #include "gtm_unistd.h"
 #include "gtm_limits.h"
+#include "gtm_un.h"
+#include "gtm_signal.h"
 
-#include <signal.h>
 #include <sys/time.h>
-#ifdef UNIX
-# include <sys/un.h>
-#endif
 #ifdef VMS
 # include <descrip.h>		/* Required for gtmsource.h */
 # include <ssdef.h>
 # include <fab.h>
 # include "desblk.h"
 #endif
+#ifdef UNIX
+# include <glob.h>
+#endif
 #include "cache.h"
+#include "gtm_multi_thread.h"
 #include "hashtab_addr.h"
 #include "hashtab_int4.h"
 #include "hashtab_int8.h"
@@ -139,22 +142,20 @@
 #ifdef UNICODE_SUPPORTED
 # include "gtm_icu_api.h"
 # include "gtm_utf8.h"
+# include "utfcgr.h"
 #endif
 
-#ifdef GTM_CRYPT
-# include "gtmcrypt.h"
-# include "gdsblk.h"
-# include "muextr.h"
-#endif
+#include "gtmcrypt.h"
+#include "gdsblk.h"
+#include "muextr.h"
 
 #ifdef GTM_TRIGGER
 # include "gv_trigger.h"
 # include "gtm_trigger.h"
 #endif
 
-#ifdef USHBIN_SUPPORTED
+#ifdef AUTORELINK_SUPPORTED
 # include "relinkctl.h"
-# include "zhist.h"
 #endif
 
 #include "gtm_threadgbl_init.h"
@@ -200,18 +201,18 @@ void gtm_threadgbl_init(void)
 		 * the best we can.
 		 */
 		FPRINTF(stderr, "GTM-F-GTMASSERT gtm_threadgbl_true_t and gtm_threadgbl_t are different sizes\n");
-		exit(ERR_GTMASSERT);
+		EXIT(ERR_GTMASSERT);
 	}
 	if (NULL != gtm_threadgbl)
 	{	/* has already been initialized - don't re-init */
 		FPRINTF(stderr, "GTM-F-GTMASSERT gtm_threadgbl is already initialized\n");
-		exit(ERR_GTMASSERT);
+		EXIT(ERR_GTMASSERT);
 	}
 	gtm_threadgbl = lcl_gtm_threadgbl = malloc(size_gtm_threadgbl_struct);
 	if (NULL == gtm_threadgbl)
 	{	/* Storage was not allocated for some reason - no error handling yet still */
 		perror("GTM-F-MEMORY Unable to allocate startup thread structure");
-		exit(UNIX_ONLY(ERR_MEMORY) VMS_ONLY(ERR_VMSMEMORY));
+		EXIT(UNIX_ONLY(ERR_MEMORY) VMS_ONLY(ERR_VMSMEMORY));
 	}
 	memset(gtm_threadgbl, 0, size_gtm_threadgbl_struct);
 	gtm_threadgbl_true = (gtm_threadgbl_true_t *)gtm_threadgbl;
@@ -227,5 +228,8 @@ void gtm_threadgbl_init(void)
 	MEMCPY_LIT(TADR(prombuf), DEFAULT_PROMPT);
 	(TREF(replgbl)).jnl_release_timeout = DEFAULT_JNL_RELEASE_TIMEOUT;
 	(TREF(window_ident)).addr = TADR(window_string);
+	ASSERT_SAFE_TO_UPDATE_THREAD_GBLS;
 	TREF(util_outbuff_ptr) = TADR(util_outbuff);	/* Point util_outbuff_ptr to the beginning of util_outbuff at first. */
+	TREF(util_outptr) = TREF(util_outbuff_ptr);
+	TREF(max_advancewindow_line) = MAX_SRCLINE;
 }
