@@ -1,6 +1,7 @@
 /****************************************************************
  *								*
- *	Copyright 2001, 2014 Fidelity Information Services, Inc	*
+ * Copyright (c) 2001-2015 Fidelity National Information 	*
+ * Services, Inc. and/or its subsidiaries. All rights reserved.	*
  *								*
  *	This source code contains the intellectual property	*
  *	of its copyright holder(s), and is made available	*
@@ -216,9 +217,6 @@ void io_init(boolean_t term_ctrl);
 bool io_is_rm(mstr *name);
 bool io_is_sn(mstr *tn);
 struct mv_stent_struct *io_find_mvstent(io_desc *io_ptr, boolean_t clear_mvstent);
-#ifdef UNIX
-bool io_is_tt(char *name);
-#endif
 bool io_open_try(io_log_name *naml, io_log_name *tl, mval *pp, int4 timeout, mval *mspace);
 enum io_dev_type io_type(mstr *tn);
 void io_init_name(void);
@@ -345,6 +343,7 @@ void get_dlr_zkey(mval *v);
 
 
 void flush_pio(void);
+void write_text_newline_and_flush_pio(mstr *text);
 
 void remove_rms(io_desc *ciod);
 void iosocket_destroy(io_desc *ciod);
@@ -422,4 +421,63 @@ LITREF unsigned char ebcdic_spaces_block[];
 }
 #define SET_ENCODING(CHSET, CHSET_MSTR)	SET_ENCODING_VALIDATE(CHSET, CHSET_MSTR,)
 
+/* Establish a GT.M I/O condition handler if one is not already active and the principal device is the current one. */
+#define ESTABLISH_GTMIO_CH(IOD, SET_CH)												\
+{																\
+	GBLREF io_pair		io_std_device;											\
+	GBLREF boolean_t	in_prin_gtmio;											\
+																\
+	intrpt_state_t		prev_intrpt_state;										\
+																\
+	if ((&gtmio_ch != active_ch->ch) && (NULL != (IOD)->out)								\
+			&& (NULL != io_std_device.out) && ((IOD)->out == io_std_device.out))					\
+	{															\
+		DEFER_INTERRUPTS(INTRPT_IN_GTMIO_CH_SET, prev_intrpt_state);							\
+		ESTABLISH(gtmio_ch);												\
+		SET_CH = TRUE;													\
+		in_prin_gtmio = TRUE;												\
+		ENABLE_INTERRUPTS(INTRPT_IN_GTMIO_CH_SET, prev_intrpt_state);							\
+	} else															\
+		SET_CH = FALSE;													\
+}
+
+/* Establish a GT.M I/O condition handler with return if one is not already active and the principal device is the current one. */
+#define ESTABLISH_RET_GTMIO_CH(IOD, VALUE, SET_CH)										\
+{																\
+	GBLREF io_pair		io_std_device;											\
+	GBLREF boolean_t	in_prin_gtmio;											\
+																\
+	intrpt_state_t		prev_intrpt_state;										\
+																\
+	if ((&gtmio_ch != active_ch->ch) && (NULL != (IOD)->out)								\
+			&& (NULL != io_std_device.out) && ((IOD)->out == io_std_device.out))					\
+	{															\
+		DEFER_INTERRUPTS(INTRPT_IN_GTMIO_CH_SET, prev_intrpt_state);							\
+		ESTABLISH_RET(gtmio_ch, VALUE);											\
+		SET_CH = TRUE;													\
+		in_prin_gtmio = TRUE;												\
+		ENABLE_INTERRUPTS(INTRPT_IN_GTMIO_CH_SET, prev_intrpt_state);							\
+	} else															\
+		SET_CH = FALSE;													\
+}
+
+/* Revert a GT.M I/O condition handler if one was set (based on the passed argument). */
+#define REVERT_GTMIO_CH(IOD, SET_CH)												\
+{																\
+	GBLREF boolean_t		in_prin_gtmio;										\
+	DEBUG_ONLY(GBLREF io_pair	io_std_device;)										\
+																\
+	intrpt_state_t		prev_intrpt_state;										\
+																\
+	if (SET_CH)														\
+	{															\
+		assert((&gtmio_ch == active_ch->ch) && (NULL != (IOD)->out)							\
+			&& (NULL != io_std_device.out) && ((IOD)->out == io_std_device.out));					\
+		DEFER_INTERRUPTS(INTRPT_IN_GTMIO_CH_SET, prev_intrpt_state);							\
+		in_prin_gtmio = FALSE;												\
+		REVERT;														\
+		ENABLE_INTERRUPTS(INTRPT_IN_GTMIO_CH_SET, prev_intrpt_state);							\
+	}															\
+}
+
 #endif /* IO_H */
