@@ -1,6 +1,7 @@
 /****************************************************************
  *								*
- *	Copyright 2001, 2014 Fidelity Information Services, Inc	*
+ * Copyright (c) 2001-2016 Fidelity National Information	*
+ * Services, Inc. and/or its subsidiaries. All rights reserved.	*
  *								*
  *	This source code contains the intellectual property	*
  *	of its copyright holder(s), and is made available	*
@@ -83,6 +84,10 @@ typedef unsigned int 	uint4;		/* 4-byte unsigned integer */
  */
 #define SHMDT(X)	shmdt((void *)(X))
 
+/* Because shmget and semget returns -1 to indicate error, -1 can never be a valid shmid or semid */
+#define INVALID_SEMID			-1
+#define INVALID_SHMID 			-1L
+
 /* constant needed for FIFO - OS390 redefines in mdefsp.h */
 #define FIFO_PERMISSION		010666 /* fifo with RW permissions for owner, group, other */
 
@@ -112,20 +117,33 @@ error_def(ERR_ASSERT);
 # define lvaddr "%08lx"
 #endif
 
+/* Need to set a flag to 0 or 1 depending on whether AUTORELINK_SUPPORTED is set or not. This is needed by gtmpcat so it can
+ * tell if autorelink is supported on the platform or not. While this could be done by always setting AUTORELINK_SUPPORTED
+ * which a value of 0 or 1, that's not how the code that uses it is structured so create this new variable with a value
+ * depending on whether AUTORELINK_SUPPORTED is defined or not. Specifically, this variable is used with a $DEFINE stmt in
+ * gtmpcat_field_defs.txt which gtmpcatfldbld.m uses to create the gtmpcat initialization file. This allows us to capture the
+ * values of #define constants and put them into gtmpcat initialization files for use by gtmpcat.
+ */
+#ifdef AUTORELINK_SUPPORTED
+# define ARLINK_ENABLED 1
+#else
+# define ARLINK_ENABLED 0
+#endif
+
 /* Define GT.M interlude functions for open, close, pipe, creat and dup system calls. This lets GT.M trace through all file
- * descriptor activity (needed for D9I11-002714). Do this on all Unix platforms. Note that only the macro GTM_FD_TRACE is
+ * descriptor activity (needed for D9I11-002714). Do this on all UNIX platforms. Note that only the macro GTM_FD_TRACE is
  * defined here. gtm_unistd.h and gtm_fcntl.h define the actual GT.M interlude functions based on this macro.
  */
-#if defined(UNIX)
+#if defined(UNIX) && ! defined(STATIC_ANALYSIS)
 #	define	GTM_FD_TRACE
 #	define	GTM_FD_TRACE_ONLY(X)	X
 #else
 #	define	GTM_FD_TRACE_ONLY(X)
 #endif
 
-/* Define what is an invalid file descriptor in Unix and VMS. */
+/* Define what is an invalid file descriptor in UNIX and VMS. */
 #if defined(UNIX)
-#	define	FD_INVALID		-1	/* fd of -1 is invalid in Unix posix calls */
+#	define	FD_INVALID		-1	/* fd of -1 is invalid in UNIX posix calls */
 #	define	FD_INVALID_NONPOSIX	FD_INVALID
 #else
 #	define	FD_INVALID		-1	/* fd of -1 is invalid in VMS if using POSIX interface (open/close etc.) */
@@ -143,18 +161,12 @@ error_def(ERR_ASSERT);
 #endif
 
 /* INTPTR_T is an integer that has the same length as a pointer on each platform.  Its basic use is for arithmetic
- * or generic parameters.  For all platforms except Tru64/VMS (alpha platforms), the [U]INTPTR_T types will be
- * equivalenced to [u]intptr_t.  But since this type is used for alignment and other checking, and since Tru64/VMS
- * (implemented as a 32 bit platform) unconditionally sets this type to its 8 char variant, on Tru64/VMS we will
- * explicitly make [U]INTPTR_T a 4 byte creature.
+ * or generic parameters.
  */
-#if !defined(__alpha)
+
 typedef intptr_t INTPTR_T;
 typedef uintptr_t UINTPTR_T;
-#else
-typedef int INTPTR_T;
-typedef unsigned int UINTPTR_T;
-#endif
+
 /* The intszofptr_t type is defined to be basically the same size as an address on the platforms it runs on. So it
    is the same size as INTPTR_T without the connotation of being a pointer. This is used in places where size_t
    or ssize_t would normally be used except they can't be used because they are the wrong size on Alpha systems.
@@ -213,8 +225,10 @@ typedef UINTPTR_T uintszofptr_t;
 #if defined(__i386) || defined(__x86_64__) || defined(_AIX) || defined (__sun)
 #	define GTM_PTHREAD
 #	define GTM_PTHREAD_ONLY(X) X
+#	define NON_GTM_PTHREAD_ONLY(X)
 #else
 #	define GTM_PTHREAD_ONLY(X)
+#	define NON_GTM_PTHREAD_ONLY(X)	X
 #endif
 
 #if defined(__ia64)
@@ -333,7 +347,7 @@ typedef struct
 #	define HAS_LITERAL_SECT
 #endif
 
-typedef long		ulimit_t;	/* NOT int4; the Unix ulimit function returns a value of type long */
+typedef long		ulimit_t;	/* NOT int4; the UNIX ulimit function returns a value of type long */
 
 /* Bit definitions for mval type (mvtype) */
 #define MV_NM		 1	/* 0x0001 */
@@ -412,7 +426,7 @@ typedef long		ulimit_t;	/* NOT int4; the Unix ulimit function returns a value of
  * 		Byte pattern			max. expanded 	   input byte	ratio
  * 						output length	 length
  * 	------------------------------------------------------------------------------
- * 		$C(129)_$ZCH(128)_		18			2	 9
+ * 		$C(127)_$ZCH(128)_		18			2	 9
  * 		$C(1536)_$ZCH(128)_		19			3	 7
  * 		$C(65279)_$ZCH(128)_		20			4	 5
  * 		$C(917585)_$ZCH(128)_		21			5	 6
@@ -455,6 +469,8 @@ mval *underr_strict(mval *start, ...);
 #define MV_FORCE_BOOL(X)	(MV_FORCE_NUM(X), (X)->m[1] ? TRUE : FALSE)
 #define MV_FORCE_INT(M)		(MV_FORCE_DEFINED(M), MV_FORCE_INTD(M))
 #define MV_FORCE_INTD(M)	(DBG_ASSERT(MV_DEFINED(M)) (M)->mvtype & MV_INT ? (M)->m[1]/MV_BIAS : mval2i(M))
+#define MV_FORCE_UINT(M)	(MV_FORCE_DEFINED(M), MV_FORCE_UINTD(M))
+#define MV_FORCE_UINTD(M)	(DBG_ASSERT(MV_DEFINED(M)) (M)->mvtype & MV_INT ? (M)->m[1]/MV_BIAS : mval2ui(M))
 #define MV_FORCE_UMVAL(M,I)	(((I) >= 1000000) ? i2usmval((M),(int)(I)) : \
 				(void)( (M)->mvtype = MV_NM | MV_INT , (M)->m[1] = (int)(I)*MV_BIAS ))
 #define MV_FORCE_MVAL(M,I)	(((I) >= 1000000 || (I) <= -1000000) ? i2mval((M),(int)(I)) : \
@@ -480,8 +496,8 @@ mval *underr_strict(mval *start, ...);
 #define MV_IS_STRING(X)		(((X)->mvtype & MV_STR) != 0)
 #define MV_DEFINED(X)		(((X)->mvtype & (MV_STR | MV_NM)) != 0)
 #define MV_IS_CANONICAL(X)	(((X)->mvtype & MV_NM) ? (((X)->mvtype & MV_NUM_APPROX) == 0) : (boolean_t)val_iscan(X))
-#define MV_INIT(X)		((X)->mvtype = 0, (X)->fnpc_indx = 0xff)
-#define MV_INIT_STRING(X, LEN, ADDR) ((X)->mvtype = MV_STR, (X)->fnpc_indx = 0xff,		\
+#define MV_INIT(X)		((X)->mvtype = 0, (X)->fnpc_indx = UNICODE_ONLY((X)->utfcgr_indx =) 0xff)
+#define MV_INIT_STRING(X, LEN, ADDR) ((X)->mvtype = MV_STR, (X)->fnpc_indx = UNICODE_ONLY((X)->utfcgr_indx =) 0xff,	\
 				      (X)->str.len = INTCAST(LEN), (X)->str.addr = (char *)ADDR)
 
 /* The MVTYPE_IS_* macros are similar to the MV_IS_* macros except that the input is an mvtype instead of an "mval *".
@@ -503,28 +519,33 @@ mval *underr_strict(mval *start, ...);
 #define DEFINE_MVAL_STRING(TYPE, EXPONENT, SIGN, LENGTH, ADDRESS, MANT_LOW, MANT_HIGH)		\
 	DEFINE_MVAL_COMMON(TYPE, EXPONENT, SIGN, 0, LENGTH, ADDRESS, MANT_LOW, MANT_HIGH)
 
-#ifdef VMS
-#define DEFINE_MVAL_COMMON(TYPE, EXPONENT, SIGN, UTF_LEN, LENGTH, ADDRESS, MANT_LOW, MANT_HIGH) \
-	{TYPE, EXPONENT, SIGN, 0xff, LENGTH, ADDRESS, MANT_LOW, MANT_HIGH}
-#else
 #ifdef BIGENDIAN
-#ifdef UNICODE_SUPPORTED
-#define DEFINE_MVAL_COMMON(TYPE, EXPONENT, SIGN, UTF_LEN, LENGTH, ADDRESS, MANT_LOW, MANT_HIGH) \
-	{TYPE, SIGN, EXPONENT, 0xff, MANT_LOW, MANT_HIGH, UTF_LEN, LENGTH, ADDRESS}
-#else
-#define DEFINE_MVAL_COMMON(TYPE, EXPONENT, SIGN, UTF_LEN, LENGTH, ADDRESS, MANT_LOW, MANT_HIGH) \
+# ifdef UNICODE_SUPPORTED
+#  ifdef GTM64
+#   define DEFINE_MVAL_COMMON(TYPE, EXPONENT, SIGN, UTF_LEN, LENGTH, ADDRESS, MANT_LOW, MANT_HIGH) \
+	{TYPE, SIGN, EXPONENT, 0xff, 0xff, MANT_LOW, MANT_HIGH, UTF_LEN, LENGTH, ADDRESS}
+#  else
+#   define DEFINE_MVAL_COMMON(TYPE, EXPONENT, SIGN, UTF_LEN, LENGTH, ADDRESS, MANT_LOW, MANT_HIGH) \
+	{TYPE, SIGN, EXPONENT, 0xff, 0xff, 0, MANT_LOW, MANT_HIGH, UTF_LEN, LENGTH, ADDRESS}
+#  endif /* GTM64 */
+# else
+#  define DEFINE_MVAL_COMMON(TYPE, EXPONENT, SIGN, UTF_LEN, LENGTH, ADDRESS, MANT_LOW, MANT_HIGH) \
 	{TYPE, SIGN, EXPONENT, 0xff, MANT_LOW, MANT_HIGH, LENGTH, ADDRESS}
-#endif
-#else	/* BIGENDIAN */
-#ifdef UNICODE_SUPPORTED
-#define DEFINE_MVAL_COMMON(TYPE, EXPONENT, SIGN, UTF_LEN, LENGTH, ADDRESS, MANT_LOW, MANT_HIGH) \
-	{TYPE, EXPONENT, SIGN, 0xff, MANT_LOW, MANT_HIGH, UTF_LEN, LENGTH, ADDRESS}
-#else
-#define DEFINE_MVAL_COMMON(TYPE, EXPONENT, SIGN, UTF_LEN, LENGTH, ADDRESS, MANT_LOW, MANT_HIGH) \
+# endif	/* UNICODE */
+#else	/* end BIGENDIAN -- start LITTLEENDIAN */
+# ifdef UNICODE_SUPPORTED
+#  ifdef GTM64
+#    define DEFINE_MVAL_COMMON(TYPE, EXPONENT, SIGN, UTF_LEN, LENGTH, ADDRESS, MANT_LOW, MANT_HIGH) \
+	{TYPE, EXPONENT, SIGN, 0xff, 0xff, MANT_LOW, MANT_HIGH, UTF_LEN, LENGTH, ADDRESS}
+#  else
+#    define DEFINE_MVAL_COMMON(TYPE, EXPONENT, SIGN, UTF_LEN, LENGTH, ADDRESS, MANT_LOW, MANT_HIGH) \
+	{TYPE, EXPONENT, SIGN, 0xff, 0xff, 0, MANT_LOW, MANT_HIGH, UTF_LEN, LENGTH, ADDRESS}
+#  endif /* GTM64 */
+# else
+#  define DEFINE_MVAL_COMMON(TYPE, EXPONENT, SIGN, UTF_LEN, LENGTH, ADDRESS, MANT_LOW, MANT_HIGH) \
 	{TYPE, EXPONENT, SIGN, 0xff, MANT_LOW, MANT_HIGH, LENGTH, ADDRESS}
-#endif	/* UNICODE */
-#endif	/* BIGENDIAN */
-#endif	/* VMS */
+# endif	/* UNICODE */
+#endif	/* BIGENDIAN/LITTLEENDIAN */
 
 #define	ASCII_MAX		(unsigned char)0x7F
 #define	IS_ASCII(X)		((uint4)(X) <= ASCII_MAX)	/* X can be greater than 255 hence the typecast to uint4 */
@@ -583,14 +604,6 @@ mval *underr_strict(mval *start, ...);
 /* Length needed to pad out to a given power of 2 boundary */
 #define PADLEN(value, bndry) (int)(ROUND_UP2((sm_long_t)(value), bndry) - (sm_long_t)(value))
 
-/* LOG2_OF_INTEGER returns the ceiling of log (base 2) of number */
-#define LOG2_OF_INTEGER(number, log2_of_number)			\
-{								\
-	int    temp = (number) - 1;				\
-	for (log2_of_number = 0; 0 < temp; log2_of_number++)	\
-	   temp = (temp) >> 1; 					\
-}
-
 #define CALLFROM	LEN_AND_LIT(__FILE__), __LINE__
 void gtm_assert(int file_name_len, char file_name[], int line_no);
 int gtm_assert2(int condlen, char *condtext, int file_name_len, char file_name[], int line_no);
@@ -672,7 +685,7 @@ int4 timeout2msec(int4 timeout);
 /* If this is unix, we have a faster sleep for short sleeps ( < 1 second) than doing a hiber start.
  * Take this chance to define UNIX_ONLY and VMS_ONLY macros.
  */
-int m_usleep(int useconds);
+void m_usleep(int useconds);
 #ifdef UNIX
 #	define SHORT_SLEEP(x) {assert(1000 > (x)); m_usleep((x) * 1000);}
 #else
@@ -723,10 +736,11 @@ int m_usleep(int useconds);
 #endif
 
 /* Note the macros below refer to the UNIX Shared Binary Support. Because the
-   support is *specifically* for the Unix platform, "NON_USHBIN_ONLY()" will
-   also be true for VMS even though that platform does have shared binary support
-   (but it does not have Unix Shared Binary support). Use "NON_USHBIN_UNIX_ONLY()"
-   for UNIX platforms that do not support Shared Binaries. */
+ * support is *specifically* for the UNIX platform, "NON_USHBIN_ONLY()" will
+ * also be true for VMS even though that platform does have shared binary support
+ * (but it does not have UNIX Shared Binary support). Use "NON_USHBIN_UNIX_ONLY()"
+ * for UNIX platforms that do not support Shared Binaries.
+ */
 #ifdef USHBIN_SUPPORTED
 #	define USHBIN_ONLY(X)		X
 #	define NON_USHBIN_ONLY(X)
@@ -740,10 +754,17 @@ int m_usleep(int useconds);
 #		define NON_USHBIN_UNIX_ONLY(X)
 #	endif
 #endif
-
-/* Unicode. Although most (all?) Unix platforms currently support Unicode, that may
-   not always be the case so a separate contingent is defined.
-*/
+/* Autorelink related macros */
+#ifdef AUTORELINK_SUPPORTED
+#	define ARLINK_ONLY(X)		X
+#	define NON_ARLINK_ONLY(X)
+#else
+#	define ARLINK_ONLY(X)
+#	define NON_ARLINK_ONLY(X)	X
+#endif
+/* Unicode. Although most (all?) UNIX platforms currently support Unicode, that may
+ * not always be the case so a separate contingent is defined.
+ */
 #ifdef UNICODE_SUPPORTED
 #	define UNICODE_ONLY(X) X
 #	define NON_UNICODE_ONLY(X)
@@ -774,6 +795,9 @@ int m_usleep(int useconds);
    typedef	int64_t			gtm_int64_t;
 #endif
 
+int ceil_log2_32bit(uint4 num);
+int ceil_log2_64bit(gtm_uint64_t num);
+
 typedef INTPTR_T	sm_off_t;
 
 /* HPPA latches (used by load_and_clear) must be 16 byte aligned.
@@ -802,8 +826,7 @@ typedef struct
 } global_latch_t;
 #define latch_image_count latch_word
 
-#define GLOBAL_LATCH_HELD_BY_US(latch) (process_id == (latch)->u.parts.latch_pid \
-					VMS_ONLY(&& image_count == (latch)->u.parts.latch_image_count))
+#define GLOBAL_LATCH_HELD_BY_US(latch) (process_id == (latch)->u.parts.latch_pid)
 
 typedef struct compswap_time_field_struct
 {	/* This structure is used where we want to do a compare-n-swap (CAS) on a time value. The CAS interfaces
@@ -1321,8 +1344,8 @@ char is_ident(mstr *v);
 int val_iscan(mval *v);
 void mcfree(void);
 int4 getprime(int4 n);
-void push_parm(UNIX_ONLY_COMMA(unsigned int totalcnt) int truth_value, ...);
-UNIX_ONLY(void suspend(int sig);)
+void push_parm(unsigned int totalcnt, int truth_value, ...);
+void suspend(int sig);
 mval *push_mval(mval *arg1);
 void mval_lex(mval *v, mstr *output);
 
@@ -1376,6 +1399,10 @@ qw_num	gtm_byteswap_64(qw_num num64);
 #define ONEDAY			86400	/* seconds in a day */
 #define MILLISECS_IN_SEC	1000	/* millseconds in a second */
 #define MICROSEC_IN_SEC		1000000 /* microseconds in a second */
+#define MICROSECS_IN_MSEC	1000	/* microseconds in a millisecond */
+#define E_6			1000000
+#define E_9			1000000000
+#define E_18			1000000000000000000LL
 
 #define ASSERT_IN_RANGE(low, x, high)	assert((low <= x) && (x <= high))
 
@@ -1384,7 +1411,7 @@ qw_num	gtm_byteswap_64(qw_num num64);
 #define VARLSTCNT1(CNT)			VARLSTCNT(CNT)
 #define PUT_SYS_ERRNO(SYS_ERRNO) 	SYS_ERRNO
 #elif defined(UNIX)
-#define DAYS		47117 /* adjust Unix returned days (seconds converted to days); Unix zero time 1970 */
+#define DAYS		47117 /* adjust UNIX returned days (seconds converted to days); UNIX zero time 1970 */
 #define VARLSTCNT1(CNT)			VARLSTCNT(CNT + 1)
 #define PUT_SYS_ERRNO(SYS_ERRNO) 	0, SYS_ERRNO
 #else
@@ -1688,24 +1715,22 @@ typedef enum
 
 #define CHK_BOUNDARY_ALIGNMENT(pointer) (((UINTPTR_T)pointer) & (SIZEOF(UINTPTR_T) - 1))
 
-/* Encryption and TLS related macros */
+/* Encryption- and TLS-related macros */
 #if defined(__ia64) || defined(__i386) || defined(__x86_64__) || defined(__sparc) || defined(_AIX) || defined(__s390__)
-# define GTM_CRYPT
-# define GTMCRYPT_ONLY(X)		X
 # define GTM_TLS
 # define GTMTLS_ONLY(X)			X
 # define GTMTLS_ONLY_COMMA(X)		, X
 # define NON_GTMTLS_ONLY(X)
 #else
-# define GTMCRYPT_ONLY(X)
 # define GTMTLS_ONLY(X)
 # define GTMTLS_ONLY_COMMA(X)
 # define NON_GTMTLS_ONLY(X)		X
 #endif
 
 #define GTMCRYPT_HASH_LEN		64
+#define GTMCRYPT_RESERVED_HASH_LEN	80
 #define GTMCRYPT_HASH_HEX_LEN		GTMCRYPT_HASH_LEN * 2
-#define GTMCRYPT_RESERVED_HASH_LEN	256
+#define GTMCRYPT_RESERVED_ENCR_SPACE	256
 
 #define GTMCRYPT_PLUGIN_DIR_NAME	"plugin"
 #define GTMCRYPT_UTIL_LIBNAME		"libgtmcryptutil.so"
@@ -1728,17 +1753,11 @@ typedef enum
 #	define	GTM_SNAPSHOT_ONLY(X)
 #endif
 
-/* Currently MUPIP REORG -TRUNCATE is only supported on Unix */
-#ifdef UNIX
-#	define	GTM_TRUNCATE
-#	define	NON_GTM_TRUNCATE_ONLY(X)
-#	define	GTM_TRUNCATE_ONLY(X)		X
-#else
-#	define	NON_GTM_TRUNCATE_ONLY(X)	X
-#	define	GTM_TRUNCATE_ONLY(X)
-#endif
+#define	GTM_TRUNCATE
+#define	NON_GTM_TRUNCATE_ONLY(X)
+#define	GTM_TRUNCATE_ONLY(X)		X
 
-/* Currently triggers are supported only on Unix */
+/* Currently triggers are supported only on UNIX */
 #if defined(UNIX) && !defined(__hppa)	/* triggers not supported on HPUX-HPPA */
 #	define	GTM_TRIGGER
 #	define	GTMTRIG_ONLY(X)			X
@@ -1777,9 +1796,14 @@ enum
 #define MAX_INDSUBSCRIPTS 32
 #define MAX_FOR_STACK 32
 
-#define	MAX_ACTUALS	32	/* Maximum number of arguments allowed in an actuallist. This value also determines
-				 * how many parameters are allowed to be passed between M and C.
-				 */
+#define PUSH_PARM_OVERHEAD	4	/* This extra space in the array is needed because push_parm() is capable of handling 32
+					 * arguments, but callg needs to accomidate space for 4 items, namely argument count
+					 * (repeated twice), return value, and the truth value. As of this writing ojchildparms.c is
+					 * the only module that relies on the extra space.
+					 */
+#define	MAX_ACTUALS		32	/* Maximum number of arguments allowed in an actuallist. This value also determines
+					 * how many parameters are allowed to be passed between M and C.
+					 */
 #if defined(DEBUG) && defined(UNIX)
 #define OPERATOR_LOG_MSG												\
 {															\
@@ -1796,7 +1820,7 @@ enum
 #ifdef GTM_PTHREAD
 /* If we detect a case when the signal came to a thread other than the main GT.M thread, this macro will redirect the signal to the
  * main thread if such is defined. Such scenarios is possible, for instance, if we are running along a JVM, which, upon receiving a
- * signal, dispatches a new thread to invoke signal handlers other than its own. The ptrhead_kill() enables us to target the signal
+ * signal, dispatches a new thread to invoke signal handlers other than its own. The pthread_kill() enables us to target the signal
  * to a specific thread rather than rethrow it to the whole process.
  */
 #define FORWARD_SIG_TO_MAIN_THREAD_IF_NEEDED(SIG)								\
@@ -1829,9 +1853,9 @@ enum
 #define MV_FORCE_DEFINED_UNLESS_SKIPARG(V)	((!M_ARG_SKIPPED(V)) ? (MV_FORCE_DEFINED(V)) : (V))
 
 #ifdef _AIX
-#define LIBPATH_ENV	"LIBPATH"
+#define LIBPATH_ENV		"LIBPATH"
 #else
-#define LIBPATH_ENV	"LD_LIBRARY_PATH"
+#define LIBPATH_ENV		"LD_LIBRARY_PATH"
 #endif
 
 #endif /* MDEF_included */
