@@ -1,6 +1,7 @@
 /****************************************************************
  *								*
- *	Copyright 2001, 2013 Fidelity Information Services, Inc	*
+ * Copyright (c) 2001-2015 Fidelity National Information	*
+ * Services, Inc. and/or its subsidiaries. All rights reserved.	*
  *								*
  *	This source code contains the intellectual property	*
  *	of its copyright holder(s), and is made available	*
@@ -15,7 +16,7 @@
 #include "gtm_stdio.h"
 #include <errno.h>
 #include "gtm_fcntl.h"
-#include <signal.h>
+#include "gtm_signal.h"
 #include "gtm_stat.h"
 #include "gtm_unistd.h"
 #include "min_max.h"
@@ -86,7 +87,7 @@ LITDEF  MIDENT_CONST(above_routine, "*above*");
 	else				\
 	{				\
 		gtm_putmsg_csa x;	\
-		exit(EXIT_FAILURE);	\
+		EXIT(EXIT_FAILURE);	\
 	}				\
 }
 #endif
