@@ -1,6 +1,7 @@
 /***************************************************************
  *								*
- *	Copyright 2001, 2013 Fidelity Information Services, Inc	*
+ * Copyright (c) 2001-2016 Fidelity National Information	*
+ * Services, Inc. and/or its subsidiaries. All rights reserved.	*
  *								*
  *	This source code contains the intellectual property	*
  *	of its copyright holder(s), and is made available	*
@@ -67,6 +68,7 @@
 #include "gtmmsg.h"
 #include "mu_getlst.h"
 #include "gvcst_protos.h"	/* for gvcst_root_search prototype */
+#include "mupip_reorg_encrypt.h"
 
 error_def(ERR_CONCURTRUNCINPROG);
 error_def(ERR_DBRDONLY);
@@ -82,8 +84,6 @@ error_def(ERR_NOEXCLUDE);
 error_def(ERR_REORGCTRLY);
 error_def(ERR_REORGINC);
 
-GTMTRIG_ONLY(LITREF mval		literal_hasht;)
-
 GBLREF bool		mu_ctrlc_occurred;
 GBLREF bool		mu_ctrly_occurred;
 GBLREF boolean_t	mu_reorg_process;
@@ -100,6 +100,9 @@ GBLREF inctn_opcode_t	inctn_opcode;
 #ifdef UNIX
 GBLREF	boolean_t	jnlpool_init_needed;
 #endif
+
+static readonly mval literal_poollimit =
+	DEFINE_MVAL_LITERAL(MV_STR | MV_NUM_APPROX, 0, 0, (SIZEOF("POOLLIMIT") - 1), "POOLLIMIT", 0, 0);
 void mupip_reorg(void)
 {
 	boolean_t		resume, reorg_success = TRUE;
@@ -108,6 +111,7 @@ void mupip_reorg(void)
 	char			cli_buff[MAX_LINE], *ptr;
 	glist			gl_head, exclude_gl_head, *gl_ptr, hasht_gl;
 	uint4			cli_status;
+	mval			reg_nam_mval = DEFINE_MVAL_STRING(MV_STR, 0 , 0 , SIZEOF(MAX_RN_LEN), 0, 0, 0);
 	unsigned short		n_len;
 	boolean_t		truncate, cur_success, restrict_reg, arg_present;
 	int			root_swap_statistic;
@@ -134,12 +138,17 @@ void mupip_reorg(void)
 	)
 
 	if ((CLI_PRESENT == cli_present("UPGRADE")) || (CLI_PRESENT == cli_present("DOWNGRADE")))
-	{
-		/* note that "mu_reorg_process" is not set to TRUE in case of MUPIP REORG UPGRADE/DOWNGRADE.
-		 * this is intentional because we are not doing any REORG kind of processing.
+	{	/* Note that "mu_reorg_process" is not set to TRUE in case of MUPIP REORG -UPGRADE/DOWNGRADE.
+		 * This is intentional because we are not doing any REORG kind of processing.
 		 */
 		mu_reorg_upgrd_dwngrd();
 		mupip_exit(SS_NORMAL);	/* does not return */
+	} else if (CLI_PRESENT == cli_present("ENCRYPT"))
+	{	/* Note that "mu_reorg_process" is not set to TRUE in case of MUPIP REORG -ENCRYPT.
+		 * This is intentional because we are not doing any REORG kind of processing.
+		 */
+		mupip_reorg_encrypt();
+		mupip_exit(SS_NORMAL);	/* does not return */
 	}
 	grlist = NULL;
 	restrict_reg = FALSE;
@@ -246,7 +255,6 @@ void mupip_reorg(void)
 	}
 	TREF(want_empty_gvts) = FALSE;
 
-	root_swap_statistic = 0;
 	mu_reorg_process = TRUE;
 	assert(NULL == gv_currkey_next_reorg);
 	GVKEYSIZE_INIT_IF_NEEDED;	/* sets "gv_keysize", "gv_currkey" and "gv_altkey" (if not already done) */
@@ -256,6 +264,21 @@ void mupip_reorg(void)
 	reorg_gv_target->alt_hist->depth = 0;
 	for (gl_ptr = gl_head.next; gl_ptr; gl_ptr = gl_ptr->next)
 	{
+		/* mu_reorg_process can't be set before gv_select above as the assert(!mu_reorg_process) fails in gvcst_search.
+		 * The option to set mu_reorg_process before calling gv_select and clearing it after being used in gvcst_init,
+		 *  can't be taken up as gvcst_init is called multiple times, once for each region.
+		 * This means gvcst_init() doesn't set csa->gbuff and we need to set csa->gbuff_limit explicitly now. */
+		if (0 == (TREF(gbuff_limit)).str.len)
+		{
+			(TREF(gbuff_limit)).str.len = SIZEOF(REORG_GBUFF_LIMIT);
+			(TREF(gbuff_limit)).str.addr = malloc(SIZEOF(REORG_GBUFF_LIMIT));
+			memcpy((TREF(gbuff_limit)).str.addr, REORG_GBUFF_LIMIT, SIZEOF(REORG_GBUFF_LIMIT));
+		}
+		reg_nam_mval.str.len = gv_cur_region->rname_len;
+		reg_nam_mval.str.addr = (char *)&gv_cur_region->rname;
+		op_view(VARLSTCNT(3) &literal_poollimit, &reg_nam_mval, &(TREF(gbuff_limit)));
+		if (WBTEST_ENABLED(WBTEST_REORG_DEBUG))
+			util_out_print("GTMPOOLLIMIT used for mupip reorg : !UL", TRUE, cs_addrs->gbuff_limit);
 		util_out_print("   ", FLUSH);
 		util_out_print("Global: !AD (region !AD)", FLUSH, GNAME(gl_ptr).len, GNAME(gl_ptr).addr, REG_LEN_STR(gl_ptr->reg));
 		if (in_exclude_list((unsigned char *)GNAME(gl_ptr).addr, GNAME(gl_ptr).len, &exclude_gl_head))
@@ -275,7 +298,6 @@ void mupip_reorg(void)
 #		ifdef GTM_TRUNCATE
 		if (truncate)
 		{	/* No need to move root blocks unless truncating */
-			cur_success &= mu_swap_root(gl_ptr, &root_swap_statistic);
 			assert(gv_cur_region == gl_ptr->reg);	/* should have been set inside "mu_reorg" call done above */
 			if (cur_success)
 			{	/* add region corresponding to this global to the set (list) of regions to truncate */
@@ -297,29 +319,22 @@ void mupip_reorg(void)
 					if (truncate)
 					{	/* Reorg ^#t in this region to move it out of the way. */
 						SET_GVTARGET_TO_HASHT_GBL(cs_addrs);	/* sets gv_target */
+						inctn_opcode = inctn_invalid_op;	/* needed for INITIAL_HASHT_ROOT_SEARCH */
+						INITIAL_HASHT_ROOT_SEARCH_IF_NEEDED;	/* sets gv_target->root */
+						DBG_CHECK_GVTARGET_GVCURRKEY_IN_SYNC(CHECK_CSA_TRUE);
 						hasht_gl.next = NULL;
 						hasht_gl.reg = gv_cur_region;
 						hasht_gl.gvt = gv_target;
-						inctn_opcode = inctn_invalid_op;	/* needed for *ROOT_SEARCH */
-						INITIAL_HASHT_ROOT_SEARCH_IF_NEEDED;	/* sets gv_target->root */
-						DBG_CHECK_GVTARGET_GVCURRKEY_IN_SYNC(CHECK_CSA_TRUE);
 						if (0 != gv_target->root)
 						{
 							util_out_print("   ", FLUSH);
 							util_out_print("Global: !AD (region !AD)", FLUSH,
 								GNAME(&hasht_gl).len, GNAME(&hasht_gl).addr,
 								REG_LEN_STR(gv_cur_region));
-							reorg_gv_target->gvname.var_name = literal_hasht.str;
+							reorg_gv_target->gvname.var_name = gv_target->gvname.var_name;
 							cur_success = mu_reorg(&hasht_gl, &exclude_gl_head, &resume,
 										index_fill_factor, data_fill_factor, reorg_op);
 							reorg_success &= cur_success;
-							SET_GV_CURRKEY_FROM_GVT(reorg_gv_target);
-							/* Recompute gv_target->root in case mu_reorg changed things around */
-							gv_target->root = 0;
-							inctn_opcode = inctn_invalid_op;	/* needed for GVCST_ROOT_SEARCH */
-							GVCST_ROOT_SEARCH;	/* set gv_target->root */
-							if (gv_target->root)
-								reorg_success &= mu_swap_root(&hasht_gl, &root_swap_statistic);
 						}
 					}
 #					endif
@@ -335,12 +350,35 @@ void mupip_reorg(void)
 	}
 	if (!reorg_success)
 	{
+		inctn_opcode = inctn_invalid_op;	/* needed by assert inside "preemptive_db_clnup" called by rts_error */
 		rts_error_csa(CSA_ARG(NULL) VARLSTCNT(1) ERR_REORGINC);
 		mupip_exit(ERR_REORGINC);
 	}
 	else if (truncate)
 	{
 #		ifdef GTM_TRUNCATE
+		/* Move GVT ROOT blocks of all global names AFTER doing regular reorg on ALL global names.
+		 * This way we ensure one pass of reorg -truncate is enough to produce an optimally truncated file.
+		 */
+		root_swap_statistic = 0;
+		for (gl_ptr = gl_head.next; gl_ptr; gl_ptr = gl_ptr->next)
+			mu_swap_root(gl_ptr, &root_swap_statistic);
+		hasht_gl.next = NULL;
+		for (reg_iter = reg_list; reg_iter; reg_iter = reg_iter->next)
+		{
+			gv_cur_region = reg_iter->reg;
+			tp_change_reg();
+			SET_GVTARGET_TO_HASHT_GBL(cs_addrs);	/* sets gv_target */
+			SET_GV_CURRKEY_FROM_GVT(gv_target);
+			gv_target->root = 0; /* Recompute gv_target->root in case mu_reorg changed things around */
+			inctn_opcode = inctn_invalid_op;	/* needed for GVCST_ROOT_SEARCH */
+			GVCST_ROOT_SEARCH;			/* set gv_target->root */
+			if (0 == gv_target->root)
+				continue;
+			hasht_gl.reg = gv_cur_region;
+			hasht_gl.gvt = gv_target;
+			mu_swap_root(&hasht_gl, &root_swap_statistic);
+		}
 		util_out_print("Total root blocks moved: !UL", FLUSH, root_swap_statistic);
 		mu_reorg_process = FALSE;
 		/* Default threshold is 0 i.e. we attempt to truncate no matter what free_blocks is. */
