@@ -1,6 +1,7 @@
 /****************************************************************
  *								*
- *	Copyright 2003, 2013 Fidelity Information Services, Inc	*
+ * Copyright (c) 2003-2016 Fidelity National Information	*
+ * Services, Inc. and/or its subsidiaries. All rights reserved.	*
  *								*
  *	This source code contains the intellectual property	*
  *	of its copyright holder(s), and is made available	*
@@ -70,13 +71,8 @@ uint4 mur_process_intrpt_recov()
 	uint4				status, status2;
 	uint4				max_autoswitchlimit, max_jnl_alq, max_jnl_deq, freeblks;
 	sgmnt_data_ptr_t		csd;
-	UNIX_ONLY(
-		jnl_private_control	*jpc;
-		jnl_buffer_ptr_t	jbp;
-	)
-	VMS_ONLY(
-		io_status_block_disk	iosb;
-	)
+	jnl_private_control		*jpc;
+	jnl_buffer_ptr_t		jbp;
 	boolean_t			jfh_changed;
 	jnl_record			*jnlrec;
 	jnl_file_header			*jfh;
@@ -236,7 +232,7 @@ uint4 mur_process_intrpt_recov()
 			 * we are guaranteed that all the updates done in the forward processing will have a timestamp that is
 			 * greater than the turn around timestamp
 			 */
-			jbp->prev_jrec_time = jctl->turn_around_time;
+			SET_JNLBUFF_PREV_JREC_TIME(jbp, jctl->turn_around_time, DO_GBL_JREC_TIME_CHECK_FALSE);
 		} else if (dba_bg == csd->acc_meth)
 		{	/* set earliest bt TN to be the turn-around TN (taken from bt_refresh()) */
 			SET_OLDEST_HIST_TN(cs_addrs, cs_addrs->ti->curr_tn - 1);
@@ -290,11 +286,16 @@ uint4 mur_process_intrpt_recov()
 		jfh = jctl->jfh;
 		jfh->turn_around_offset = jctl->turn_around_offset;	/* save progress in file header for 	*/
 		jfh->turn_around_time = jctl->turn_around_time;		/* possible re-issue of recover 	*/
-		UNIX_ONLY(
-			for (idx = 0; idx < MAX_SUPPL_STRMS; idx++)
-				jfh->strm_end_seqno[idx] = csd->strm_reg_seqno[idx];
-		)
+		for (idx = 0; idx < MAX_SUPPL_STRMS; idx++)
+			jfh->strm_end_seqno[idx] = csd->strm_reg_seqno[idx];
 		jfh_changed = TRUE;
+		/* We are about to update the journal file header of the turnaround-point journal file to store the
+		 * non-zero jfh->turn_around_offset. Ensure corresponding database is considered updated.
+		 * This is needed in case journal recovery/rollback terminates abnormally and we go to mur_close_files.
+		 * We need to ensure csd->recov_interrupted does not get reset to FALSE even if this region did not have
+		 * have any updates to the corresponding database file otherwise. (GTM-8394)
+		 */
+		rctl->db_updated = TRUE;
 		for ( ; NULL != jctl; jctl = jctl->next_gen)
 		{	/* setup the next_jnl links. note that in the case of interrupted recovery, next_jnl links
 			 * would have been already set starting from the turn-around point journal file of the
@@ -321,7 +322,7 @@ uint4 mur_process_intrpt_recov()
 				 * recoveries got interrupted in this loop) that need to be reset.
 				 */
 				assert(!jctl->turn_around_offset);
-				assert(rctl->recov_interrupted);	/* rctl->jfh_recov_interrupted can fail */
+				assert(rctl->recov_interrupted || rctl->jctl_apply_pblk); /* rctl->jfh_recov_interrupted can fail */
 				jfh->turn_around_offset = 0;
 				jfh->turn_around_time = 0;
 				jfh_changed = TRUE;
@@ -345,18 +346,16 @@ uint4 mur_process_intrpt_recov()
 							jctl->jnl_fn, jctl->status, PUT_SYS_ERRNO(jctl->status2));
 					return jctl->status;
 				}
-				UNIX_ONLY(
-					GTM_JNL_FSYNC(rctl->csa, jctl->channel, jctl->status);
-					if (-1 == jctl->status)
-					{
-						jctl->status2 = errno;
-						assert(FALSE);
-						gtm_putmsg_csa(CSA_ARG(rctl->csa) VARLSTCNT(9) ERR_JNLFSYNCERR, 2,
-							jctl->jnl_fn_len, jctl->jnl_fn,
-							ERR_TEXT, 2, RTS_ERROR_TEXT("Error with fsync"), jctl->status2);
-						return ERR_JNLFSYNCERR;
-					}
-				)
+				GTM_JNL_FSYNC(rctl->csa, jctl->channel, jctl->status);
+				if (-1 == jctl->status)
+				{
+					jctl->status2 = errno;
+					assert(FALSE);
+					gtm_putmsg_csa(CSA_ARG(rctl->csa) VARLSTCNT(9) ERR_JNLFSYNCERR, 2,
+						jctl->jnl_fn_len, jctl->jnl_fn,
+						ERR_TEXT, 2, RTS_ERROR_TEXT("Error with fsync"), jctl->status2);
+					return ERR_JNLFSYNCERR;
+				}
 			}
 			jfh_changed = FALSE;
 		}
