@@ -1,6 +1,7 @@
 /****************************************************************
  *								*
- *	Copyright 2001, 2014 Fidelity Information Services, Inc	*
+ * Copyright (c) 2001-2016 Fidelity National Information	*
+ * Services, Inc. and/or its subsidiaries. All rights reserved.	*
  *								*
  *	This source code contains the intellectual property	*
  *	of its copyright holder(s), and is made available	*
@@ -44,7 +45,7 @@
 #include "gvcmz.h"
 #include "testpt.h"
 #include "mvalconv.h"
-#include "dpgbldir.h"	/* for get_next_gdr() prototype */
+#include "dpgbldir.h"		/* For get_next_gdr() prototype */
 #include "ast.h"
 #include "wcs_flu.h"
 #include "stringpool.h"
@@ -57,7 +58,7 @@
 #include "targ_alloc.h"
 #include "gvcst_protos.h"
 #ifdef GTM_TRIGGER
-# include "rtnhdr.h"		/* for rtn_tabent in gv_trigger.h */
+# include "rtnhdr.h"		/* For rtn_tabent in gv_trigger.h */
 # include "gv_trigger.h"
 # include "gtm_trigger.h"
 #endif
@@ -69,12 +70,13 @@
 #  include "gtmsecshr.h"
 # endif
 #endif
-#if defined(DEBUG) && defined(USHBIN_SUPPORTED)
+#ifdef AUTORELINK__SUPPORTED
 # include "relinkctl.h"
 #endif
 #include "gtmimagename.h"
+#include "cache.h"
 
-STATICFNDCL void view_dbflushop(unsigned char keycode, viewparm *parmblkptr, mval *thirdarg);
+STATICFNDCL void view_dbop(unsigned char keycode, viewparm *parmblkptr, mval *thirdarg);
 
 GBLREF	volatile int4 		db_fsync_in_prog;
 GBLREF	boolean_t		certify_all_blocks;
@@ -91,9 +93,9 @@ GBLREF	gv_namehead		*reset_gv_target;
 GBLREF	sgmnt_addrs		*cs_addrs;
 GBLREF	sgmnt_data_ptr_t	cs_data;
 GBLREF	symval			*curr_symval;
-GBLREF	trans_num		local_tn;	/* transaction number for THIS PROCESS */
+GBLREF	trans_num		local_tn;	/* Transaction number for THIS PROCESS */
 GBLREF	int4			zdir_form;
-GBLREF	boolean_t		gvdupsetnoop; /* if TRUE, duplicate SETs update journal but not database (except for curr_tn++) */
+GBLREF	boolean_t		gvdupsetnoop;	/* If TRUE, duplicate SETs update journal but not database (except for curr_tn++) */
 GBLREF	boolean_t		badchar_inhibit;
 GBLREF	int			gv_fillfactor;
 GBLREF	symval			*curr_symval;
@@ -130,7 +132,7 @@ error_def(ERR_ZDEFACTIVE);
 
 #define WRITE_LITERAL(x) (outval.str.len = SIZEOF(x) - 1, outval.str.addr = (x), op_write(&outval))
 
-/* if changing noisolation status within TP and already referenced the global, then error */
+/* If changing noisolation status within TP and already referenced the global, then error */
 #define SET_GVNH_NOISOLATION_STATUS(gvnh, status)								\
 {														\
 	GBLREF	uint4			dollar_tlevel;								\
@@ -144,57 +146,52 @@ error_def(ERR_ZDEFACTIVE);
 
 void	op_view(UNIX_ONLY_COMMA(int numarg) mval *keyword, ...)
 {
-	int4			testvalue, tmpzdefbufsiz;
-	uint4			jnl_status, dummy_errno;
-	int			status, lcnt, icnt, recnum;
+	boolean_t		dbgdmpenabled, old_bool, was_crit, was_skip_gtm_putmsg;
+	char			*chptr;
+	collseq			*new_lcl_collseq;
+	gd_addr			*addr_ptr;
 	gd_region		*reg, *r_top, *save_reg;
+	gvnh_reg_t		*gvnh_reg;
+	gvnh_spanreg_t		*gvspan;
 	gv_namehead		*gvnh;
-	mval			*arg, *nextarg, outval;
+	hash_table_mname	*table;
+	ht_ent_mname		*tabent, *table_base_orig, *topent;
+	int			clrlen, lcnt, lct, icnt, ncol, nct, status, table_size_orig;
+	int4			testvalue, tmpzdefbufsiz;
+	jnl_buffer_ptr_t	jb;
+	lv_blk			*lvbp;
+	lv_val			*lv, *lvp, *lvp_top;
 	mstr			tmpstr;
+	mval			*arg, *nextarg, outval;
+	noisolation_element	*gvnh_entry;
+	sgmnt_addrs		*csa;
+	sgmnt_data_ptr_t	csd;
+	symval			*cstab, *lvlsymtab;
+	uint4			jnl_status, dummy_errno;
 	va_list			var;
 	viewparm		parmblk, parmblk2;
 	viewtab_entry		*vtp;
-	gd_addr			*addr_ptr;
-	gvnh_reg_t		*gvnh_reg;
-	gvnh_spanreg_t		*gvspan;
-	noisolation_element	*gvnh_entry;
-	int			lct, ncol, nct;
-	collseq			*new_lcl_collseq;
-	ht_ent_mname		*tabent, *topent;
-	lv_val			*lv;
-	symval			*cstab;
-	sgmnt_addrs		*csa;
-	sgmnt_data_ptr_t	csd;
-	jnl_buffer_ptr_t	jb;
-	int			table_size_orig;
-	ht_ent_mname		*table_base_orig;
-	hash_table_mname	*table;
-	boolean_t		dbgdmpenabled, was_crit, was_skip_gtm_putmsg;
-	symval			*lvlsymtab;
-	lv_blk			*lvbp;
-	lv_val			*lvp, *lvp_top;
-#	if defined(DEBUG) && defined(USHBIN_SUPPORTED)
+#	ifdef AUTORELINK_SUPPORTED
 	open_relinkctl_sgm	*linkctl;
 	relinkrec_t		*linkrec;
 #	endif
 	VMS_ONLY(int		numarg;)
-
 	static readonly char msg1[] = "Caution: Database Block Certification Has Been ";
-	static readonly char msg2[] = "Disabled";
-	static readonly char msg3[] = "Enabled";
 	static readonly char lv_msg1[] =
 		"Caution: GT.M reserved local variable string pointer duplicate check diagnostic has been";
-	static readonly char upper[] = "UPPER";
+	static readonly char msg2[] = "Disabled";
+	static readonly char msg3[] = "Enabled";
 	static readonly char lower[] = "LOWER";
+	static readonly char upper[] = "UPPER";
 	DCL_THREADGBL_ACCESS;
 
 	SETUP_THREADGBL_ACCESS;
 	VAR_START(var, keyword);
 	VMS_ONLY(va_count(numarg));
-		jnl_status = 0;
+	jnl_status = 0;
 	assertpro(1 <= numarg);
 	MV_FORCE_STR(keyword);
-	numarg--;	/* remove keyword from count */
+	numarg--;	/* Remove keyword from count */
 	if (0 < numarg)
 	{
 		arg = va_arg(var, mval *);
@@ -232,13 +229,14 @@ void	op_view(UNIX_ONLY_COMMA(int numarg) mval *keyword, ...)
 			outval.mvtype = MV_STR;
 			view_debug4 = (0 != MV_FORCE_INT(parmblk.value));
 			break;
+		case VTK_DBFLUSH:
 		case VTK_DBSYNC:
 		case VTK_EPOCH:
 		case VTK_FLUSH:
+		case VTK_GVSRESET:
 		case VTK_JNLFLUSH:
-		case VTK_DBFLUSH:
-			arg = (numarg > 1) ? va_arg(var, mval *) : NULL;
-			view_dbflushop(vtp->keycode, &parmblk, arg);
+		case VTK_POOLLIMIT:
+			view_dbop(vtp->keycode, &parmblk, (numarg > 1) ? va_arg(var, mval *) : (mval *)NULL);
 			break;
 #		ifdef UNIX
 		case VTK_DMTERM:
@@ -249,32 +247,16 @@ void	op_view(UNIX_ONLY_COMMA(int numarg) mval *keyword, ...)
 			break;
 #		endif
 		case VTK_FULLBOOL:
-			TREF(gtm_fullbool) = FULL_BOOL;
-			break;
 		case VTK_FULLBOOLWARN:
-			TREF(gtm_fullbool) = FULL_BOOL_WARN;
-			break;
 		case VTK_NOFULLBOOL:
-			if (OLD_SE == TREF(side_effect_handling))
-				TREF(gtm_fullbool) = GTM_BOOL;
-			else
+
+			if ((VTK_NOFULLBOOL == vtp->keycode) && (OLD_SE != TREF(side_effect_handling)))
 				rts_error_csa(CSA_ARG(NULL) VARLSTCNT(1) ERR_SEFCTNEEDSFULLB);
-			break;
-		case VTK_GDSCERT0:
-			outval.mvtype = MV_STR;
-			op_wteol(1);
-			WRITE_LITERAL(msg1);
-			certify_all_blocks = FALSE;
-			WRITE_LITERAL(msg2);
-			op_wteol(1);
-			break;
-		case VTK_GDSCERT1:
-			outval.mvtype = MV_STR;
-			op_wteol(1);
-			WRITE_LITERAL(msg1);
-			certify_all_blocks = TRUE;
-			WRITE_LITERAL(msg3);
-			op_wteol(1);
+			old_bool = TREF(gtm_fullbool);
+			TREF(gtm_fullbool) = (VTK_FULLBOOL == vtp->keycode) ? FULL_BOOL
+				: (VTK_FULLBOOLWARN == vtp->keycode) ? FULL_BOOL_WARN : GTM_BOOL;
+			if (old_bool != TREF(gtm_fullbool))
+				cache_table_rebuild();
 			break;
 		case VTK_GDSCERT:
 			outval.mvtype = MV_STR;
@@ -317,39 +299,13 @@ void	op_view(UNIX_ONLY_COMMA(int numarg) mval *keyword, ...)
 				TREF(error_on_jnl_file_lost) = MV_FORCE_INT(parmblk.value);
 				if (MAX_JNL_FILE_LOST_OPT < TREF(error_on_jnl_file_lost))
 					TREF(error_on_jnl_file_lost) = JNL_FILE_LOST_TURN_OFF;
-				if (NULL == gd_header)		/* open gbldir */
-					gvinit();
-				save_reg = gv_cur_region;
-				/* change all regions */
-				reg = gd_header->regions;
-				r_top = reg + gd_header->n_regions - 1;
-				for (;  reg <= r_top;  reg++)
-				{
-					if (!reg->open)
-						gv_init_reg(reg);
-					if (!reg->read_only)
-					{
-						gv_cur_region = reg;
-						change_reg();
-						csa = cs_addrs;
-						csd = csa->hdr;
-						if (JNL_ENABLED(csd))
-						{
-							was_crit = csa->now_crit;
-							if (!was_crit)
-								grab_crit(reg);
-							if (JNL_ENABLED(csd))
-								csa->jnl->error_reported = FALSE;
-							if (!was_crit)
-								rel_crit(reg);
-						}
-					}
-				}
+				parmblk.gv_ptr = NULL;
+				view_dbop(vtp->keycode, &parmblk, (mval *)NULL);
 			}
 			break;
 #		endif
 		case VTK_JNLWAIT:
-			/* go through all regions that could have possibly been open across all global directories */
+			/* Go through all regions that could have possibly been open across all global directories */
 			if (!dollar_tlevel)
 			{	/* Only if we're not in a TP transaction */
 				for (addr_ptr = get_next_gdr(NULL); addr_ptr; addr_ptr = get_next_gdr(addr_ptr))
@@ -442,8 +398,9 @@ void	op_view(UNIX_ONLY_COMMA(int numarg) mval *keyword, ...)
 				rts_error_csa(CSA_ARG(NULL) VARLSTCNT(3) ERR_ACTRANGE, 1, lct);
 			}
 			was_skip_gtm_putmsg = TREF(skip_gtm_putmsg);
-			TREF(skip_gtm_putmsg) = TRUE;	/* to avoid ready_collseq from doing gtm_putmsg in case of errors.
-							 * not doing so will cause GDECHECK errors in caller (GDE). */
+			TREF(skip_gtm_putmsg) = TRUE;	/* To avoid ready_collseq from doing gtm_putmsg in case of errors.
+							 * not doing so will cause GDECHECK errors in caller (GDE).
+							 */
 			new_lcl_collseq = ready_collseq(lct);
 			TREF(skip_gtm_putmsg) = was_skip_gtm_putmsg;
 			if (0 == new_lcl_collseq)
@@ -487,7 +444,7 @@ void	op_view(UNIX_ONLY_COMMA(int numarg) mval *keyword, ...)
 				view_arg_convert(vtp, VTP_DBREGION, arg, &parmblk2, IS_DOLLAR_VIEW_FALSE);
 				reg = parmblk2.gv_ptr;
 				/* Determine if "reg" is mapped to by global name. If not issue error */
-				gvnh_reg = TREF(gd_targ_gvnh_reg);	/* set up by op_gvname */
+				gvnh_reg = TREF(gd_targ_gvnh_reg);	/* Set up by op_gvname */
 				gvspan = (NULL == gvnh_reg) ? NULL : gvnh_reg->gvspan;
 				if (((NULL != gvspan) && !gvnh_spanreg_ismapped(gvnh_reg, gd_header, reg))
 					|| ((NULL == gvspan) && (reg != gv_cur_region)))
@@ -546,7 +503,7 @@ void	op_view(UNIX_ONLY_COMMA(int numarg) mval *keyword, ...)
 					rts_error_csa(CSA_ARG(NULL) VARLSTCNT(3) ERR_ACTRANGE, 1, lct);
 				}
 			}
-			/* at this point, verify that there is no local data with subscripts */
+			/* At this point, verify that there is no local data with subscripts */
 			for (cstab = curr_symval; cstab; cstab = cstab->last_tab)
 			{
 				assert(cstab->h_symtab.top == cstab->h_symtab.base + cstab->h_symtab.size);
@@ -619,9 +576,9 @@ void	op_view(UNIX_ONLY_COMMA(int numarg) mval *keyword, ...)
 			tmpzdefbufsiz = MV_FORCE_INT(parmblk.value);
 			if (1 == tmpzdefbufsiz)
 				tmpzdefbufsiz = ZDEFDEF;
-			else  if (tmpzdefbufsiz < ZDEFMIN)
+			else if (tmpzdefbufsiz < ZDEFMIN)
 				tmpzdefbufsiz = ZDEFMIN;
-			else  if (tmpzdefbufsiz > ZDEFMAX)
+			else if (tmpzdefbufsiz > ZDEFMAX)
 				tmpzdefbufsiz = ZDEFMAX;
 			zdefbufsiz = (unsigned short)tmpzdefbufsiz;
 			break;
@@ -675,14 +632,15 @@ void	op_view(UNIX_ONLY_COMMA(int numarg) mval *keyword, ...)
 			gv_fillfactor = testvalue;
 			break;
 		case VTK_STPGCOL:
-			INVOKE_STP_GCOL(INTCAST(stringpool.top - stringpool.free) + 1);/* computation to avoid assert in stp_gcol */
+			INVOKE_STP_GCOL(INTCAST(stringpool.top - stringpool.free) + 1);/* Computation to avoid assert in stp_gcol */
 			break;
 		case VTK_LVGCOL:
 			als_lvval_gc();
 			break;
 		case VTK_LVREHASH:
 			/* This doesn't actually expand or contract the local variable hash table but does cause it to be
-			 * rebuilt. Then we need to do the same sort of cleanup that add_hashtab_mname_symval does. */
+			 * rebuilt. Then we need to do the same sort of cleanup that add_hashtab_mname_symval does.
+			 */
 			/* Step 1: remember table we started with */
 			table = &curr_symval->h_symtab;
 			table_base_orig = table->base;
@@ -702,7 +660,7 @@ void	op_view(UNIX_ONLY_COMMA(int numarg) mval *keyword, ...)
 		case VTK_STORDUMP:
 #			if defined(DEBUG) && defined(UNIX)
 			if (gtm_white_box_test_case_enabled
-			    && (WBTEST_HOLD_CRIT_TILL_LCKALERT == gtm_white_box_test_case_number))
+				&& (WBTEST_HOLD_CRIT_TILL_LCKALERT == gtm_white_box_test_case_number))
 			{	/* Hold crit for a long enough interval to generate lock alert which then does a continue_proc */
 				grab_crit(gv_cur_region);
 				icnt = TREF(continue_proc_cnt);
@@ -748,8 +706,7 @@ void	op_view(UNIX_ONLY_COMMA(int numarg) mval *keyword, ...)
 			{
 				if (!TREF(tprestart_syslog_delta))
 					TREF(tprestart_syslog_delta) = 1;
-			}
-			else
+			} else
 			{
 				TREF(tprestart_syslog_delta) = MV_FORCE_INT(parmblk.value);
 				if (0 > TREF(tprestart_syslog_delta))
@@ -760,6 +717,22 @@ void	op_view(UNIX_ONLY_COMMA(int numarg) mval *keyword, ...)
 		case VTK_NOLOGTPRESTART:
 			TREF(tprestart_syslog_delta) = 0;
 			break;
+		case VTK_LOGNONTP:
+			if (!numarg)
+			{
+				if (!TREF(nontprestart_log_delta))
+					TREF(nontprestart_log_delta) = 1;
+			} else
+			{
+				TREF(nontprestart_log_delta) = MV_FORCE_INT(parmblk.value);
+				if (0 > TREF(nontprestart_log_delta))
+					TREF(nontprestart_log_delta) = 0;
+			}
+			TREF(nontprestart_count) = 0;
+			break;
+		case VTK_NOLOGNONTP:
+			TREF(nontprestart_log_delta) = 0;
+			break;
 #		ifdef UNIX
 		case VTK_LINK:
 			init_relink_allowed(&parmblk.value->str);
@@ -796,35 +769,14 @@ void	op_view(UNIX_ONLY_COMMA(int numarg) mval *keyword, ...)
 			lv = (lv_val *)parmblk.value;
 			util_out_print("", RESET);	/* Reset the buffer */
 			util_out_print("LV: !AD  addr: 0x!XJ  mvtype: 0x!4XW  sign: !UB  exp: !UL  m[0]: !UL [0x!XL]  "
-				       "m[1]: !UL [0x!XL]  str.len: !UL  str.addr: 0x!XJ", SPRINT, arg->str.len, arg->str.addr,
-				       lv, lv->v.mvtype, lv->v.sgn, lv->v.e, lv->v.m[0], lv->v.m[0], lv->v.m[1], lv->v.m[1],
-				       lv->v.str.len, lv->v.str.addr);
+					"m[1]: !UL [0x!XL]  str.len: !UL  str.addr: 0x!XJ", SPRINT, arg->str.len, arg->str.addr,
+					lv, lv->v.mvtype, lv->v.sgn, lv->v.e, lv->v.m[0], lv->v.m[0], lv->v.m[1], lv->v.m[1],
+					lv->v.str.len, lv->v.str.addr);
 			outval.str.addr = TREF(util_outptr);
 			outval.str.len = STRLEN(TREF(util_outptr));
 			op_write(&outval);
 			op_wteol(1);
 			break;
-#		endif
-#		if defined(DEBUG) && defined(USHBIN_SUPPORTED)
-		case VTK_RCTLDUMP:
-			util_out_print("", RESET);	/* Reset output buffer */
-			for (linkctl = TREF(open_relinkctl_list); NULL != linkctl; linkctl = linkctl->next)
-			{
-				util_out_print_gtmio("relinkctl: 0x!XJ  entryname: !AD  records: !UL  hdr: 0x!XJ  base: 0x!XJ"
-						     "  locked: !UL", FLUSH, linkctl, linkctl->zro_entry_name.len,
-						     linkctl->zro_entry_name.addr, linkctl->n_records, linkctl->hdr,
-						     linkctl->rec_base, linkctl->locked);
-				util_out_print_gtmio("  hdr: records: !UL  nattached: !UL", FLUSH, linkctl->hdr->n_records,
-						     linkctl->hdr->nattached);
-				for (linkrec = linkctl->rec_base, recnum = 1; recnum <= linkctl->hdr->n_records;
-				     linkrec++, recnum++)
-				{
-					util_out_print_gtmio("    rec#!4UL: rtnname: !AD  cycle: !UL", FLUSH, recnum,
-							     mid_len(&linkrec->rtnname_fixed), &linkrec->rtnname_fixed.c,
-							     linkrec->cycle);
-				}
-			}
-			break;
 #		endif
 		default:
 			va_end(var);
@@ -834,40 +786,45 @@ void	op_view(UNIX_ONLY_COMMA(int numarg) mval *keyword, ...)
 	return;
 }
 
-void view_dbflushop(unsigned char keycode, viewparm *parmblkptr, mval *thirdarg)
+void view_dbop(unsigned char keycode, viewparm *parmblkptr, mval *thirdarg)
 {
-	uint4			jnl_status, dummy_errno;
-	int4			nbuffs;
-	int			save_errno;
-	int			status, lcnt, icnt;
+	boolean_t		was_crit;
 	gd_region		*reg, *r_top, *save_reg;
+	int			icnt, lcnt, save_errno, status;
+	int4			nbuffs;
+	jnl_buffer_ptr_t	jb;
 	sgmnt_addrs		*csa;
 	sgmnt_data_ptr_t	csd;
-	jnl_buffer_ptr_t	jb;
-	boolean_t		was_crit;
+	uint4			jnl_status, dummy_errno;
 	UNIX_ONLY(unix_db_info	*udi;)
 
-	if (NULL == gd_header)		/* open gbldir */
+	if (NULL == gd_header)		/* Open gbldir */
 		gvinit();
 	save_reg = gv_cur_region;
 	if (NULL == parmblkptr->gv_ptr)
-	{	/* operate on all regions */
+	{	/* Operate on all regions */
 		reg = gd_header->regions;
 		r_top = reg + gd_header->n_regions - 1;
-	} else	/* operate on selected region */
+	} else	/* Operate on selected region */
 		r_top = reg = parmblkptr->gv_ptr;
 	for (; reg <= r_top; reg++)
 	{
 		if (!reg->open)
 			gv_init_reg(reg);
-		gv_cur_region = reg;
+		TP_CHANGE_REG(reg);
 		switch(keycode)
 		{
+			case VTK_DBFLUSH:
+				if (!reg->read_only)
+				{
+					nbuffs = (NULL != thirdarg) ? MV_FORCE_INT(thirdarg) : cs_addrs->nl->wcs_active_lvl;
+					JNL_ENSURE_OPEN_WCS_WTSTART(cs_addrs, reg, nbuffs, dummy_errno);
+				}
+				break;
 			case VTK_DBSYNC:
 #				ifdef UNIX
 				if (!reg->read_only)
 				{
-					change_reg();
 					csa = cs_addrs;
 					udi = FILE_INFO(reg);
 					DB_FSYNC(reg, udi, csa, db_fsync_in_prog, save_errno);
@@ -880,11 +837,10 @@ void view_dbflushop(unsigned char keycode, viewparm *parmblkptr, mval *thirdarg)
 				}
 #				endif
 				break;
-			case VTK_FLUSH:
 			case VTK_EPOCH:
+			case VTK_FLUSH:
 				if (!reg->read_only)
 				{
-					change_reg(); /* for jnl_ensure_open */
 					ENSURE_JNL_OPEN(cs_addrs, gv_cur_region);
 					/* We should NOT invoke wcs_recover here because it's possible we are in the final retry
 					 * of a TP transaction. In this case, we likely have pointers to non-dirty global buffers
@@ -896,8 +852,32 @@ void view_dbflushop(unsigned char keycode, viewparm *parmblkptr, mval *thirdarg)
 					wcs_flu(WCSFLU_FLUSH_HDR | WCSFLU_WRITE_EPOCH | WCSFLU_IN_COMMIT | WCSFLU_SPEEDUP_NOBEFORE);
 				}
 				break;
-			case VTK_JNLFLUSH:
+			case VTK_GVSRESET:
 				change_reg();
+				if (!reg->read_only)
+					CLRGVSTATS(cs_addrs);
+				memset((char *)&cs_addrs->gvstats_rec, 0, SIZEOF(gvstats_rec_t));	/* Always process-private */
+				break;
+#			ifndef VMS
+			case VTK_JNLERROR:
+				if (!reg->read_only)
+				{
+					csa = cs_addrs;
+					csd = csa->hdr;
+					if (JNL_ENABLED(csd))
+					{
+						was_crit = csa->now_crit;
+						if (!was_crit)
+							grab_crit(reg);
+						if (JNL_ENABLED(csd))
+							csa->jnl->error_reported = FALSE;
+						if (!was_crit)
+							rel_crit(reg);
+					}
+				}
+				break;
+#				endif
+			case VTK_JNLFLUSH:
 				csa = cs_addrs;
 				csd = csa->hdr;
 				if (JNL_ENABLED(csd))
@@ -932,13 +912,17 @@ void view_dbflushop(unsigned char keycode, viewparm *parmblkptr, mval *thirdarg)
 						rel_crit(reg);
 				}
 				break;
-			case VTK_DBFLUSH:
-				if (!reg->read_only)
-				{
-					change_reg(); /* for jnl_ensure_open */
-					nbuffs = (NULL != thirdarg) ? MV_FORCE_INT(thirdarg) : cs_addrs->nl->wcs_active_lvl;
-					JNL_ENSURE_OPEN_WCS_WTSTART(cs_addrs, reg, nbuffs, dummy_errno);
-				}
+			case VTK_POOLLIMIT:
+				csa = cs_addrs;
+				csd = csa->hdr;
+				nbuffs = ((dba_bg == csd->acc_meth) && (NULL != thirdarg)) ? MV_FORCE_INT(thirdarg) : 0;
+				if (nbuffs && (MV_STR & thirdarg->mvtype) && ('%' == thirdarg->str.addr[thirdarg->str.len - 1]))
+					nbuffs = (100 == nbuffs) ? 0 : (csd->n_bts * nbuffs) / 100;		/* Percentage */
+				csa->gbuff_limit = (0 == nbuffs) ? 0 : MAX(MIN(nbuffs, csd->n_bts * .5), MIN_GBUFF_LIMIT);
+				/* To pick the current "hand" as a pseudo-random spot for our area see dbg code in gvcst_init
+				 * but for the first release of this always pick the end of the buffer
+				 */
+				csa->our_midnite = csa->acc_meth.bg.cache_state->cache_array + csd->bt_buckets + csd->n_bts;
 				break;
 		}
 	}
