@@ -1,6 +1,7 @@
 /****************************************************************
  *								*
- *	Copyright 2001, 2014 Fidelity Information Services, Inc	*
+ * Copyright (c) 2001-2015 Fidelity National Information 	*
+ * Services, Inc. and/or its subsidiaries. All rights reserved.	*
  *								*
  *	This source code contains the intellectual property	*
  *	of its copyright holder(s), and is made available	*
@@ -36,6 +37,7 @@ void op_zshow(mval *func, int type, lv_val *lvn)
 {
 	const char	*ptr;
 	boolean_t	do_all = FALSE,
+			done_a = FALSE,
 			done_b = FALSE,
 			done_c = FALSE,
 			done_d = FALSE,
@@ -55,6 +57,8 @@ void op_zshow(mval *func, int type, lv_val *lvn)
 	{
 		switch (*ptr)
 		{
+			case 'A':
+			case 'a':
 			case 'B':
 			case 'b':
 			case 'C':
@@ -97,6 +101,7 @@ void op_zshow(mval *func, int type, lv_val *lvn)
 	{
 		output.out_var.gv.end = gv_currkey->end;
 		output.out_var.gv.prev = gv_currkey->prev;
+		output.line_cont = 0;	/* currently only used for global output and only initialized here */
 	}
 	MAXSTR_BUFF_INIT;
 	output.type = type;
@@ -107,7 +112,16 @@ void op_zshow(mval *func, int type, lv_val *lvn)
 	{
 		output.line_num = 1;
 		switch (*ptr)
-		{	case 'B':
+		{
+			case 'A':
+			case 'a':
+				if (done_a)
+					break;
+				done_a = TRUE;
+				output.code = 'A';
+				ARLINK_ONLY(zshow_rctldump(&output));
+				break;
+			case 'B':
 			case 'b':
 				if (done_b)
 					break;
@@ -190,6 +204,6 @@ void op_zshow(mval *func, int type, lv_val *lvn)
 	/* If ZSHOW was done onto a subscripted lvn but no zshow records got dumped in that lvn, it might have $data = 0.
 	 * Kill it in that case as otherwise it will create an out-of-design situation for $query(lvn).
 	 */
-	if ((type == ZSHOW_LOCAL) && (active_lv == lvn))
+	if ((type == ZSHOW_LOCAL) && (NULL != active_lv) && lcl_arg1_is_desc_of_arg2(active_lv, lvn))
 		UNDO_ACTIVE_LV(actlv_op_zshow);
 }
