@@ -1,6 +1,7 @@
 /****************************************************************
  *								*
- *	Copyright 2001, 2013 Fidelity Information Services, Inc	*
+ * Copyright (c) 2001-2016 Fidelity National Information	*
+ * Services, Inc. and/or its subsidiaries. All rights reserved.	*
  *								*
  *	This source code contains the intellectual property	*
  *	of its copyright holder(s), and is made available	*
@@ -201,7 +202,7 @@ cw_set_element *t_write (
 			 * access the buffer more than the db blk_size.
 			 */
 			bsiz = MIN(old_block->bsiz, csa->hdr->blk_size);
-			cse->blk_checksum = jnl_get_checksum((uint4*)old_block, csa, bsiz);
+			cse->blk_checksum = jnl_get_checksum(old_block, csa, bsiz);
 		}
 		/* the buffer in shared memory holding the GDS block contents currently does not have in its block header the
 		 * on-disk format of that block. if it had, we could have easily copied that over to the cw-set-element.
