@@ -1,6 +1,7 @@
 /****************************************************************
  *								*
- *	Copyright 2001, 2014 Fidelity Information Services, Inc	*
+ * Copyright (c) 2001-2015 Fidelity National Information 	*
+ * Services, Inc. and/or its subsidiaries. All rights reserved.	*
  *								*
  *	This source code contains the intellectual property	*
  *	of its copyright holder(s), and is made available	*
@@ -45,6 +46,7 @@
 #include "op.h"
 #include "gt_timer.h"
 #include "iosocketdef.h"
+#include "have_crit.h"
 #ifdef UNIX
 #include "iormdef.h"
 #include "iottdef.h"
@@ -75,6 +77,7 @@ GBLREF lv_xnew_var		*xnewvar_anchor;
 #ifdef GTM_TRIGGER
 GBLREF mstr			*dollar_ztname;
 GBLREF mval			*dollar_ztdata;
+GBLREF mval			*dollar_ztdelim;
 GBLREF mval			*dollar_ztoldval;
 GBLREF mval			*dollar_ztriggerop;
 GBLREF mval			*dollar_ztupdate;
@@ -117,6 +120,7 @@ void unw_mv_ent(mv_stent *mv_st_ent)
 	socket_interrupt 	*sockintr;
 	socket_struct		*socketptr;
 	zintcmd_ops		zintcmd_command;
+	intrpt_state_t		prev_intrpt_state;
 	UNIX_ONLY(d_tt_struct	*tt_ptr;)
 	DBGRFCT_ONLY(mident_fixed vname;)
 	DCL_THREADGBL_ACCESS;
@@ -443,6 +447,7 @@ void unw_mv_ent(mv_stent *mv_st_ent)
 			dollar_ztvalue = mv_st_ent->mv_st_cont.mvs_trigr.ztvalue_save;
 			dollar_ztname = mv_st_ent->mv_st_cont.mvs_trigr.ztname_save;
 			dollar_ztdata = mv_st_ent->mv_st_cont.mvs_trigr.ztdata_save;
+			dollar_ztdelim = mv_st_ent->mv_st_cont.mvs_trigr.ztdelim_save;
 			dollar_ztoldval = mv_st_ent->mv_st_cont.mvs_trigr.ztoldval_save;
 			dollar_ztriggerop = mv_st_ent->mv_st_cont.mvs_trigr.ztriggerop_save;
 			dollar_ztupdate = mv_st_ent->mv_st_cont.mvs_trigr.ztupdate_save;
@@ -460,20 +465,24 @@ void unw_mv_ent(mv_stent *mv_st_ent)
 				dollar_ztrap = mv_st_ent->mv_st_cont.mvs_trigr.dollar_ztrap_save;
 				ztrap_explicit_null = mv_st_ent->mv_st_cont.mvs_trigr.ztrap_explicit_null_save;
 			}
+			DEFER_INTERRUPTS(INTRPT_IN_CONDSTK, prev_intrpt_state);
 			CHECKHIGHBOUND(mv_st_ent->mv_st_cont.mvs_trigr.ctxt_save);
 			CHECKLOWBOUND(mv_st_ent->mv_st_cont.mvs_trigr.ctxt_save);
 			ctxt = mv_st_ent->mv_st_cont.mvs_trigr.ctxt_save;
 			/* same assert as in gtm_trigger.c */
 			assert(((0 == gtm_trigger_depth)
-				&& (((ch_at_trigger_init == ctxt->ch)
-				     || ((ch_at_trigger_init == (ctxt - 1)->ch)
-					 && ((&gvcst_put_ch == ctxt->ch) || (&gvcst_kill_ch == ctxt->ch))))))
-			       || ((0 < gtm_trigger_depth)
-				   && (((&mdb_condition_handler == ctxt->ch)
-					|| ((&mdb_condition_handler == (ctxt - 1)->ch)
-					    && ((&gvcst_put_ch == ctxt->ch) || (&gvcst_kill_ch == ctxt->ch)))))));
+					&& (((ch_at_trigger_init == ctxt->ch)
+						|| ((ch_at_trigger_init == (ctxt - 1)->ch)
+							&& ((&gvcst_put_ch == ctxt->ch) || (&gvcst_kill_ch == ctxt->ch)
+								|| (&gvcst_spr_kill_ch == ctxt->ch))))))
+				|| ((0 < gtm_trigger_depth)
+					&& (((&mdb_condition_handler == ctxt->ch)
+						|| ((&mdb_condition_handler == (ctxt - 1)->ch)
+							&& ((&gvcst_put_ch == ctxt->ch) || (&gvcst_kill_ch == ctxt->ch)
+								|| (&gvcst_spr_kill_ch == ctxt->ch)))))));
 			active_ch = ctxt;
 			ctxt->ch_active = FALSE;
+			ENABLE_INTERRUPTS(INTRPT_IN_CONDSTK, prev_intrpt_state);
 			if (tp_timeout_deferred && !((0 < dollar_ecode.index) && (ETRAP_IN_EFFECT))
 			    && !dollar_zininterrupt)
 			{	/* A tp timeout was deferred. Now that $ETRAP is no longer in effect and/or we are no
