@@ -1,6 +1,7 @@
 /****************************************************************
  *								*
- *	Copyright 2005, 2014 Fidelity Information Services, Inc.	*
+ * Copyright (c) 2005-2016 Fidelity National Information	*
+ * Services, Inc. and/or its subsidiaries. All rights reserved.	*
  *								*
  *	This source code contains the intellectual property	*
  *	of its copyright holder(s), and is made available	*
@@ -72,6 +73,7 @@
 #include "performcaslatchcheck.h"
 #include "updproc_get_gblname.h"
 #include "gtmmsg.h"
+#include "mupip_reorg_encrypt.h"
 
 #ifdef REPL_DEBUG
 #include "format_targ_key.h"
@@ -109,7 +111,10 @@ GBLREF 	sgmnt_data_ptr_t 	cs_data;
 GBLREF  uint4                   image_count;
 #endif
 GBLREF boolean_t		disk_blk_read;
-LITREF	mval			literal_hasht;
+#ifdef DEBUG
+GBLREF	sgmnt_addrs		*reorg_encrypt_restart_csa;
+#endif
+
 static	uint4			last_pre_read_offset;
 
 error_def(ERR_DBCCERR);
@@ -171,10 +176,10 @@ boolean_t updproc_preread(void)
 	gtmrecv_local_ptr_t	gtmrecv_local;
 	upd_helper_ctl_ptr_t	upd_helper_ctl;
 	gvnh_reg_t		*gvnh_reg;
-#ifdef REPL_DEBUG
+#	ifdef REPL_DEBUG
 	unsigned char 		buff[MAX_ZWR_KEY_SZ], *end;
 	uint4			write, write_wrap;
-#endif
+#	endif
 	DCL_THREADGBL_ACCESS;
 
 	SETUP_THREADGBL_ACCESS;
@@ -202,7 +207,7 @@ boolean_t updproc_preread(void)
 			return TRUE;
 		}
 		for (retries = LOCK_TRIES - 1; 0 < retries; retries--)	/* - 1 so do rel_quant 3 times first */
-		{
+		{	/* seems like this might be a legitimate spin lock - might could use some work to tighten it up */
 			for (spins = maxspins; 0 < spins; spins--)
 			{
 				if (GET_SWAPLOCK(&upd_helper_ctl->pre_read_lock))
@@ -216,7 +221,8 @@ boolean_t updproc_preread(void)
 						return FALSE;
 					/* On every 4th pass, we bide for awhile */
 					wcs_sleep(LOCK_SLEEP);
-					if (RETRY_CASLATCH_CUTOFF == retries)
+					/* Check if we're due to check for lock abandonment check or holder wakeup */
+					if (0 == (retries & (LOCK_CASLATCH_CHKINTVL - 1)))
 						performCASLatchCheck(&upd_helper_ctl->pre_read_lock, TRUE);
 				}
 			}
@@ -229,10 +235,10 @@ boolean_t updproc_preread(void)
 					ERR_ERRCALL, 3, CALLFROM);
 			return FALSE;
 		}
-#ifdef REPL_DEBUG
+#		ifdef REPL_DEBUG
 		write_wrap = recvpool_ctl->write_wrap;
 		write = recvpool_ctl->write;
-#endif
+#		endif
 		if (pre_read_offset >= recvpool_ctl->write_wrap)
 		{
 			REPL_DPRINT4("Wrapped: pre_read_offset = %x write_wrap = %x write = %x\n",
@@ -379,7 +385,8 @@ boolean_t updproc_preread(void)
 							 * This is done for performance reasons so that n_pre_read
 							 * doesn't have to be an atomic counter.
 							 */
-							csa = &FILE_INFO(gv_cur_region)->s_addrs;
+							reg = gv_cur_region;
+							csa = &FILE_INFO(reg)->s_addrs;
 							assert(!csa->now_crit);
 							status = gvcst_search(gv_currkey, NULL);
 							assert(!csa->now_crit);
@@ -400,10 +407,12 @@ boolean_t updproc_preread(void)
 								 * gvcst_search for this gv_target is called.
 								 */
 								gv_target->clue.end = 0;
+								assert(cdb_sc_reorg_encrypt != status);
 							}
+							assert(NULL == reorg_encrypt_restart_csa);
 							if (disk_blk_read)
 								csa->nl->n_pre_read--;
-#ifdef REPL_DEBUG
+#							ifdef REPL_DEBUG
 							if (NULL == (end = format_targ_key(buff,
 											   MAX_ZWR_KEY_SZ, gv_currkey, TRUE)))
 								end = &buff[MAX_ZWR_KEY_SZ - 1];
@@ -415,7 +424,7 @@ boolean_t updproc_preread(void)
 								" Seqno = 0x!16@XQ Rectype = !SL gv_currkey = !AD status = !SL",
 								TRUE, &recvpool.recvpool_ctl->jnl_seqno,
 								rectype, end - buff, buff, status);
-#endif
+#							endif
 						} else
 						{
 							REPL_DPRINT1("Unexpected bad record\n");
@@ -429,12 +438,12 @@ boolean_t updproc_preread(void)
 		}
 		if (!good_record)
 		{
-#ifdef REPL_DEBUG
+#			ifdef REPL_DEBUG
 			REPL_DPRINT5("Skipping record: pre_read_offset = %x read = %x write_wrap = %x write = %x\n",
 				pre_read_offset, upd_proc_local->read, write_wrap, write);
 			REPL_DPRINT3("New values: write_wrap = %x write = %x\n",
 					recvpool_ctl->write_wrap, recvpool_ctl->write);
-#endif
+#			endif
 			return TRUE;
 		}
 	} /* end while */
