@@ -1,6 +1,7 @@
 /****************************************************************
  *								*
- *	Copyright 2001, 2014 Fidelity Information Services, Inc	*
+ * Copyright (c) 2001-2015 Fidelity National Information 	*
+ * Services, Inc. and/or its subsidiaries. All rights reserved.	*
  *								*
  *	This source code contains the intellectual property	*
  *	of its copyright holder(s), and is made available	*
@@ -62,6 +63,7 @@
 #ifdef GTM_TRIGGER
 #include <rtnhdr.h>			/* for rtn_tabent in gv_trigger.h */
 #include "gv_trigger.h"
+# include "gtm_trigger.h"
 #include "targ_alloc.h"
 #include "trigger.h"
 #include "hashtab_str.h"
@@ -150,16 +152,16 @@ GBLREF	int			tprestart_state;        /* When triggers restart, multiple states p
 GBLREF	dollar_ecode_type	dollar_ecode;		/* structure containing $ECODE related information */
 GBLREF	mval			dollar_ztwormhole;
 GBLREF	boolean_t		dollar_ztrigger_invoked;
+GBLREF	stack_frame		*frame_pointer;	/* needed by TRIGGER_BASE_FRAME_UNWIND_IF_NOMANSLAND macro */
 #endif
 GBLREF	boolean_t		skip_dbtriggers;
 GBLREF	gv_namehead		*gv_target;
 GBLREF	boolean_t		gv_play_duplicate_kills;
 #ifdef UNIX
-GBLREF		int4			strm_index;
-STATICDEF	boolean_t		set_onln_rlbk_flg;
+GBLREF		int4		strm_index;
+STATICDEF	boolean_t	set_onln_rlbk_flg;
 #endif
 
-LITREF	mval			literal_hasht;
 static	boolean_t		updproc_continue = TRUE;
 
 error_def(ERR_GBLOFLOW);
@@ -170,6 +172,7 @@ error_def(ERR_REPEATERROR);
 error_def(ERR_REPLONLNRLBK);
 error_def(ERR_SECONDAHEAD);
 error_def(ERR_STRMSEQMISMTCH);
+error_def(ERR_SYSCALL);
 error_def(ERR_TEXT);
 error_def(ERR_TPRETRY);
 error_def(ERR_TRIGDEFNOSYNC);
@@ -265,9 +268,12 @@ error_def(ERR_UPDREPLSTATEOFF);
 
 CONDITION_HANDLER(updproc_ch)
 {
-	int		rc;
-	unsigned char	seq_num_str[32], *seq_num_ptr;
-	unsigned char	seq_num_strx[32], *seq_num_ptrx;
+	int			rc;
+	unsigned char		seq_num_str[32], *seq_num_ptr;
+	unsigned char		seq_num_strx[32], *seq_num_ptrx;
+#	ifdef GTM_TRIGGER
+	condition_handler	*save_active_ch;
+#	endif
 
 	START_CH(TRUE);
 	if ((int)ERR_TPRETRY == SIGNAL)
@@ -287,6 +293,24 @@ CONDITION_HANDLER(updproc_ch)
 		if (first_sgm_info GTMTRIG_ONLY( || (TPRESTART_STATE_NORMAL != tprestart_state)))
 		{
 			VMS_ONLY(assert(FALSE == tp_restart_fail_sig_used);)
+#			ifdef GTM_TRIGGER
+			/* Note: Below GTM_TRIGGER-only code is similar to mdb_condition_handler. But unlike there,
+			 * where MUM_TSTART is done inside the condition handler to transfer control, we do an UNWIND
+			 * here in updproc_ch. The UNWIND macro does an active_ch++ (to balance out the active_ch-- done
+			 * by START_CH done at the entry of every condition handler). We want to preserve active_ch in
+			 * the state that UNWIND expects. But the macro call done below calls gtm_trigger_fini which in
+			 * turn calls unw_mv_ent which could modify active_ch as part of popping MVST_TRIGR. Work around
+			 * that by saving and restoring active_ch after the macro call so UNWIND is fine. But what this
+			 * means is that for a short code window (from when unw_mv_ent sets active_ch to when we reset it
+			 * after the macro call below), active_ch would be incorrectly set to updproc_ch (instead of
+			 * util_base_ch which START_CH would have set it to above). But that is okay since we dont expect
+			 * any TPRETRY errors in this short window and updproc_ch would anyways transfer control to
+			 * util_base_ch (through NEXTCH) for all non-TPRETRY errors.
+			 */
+			save_active_ch = active_ch;
+			TRIGGER_BASE_FRAME_UNWIND_IF_NOMANSLAND;
+			active_ch = save_active_ch;
+#			endif
 			rc = tp_restart(1, TP_RESTART_HANDLES_ERRORS); /* any nested errors will set SIGNAL accordingly */
 			assert(0 == rc);			/* No partials restarts can happen at this final level */
 			GTMTRIG_ONLY(assert(TPRESTART_STATE_NORMAL == tprestart_state));
@@ -340,6 +364,10 @@ CONDITION_HANDLER(updproc_ch)
 	else if (ERR_REPLONLNRLBK == SIGNAL)
 	{
 		preemptive_db_clnup(SEVERITY);
+		/* We are about to abort the current in-progress transaction (TP or Non-TP). And restart from a potentially older
+		 * transaction. Release crit on any regions we hold (possible in some edge cases).
+		 */
+		have_crit(CRIT_HAVE_ANY_REG | CRIT_RELEASE); /* Note: "have_crit" function is used to release crit in this case */
 		assert(INVALID_GV_TARGET == reset_gv_target);
 		set_onln_rlbk_flg = TRUE;
 		/* Just like the UNWIND done above in the tprestart case, this is a case where an online rollback is signaled
@@ -352,6 +380,10 @@ CONDITION_HANDLER(updproc_ch)
 		UNWIND(NULL, NULL);
 	}
 #	endif
+	/* Assumes we don't fall through and unlock the mutex later
+	 * We don't care if it fails, so ignore the status.
+	 */
+	pthread_mutex_unlock(&recvpool.recvpool_ctl->write_updated_ctl);
 	NEXTCH;
 }
 
@@ -512,8 +544,17 @@ int updproc(void)
 			|| (jnlpool.repl_inst_filehdr->is_supplementary && !jnlpool.jnlpool_ctl->upd_disabled));
 #		endif
 		UNIX_ONLY(assert(updproc_continue && !set_onln_rlbk_flg));
+		/* We should be the only one acquiring the lock, so a single try should be sufficient. */
+		status = pthread_mutex_trylock(&recvpool.recvpool_ctl->write_updated_ctl);
+		if (0 != status)
+			rts_error_csa(CSA_ARG(NULL) VARLSTCNT(8) ERR_SYSCALL, 5,
+					LEN_AND_LIT("pthread_mutex_trylock"), CALLFROM, status, 0);
 		while (updproc_continue UNIX_ONLY(&& !set_onln_rlbk_flg))
 			updproc_actions(gld_db_files);
+		status = pthread_mutex_unlock(&recvpool.recvpool_ctl->write_updated_ctl);
+		if (0 != status)
+			rts_error_csa(CSA_ARG(NULL) VARLSTCNT(8) ERR_SYSCALL, 5,
+					LEN_AND_LIT("pthread_mutex_unlock"), CALLFROM, status, 0);
 #		ifdef UNIX
 		if (set_onln_rlbk_flg)
 		{	/* A concurrent online rollback happened which drove the updproc_ch and called us. Need to let the receiver
@@ -580,6 +621,8 @@ void updproc_actions(gld_dbname_list *gld_db_files)
 	repl_old_triple_jnl_t	*input_old_triple;
 	repl_histrec_jnl_ptr_t	input_histjrec;
 	uint4			expected_rec_len;
+	int			status;
+	struct timespec		waketime;
 #	endif
 	jnl_private_control	*jpc;
 	gld_dbname_list		*curr;
@@ -704,7 +747,17 @@ void updproc_actions(gld_dbname_list *gld_db_files)
 			 */
 			assert((0 == recvpool.recvpool_ctl->jnl_seqno) || (jnl_seqno <= recvpool.recvpool_ctl->jnl_seqno));
 				/* the 0 == check takes care of the startup case where jnl_seqno is 0 in the recvpool_ctl */
-			SHORT_SLEEP(10);
+			status = clock_gettime(CLOCK_REALTIME, &waketime);
+			if (0 != status)
+				rts_error_csa(CSA_ARG(NULL) VARLSTCNT(8) ERR_SYSCALL, 5,
+						LEN_AND_LIT("clock_gettime"), CALLFROM, errno, 0);
+			waketime.tv_sec += 10;
+			status = pthread_cond_timedwait(&recvpool.recvpool_ctl->write_updated,
+							&recvpool.recvpool_ctl->write_updated_ctl,
+							&waketime);
+			if ((0 != status) && (ETIMEDOUT != status))
+				rts_error_csa(CSA_ARG(NULL) VARLSTCNT(8) ERR_SYSCALL, 5,
+						LEN_AND_LIT("pthread_mutex_timedwait"), CALLFROM, status, 0);
 #			ifdef UNIX
 			if (!upd_proc_local->onln_rlbk_flg && (repl_csa->onln_rlbk_cycle != jnlpool.jnlpool_ctl->onln_rlbk_cycle))
 			{	/* A concurrent online rollback happened. Start afresh */
@@ -1382,7 +1435,7 @@ void updproc_actions(gld_dbname_list *gld_db_files)
 						assert(NULL != tmpcsa);
 						rts_error_csa(CSA_ARG(tmpcsa) VARLSTCNT(10) ERR_REC2BIG, 4,
 							VMS_ONLY(gv_currkey->end + 1 + SIZEOF(rec_hdr) +) val_mv.str.len,
-							(int4)gv_cur_region->max_rec_size, REG_LEN_STR(gv_cur_region),
+							(int4)(tmpcsa->region)->max_rec_size, REG_LEN_STR(tmpcsa->region),
 							ERR_GVIS, 2, end - buff, buff);
 						break;
 				}
