@@ -1,6 +1,7 @@
 /****************************************************************
  *								*
- *	Copyright 2001, 2014 Fidelity Information Services, Inc	*
+ * Copyright (c) 2001-2015 Fidelity National Information 	*
+ * Services, Inc. and/or its subsidiaries. All rights reserved.	*
  *								*
  *	This source code contains the intellectual property	*
  *	of its copyright holder(s), and is made available	*
@@ -12,14 +13,16 @@
 /* WARNING: this module contains a mixture of ASCII and EBCDIC on S390*/
 #include "mdef.h"
 
-#include "gtm_string.h"
-#include <errno.h>
-#include <wctype.h>
-#include <wchar.h>
-#include <signal.h>
+#include "gtm_signal.h"
 #include "gtm_unistd.h"
 #include "gtm_stdlib.h"
 #include "gtm_select.h"
+#include "gtm_string.h"
+
+#include <errno.h>
+#include <wctype.h>
+#include <wchar.h>
+
 #include "io.h"
 #include "trmdef.h"
 #include "iottdef.h"
@@ -55,6 +58,7 @@ GBLREF spdesc 		stringpool;
 GBLREF unsigned char	*msp, *stackbase, *stacktop, *stackwarn;
 GBLREF volatile int4	outofband;
 GBLREF	boolean_t	dmterm_default;
+GBLREF	volatile boolean_t	timer_in_handler;
 
 LITREF unsigned char	lower_to_upper_table[];
 #ifdef UNICODE_SUPPORTED
@@ -150,6 +154,7 @@ void	dm_read (mval *v)
 #	endif
 	char		*argv[3];
 	char		temp_str[MAX_RECALL_NUMBER_LENGTH + 1];
+	char		*strtokptr;
 	const char	delimiter_string[] = " \t";
 	d_tt_struct 	*tt_ptr;
 	enum RECALL_ERR_CODE	err_recall = NO_ERROR;
@@ -378,7 +383,7 @@ void	dm_read (mval *v)
 				if (!prin_in_dev_failure)
 					prin_in_dev_failure = TRUE;
 				else
-					exit(errno);
+					EXIT(errno);
 			}
 			tt_ptr->discard_lf = FALSE;
 			rts_error_csa(CSA_ARG(NULL) VARLSTCNT(1) ERR_IOEOF);
@@ -532,8 +537,8 @@ void	dm_read (mval *v)
 					if (((strlen(REC) == match_length) || (strlen(RECALL) == match_length))
 						&& (0 == strncmp((const char *)buffer_start, RECALL, match_length)))
 					{
-						strtok((char *)buffer_start, delimiter_string);
-						argv[1] = strtok(NULL, "");
+						STRTOK_R((char *)buffer_start, delimiter_string, &strtokptr);
+						argv[1] = STRTOK_R(NULL, "", &strtokptr);
 					} else
 						break;		/* not RECALL so end of line */
 #				ifdef UNICODE_SUPPORTED
