@@ -1,6 +1,7 @@
 /****************************************************************
  *								*
- *	Copyright 2001, 2014 Fidelity Information Services, Inc	*
+ * Copyright (c) 2001-2015 Fidelity National Information 	*
+ * Services, Inc. and/or its subsidiaries. All rights reserved.	*
  *								*
  *	This source code contains the intellectual property	*
  *	of its copyright holder(s), and is made available	*
@@ -24,16 +25,19 @@
 #include "gtm_ipc.h"
 #include <errno.h>
 #include "gtm_string.h"
+#include "cli.h"
 
 #define DEFAULT_ID	43
 #define ID_PREFIX	"-id="
 
+GBLDEF   CLI_ENTRY       *cmd_ary = NULL; /* "ftok" utility does not have any command tables so initialize command array to NULL */
+
 #define PrintUsage \
 	{ \
 		FPRINTF(stderr, "\nUsage:\n"); \
 		FPRINTF(stderr, "\t%s [%s<number>] <file1> <file2> ... <filen>\n\n", argv[0], ID_PREFIX); \
 		FPRINTF(stderr, "Reports IPC Key(s) (using id 1, or <number>) of <file1> <file2> ... <filen>\n\n"); \
-		exit(EXIT_FAILURE); \
+		EXIT(EXIT_FAILURE); \
 	}
 
 int main (int argc, char *argv[])
