@@ -1,6 +1,7 @@
 /****************************************************************
  *								*
- *	Copyright 2004, 2014 Fidelity Information Services, Inc	*
+ * Copyright (c) 2004-2016 Fidelity National Information	*
+ * Services, Inc. and/or its subsidiaries. All rights reserved.	*
  *								*
  *	This source code contains the intellectual property	*
  *	of its copyright holder(s), and is made available	*
@@ -17,6 +18,7 @@
 #include <errno.h>
 #include <sys/sem.h>
 #include <sys/types.h>
+#include <locale.h>
 #include "gtm_stat.h"
 #include "gtm_string.h"
 #include "gtm_strings.h"
@@ -51,6 +53,7 @@
 #include "gtmlink.h"
 #include "send_msg.h"
 #include "eintr_wrappers.h"
+#include "utfcgr.h"
 #ifdef __linux__
 #include "hugetlbfs_overrides.h"
 #endif
@@ -66,6 +69,12 @@
 # define GTM_USESECSHR				"$gtm_usesecshr"
 /* GTM_TEST_FAKE_ENOSPC is used only in debug code so it does not have to go in gtm_logicals.h */
 # define GTM_TEST_FAKE_ENOSPC			"$gtm_test_fake_enospc"
+/* GTM_TEST_AUTORELINK_ALWAYS is used only in debug code so it does not have to go in gtm_logicals.h */
+# define GTM_TEST_AUTORELINK_ALWAYS		"$gtm_test_autorelink_always"
+/* GTM_DB_COUNTER_SEM_INCR is used only in debug code so it does not have to go in gtm_logicals.h */
+# define GTM_DB_COUNTER_SEM_INCR		"$gtm_db_counter_sem_incr"
+/* GTM_TEST_JNLPOOL_SYNC is used only in debug code so it does not have to go in gtm_logicals.h */
+# define GTM_TEST_JNLPOOL_SYNC			"$gtm_test_jnlpool_sync"
 #endif
 #define DEFAULT_MUPIP_TRIGGER_ETRAP 		"IF $ZJOBEXAM()"
 
@@ -77,7 +86,13 @@
 #endif
 #define MAX_TRANS_NAME_LEN			GTM_PATH_MAX
 
-GBLREF	int4			gtm_shmflags;			/* Shared memory flags for shmat() */
+/* Remove trailing '/' from path (unless only '/') */
+#define	REMOVE_TRAILING_SLASH_FROM_MSTR(TRANS)				\
+{									\
+	while ((1 < TRANS.len) && ('/' == TRANS.addr[TRANS.len - 1]))	\
+		TRANS.len--;						\
+}
+
 GBLREF	uint4			gtm_principal_editing_defaults;	/* ext_cap flags if tt */
 GBLREF	boolean_t		is_gtm_chset_utf8;
 GBLREF	boolean_t		utf8_patnumeric;
@@ -93,6 +108,7 @@ GBLREF	boolean_t		ipv4_only;		/* If TRUE, only use AF_INET. */
 ZOS_ONLY(GBLREF	char		*gtm_utf8_locale_object;)
 ZOS_ONLY(GBLREF	boolean_t	gtm_tag_utf8_as_ascii;)
 GTMTRIG_ONLY(GBLREF	mval	gtm_trigger_etrap;)
+GBLREF	volatile boolean_t	timer_in_handler;
 
 #ifdef GTM_TRIGGER
 LITDEF mval default_mupip_trigger_etrap = DEFINE_MVAL_LITERAL(MV_STR, 0 , 0 , (SIZEOF(DEFAULT_MUPIP_TRIGGER_ETRAP) - 1),
@@ -117,17 +133,22 @@ static readonly unsigned char editing_index[27] =
 };
 static readonly unsigned char init_break[1] = {'B'};
 
+error_def(ERR_INVLOCALE);
 error_def(ERR_INVLINKTMPDIR);
+error_def(ERR_INVTMPDIR);
+error_def(ERR_ARCTLMAXHIGH);
+error_def(ERR_ARCTLMAXLOW);
 
 void	gtm_env_init_sp(void)
 {	/* Unix only environment initializations */
 	mstr		val, trans;
 	int4		status, index, len, hrtbt_cntr_delta, stat_res;
 	size_t		cwdlen;
-	boolean_t	ret, is_defined;
+	boolean_t	ret, is_defined, novalidate;
 	char		buf[MAX_TRANS_NAME_LEN], *token, cwd[GTM_PATH_MAX];
-	char		*cwdptr, *trigger_etrap, *c, *end;
+	char		*cwdptr, *trigger_etrap, *c, *end, *strtokptr;
 	struct stat	outbuf;
+	int		gtm_autorelink_shm_min;
 	DCL_THREADGBL_ACCESS;
 
 	SETUP_THREADGBL_ACCESS;
@@ -135,8 +156,9 @@ void	gtm_env_init_sp(void)
 	libhugetlbfs_init();
 #	endif
 #	ifdef __MVS__
-	/* For now OS/390 only. Eventually, this will be added to all UNIX platforms along with the
-	 * capability to specify the desired directory to put a core file in.
+	/* For now OS/390 only. Eventually, this could be added to all UNIX platforms along with the
+	 * capability to specify the desired directory to put a core file in. Needs to be setup before
+	 * much of anything else is.
 	 */
 	if (NULL == gtm_core_file)
 	{
@@ -156,9 +178,25 @@ void	gtm_env_init_sp(void)
 		} /* else gtm_core_file/gtm_core_putenv remain null and we likely cannot generate proper core files */
 	}
 #	endif
-	val.addr = GTM_SHMFLAGS;
-	val.len = SIZEOF(GTM_SHMFLAGS) - 1;
-	gtm_shmflags = (int4)trans_numeric(&val, &is_defined, TRUE);	/* Flags vlaue (0 is undefined or bad) */
+	/* Validate $gtm_tmp if specified, else that default is available */
+	val.addr = GTM_TMP_ENV;
+	val.len = SIZEOF(GTM_TMP_ENV) - 1;
+	if (SS_NORMAL != (status = TRANS_LOG_NAME(&val, &trans, buf, SIZEOF(buf), do_sendmsg_on_log2long)))
+	{	/* Nothing for $gtm_tmp either - use DEFAULT_GTM_TMP which is already a string */
+		MEMCPY_LIT(buf, DEFAULT_GTM_TMP);
+		trans.addr = buf;
+		trans.len = SIZEOF(DEFAULT_GTM_TMP) - 1;
+	}
+	assert(GTM_PATH_MAX > trans.len);
+	REMOVE_TRAILING_SLASH_FROM_MSTR(trans); /* Remove trailing '/' from trans.addr */
+	trans.addr[trans.len] = '\0';
+	STAT_FILE(trans.addr, &outbuf, stat_res);
+	if ((-1 == stat_res) || !S_ISDIR(outbuf.st_mode))
+	{
+		/* Either the directory doesn't exist or the specified or defaulted entity is not a directory */
+		send_msg_csa(CSA_ARG(NULL) VARLSTCNT(4)	ERR_INVTMPDIR, 2, trans.len, trans.addr);
+	}
+	/* Check for and and setup gtm_quiet_halt if specified */
 	val.addr = GTM_QUIET_HALT;
 	val.len = SIZEOF(GTM_QUIET_HALT) - 1;
 	ret = logical_truth_value(&val, FALSE, &is_defined);
@@ -196,7 +234,7 @@ void	gtm_env_init_sp(void)
 	{
 		assert(trans.len < SIZEOF(buf));
 		trans.addr[trans.len] = '\0';
-		token = strtok(trans.addr, ":");
+		token = STRTOK_R(trans.addr, ":", &strtokptr);
 		while (NULL != token)
 		{
 			if (ISALPHA_ASCII(token[0]))
@@ -207,33 +245,33 @@ void	gtm_env_init_sp(void)
 			{
 				switch (index)
 				{
-				case 0:	/* EDITING */
-					gtm_principal_editing_defaults |= TT_EDITING;
-					break;
-				case 1:	/* EMPTERM */
-					gtm_principal_editing_defaults |= TT_EMPTERM;
-					break;
-				case 2:	/* INSERT */
-					gtm_principal_editing_defaults &= ~TT_NOINSERT;
-					break;
-				case 3:	/* NOEDITING */
-					gtm_principal_editing_defaults &= ~TT_EDITING;
-					break;
-				case 4:	/* NOEMPTERM */
-					gtm_principal_editing_defaults &= ~TT_EMPTERM;
-					break;
-				case 5:	/* NOINSERT */
-					gtm_principal_editing_defaults |= TT_NOINSERT;
-					break;
+					case 0:	/* EDITING */
+						gtm_principal_editing_defaults |= TT_EDITING;
+						break;
+					case 1:	/* EMPTERM */
+						gtm_principal_editing_defaults |= TT_EMPTERM;
+						break;
+					case 2:	/* INSERT */
+						gtm_principal_editing_defaults &= ~TT_NOINSERT;
+						break;
+					case 3:	/* NOEDITING */
+						gtm_principal_editing_defaults &= ~TT_EDITING;
+						break;
+					case 4:	/* NOEMPTERM */
+						gtm_principal_editing_defaults &= ~TT_EMPTERM;
+						break;
+					case 5:	/* NOINSERT */
+						gtm_principal_editing_defaults |= TT_NOINSERT;
+						break;
 				}
 			}
-			token = strtok(NULL, ":");
+			token = STRTOK_R(NULL, ":", &strtokptr);
 		}
 	}
 	val.addr = GTM_CHSET_ENV;
 	val.len = STR_LIT_LEN(GTM_CHSET_ENV);
 	if (SS_NORMAL == (status = TRANS_LOG_NAME(&val, &trans, buf, SIZEOF(buf), do_sendmsg_on_log2long))
-		&& STR_LIT_LEN(UTF8_NAME) == trans.len)
+	    && STR_LIT_LEN(UTF8_NAME) == trans.len)
 	{
 		if (!strncasecmp(buf, UTF8_NAME, STR_LIT_LEN(UTF8_NAME)))
 		{
@@ -242,7 +280,7 @@ void	gtm_env_init_sp(void)
 			val.addr = GTM_CHSET_LOCALE_ENV;
 			val.len = STR_LIT_LEN(GTM_CHSET_LOCALE_ENV);
 			if ((SS_NORMAL == (status = TRANS_LOG_NAME(&val, &trans, buf, SIZEOF(buf), do_sendmsg_on_log2long))) &&
-				(0 < trans.len))
+			    (0 < trans.len))
 			{	/* full path to 64 bit ASCII UTF-8 locale object */
 				gtm_utf8_locale_object = malloc(trans.len + 1);
 				strcpy(gtm_utf8_locale_object, buf);
@@ -259,8 +297,8 @@ void	gtm_env_init_sp(void)
 			val.addr = GTM_PATNUMERIC_ENV;
 			val.len = STR_LIT_LEN(GTM_PATNUMERIC_ENV);
 			if (SS_NORMAL == (status = TRANS_LOG_NAME(&val, &trans, buf, SIZEOF(buf), do_sendmsg_on_log2long))
-				&& STR_LIT_LEN(UTF8_NAME) == trans.len
-				&& !strncasecmp(buf, UTF8_NAME, STR_LIT_LEN(UTF8_NAME)))
+			    && STR_LIT_LEN(UTF8_NAME) == trans.len
+			    && !strncasecmp(buf, UTF8_NAME, STR_LIT_LEN(UTF8_NAME)))
 			{
 				utf8_patnumeric = TRUE;
 			}
@@ -345,38 +383,83 @@ void	gtm_env_init_sp(void)
 	{
 		init_relink_allowed(&trans); /* set TREF(relink_allowed) */
 	}
-#	ifdef USHBIN_SUPPORTED
-	/* Set default or supplied value for $gtm_linktmpdir */
-	val.addr = GTM_LINKTMPDIR;
-	val.len = SIZEOF(GTM_LINKTMPDIR) - 1;
-	if (SS_NORMAL != (status = TRANS_LOG_NAME(&val, &trans, buf, SIZEOF(buf), do_sendmsg_on_log2long)))
-	{	/* Else use default $gtm_tmp value or its default */
-		val.addr = GTM_TMP_ENV;
-		val.len = SIZEOF(GTM_TMP_ENV) - 1;
+#	ifdef AUTORELINK_SUPPORTED
+	if (!IS_GTMSECSHR_IMAGE)
+	{	/* Set default or supplied value for $gtm_linktmpdir */
+		val.addr = GTM_LINKTMPDIR;
+		val.len = SIZEOF(GTM_LINKTMPDIR) - 1;
 		if (SS_NORMAL != (status = TRANS_LOG_NAME(&val, &trans, buf, SIZEOF(buf), do_sendmsg_on_log2long)))
-		{	/* Nothing for $gtm_tmp either - use DEFAULT_GTM_TMP which is already a string */
-			trans.addr = DEFAULT_GTM_TMP;
-			trans.len = SIZEOF(DEFAULT_GTM_TMP) - 1;
+		{	/* Else use default $gtm_tmp value or its default */
+			val.addr = GTM_TMP_ENV;
+			val.len = SIZEOF(GTM_TMP_ENV) - 1;
+			if (SS_NORMAL != (status = TRANS_LOG_NAME(&val, &trans, buf, SIZEOF(buf), do_sendmsg_on_log2long)))
+			{	/* Nothing for $gtm_tmp either - use DEFAULT_GTM_TMP which is already a string */
+				trans.addr = DEFAULT_GTM_TMP;
+				trans.len = SIZEOF(DEFAULT_GTM_TMP) - 1;
+			}
+			novalidate = TRUE;		/* Don't validate gtm_linktmpdir if is defaulting to $gtm_tmp */
+		} else
+			novalidate = FALSE;
+		assert(GTM_PATH_MAX > trans.len);
+		REMOVE_TRAILING_SLASH_FROM_MSTR(trans); /* Remove trailing '/' from trans.addr */
+		(TREF(gtm_linktmpdir)).addr = malloc(trans.len + 1); /* +1 for '\0'; This memory is never freed */
+		(TREF(gtm_linktmpdir)).len = trans.len;
+		/* For now, we assume that if the environment variable is defined to NULL, anticipatory freeze is NOT in effect */
+		if (0 < trans.len)
+			memcpy((TREF(gtm_linktmpdir)).addr, trans.addr, trans.len);
+		((TREF(gtm_linktmpdir)).addr)[trans.len] = '\0';
+		if (!novalidate)
+		{	/* $gtm_linktmpdir was specified - validate it (bypass if using $gtm_tmp or its default as that was
+			 * already checked earlier.
+			 */
+			STAT_FILE((TREF(gtm_linktmpdir)).addr, &outbuf, stat_res);
+			if ((-1 == stat_res) || !S_ISDIR(outbuf.st_mode))
+			{	/* Either the directory doesn't exist or the entity is not a directory */
+				send_msg_csa(CSA_ARG(NULL) VARLSTCNT(4)
+					     ERR_INVLINKTMPDIR, 2, (TREF(gtm_linktmpdir)).len, (TREF(gtm_linktmpdir)).addr);
+				free((TREF(gtm_linktmpdir)).addr);
+				trans.len = SIZEOF(DEFAULT_GTM_TMP) - 1;
+				trans.addr = DEFAULT_GTM_TMP;
+				REMOVE_TRAILING_SLASH_FROM_MSTR(trans); /* Remove trailing '/' from trans.addr */
+				(TREF(gtm_linktmpdir)) = trans;
+			}
 		}
 	}
-	assert(GTM_PATH_MAX > trans.len);
-	(TREF(gtm_linktmpdir)).addr = malloc(trans.len + 1); /* +1 for '\0'; This memory is never freed */
-	(TREF(gtm_linktmpdir)).len = trans.len;
-	/* For now, we assume that if the environment variable is defined to NULL, anticipatory freeze is NOT in effect */
-	if (0 < trans.len)
+	/* See if gtm_autorelink_shm is set */
+	val.addr = GTM_AUTORELINK_SHM;
+	val.len = SIZEOF(GTM_AUTORELINK_SHM) - 1;
+	gtm_autorelink_shm_min = trans_numeric(&val, &is_defined, TRUE);
+	if (!is_defined || !gtm_autorelink_shm_min)
+		TREF(relinkctl_shm_min_index) = 0;
+	else
 	{
-		memcpy((TREF(gtm_linktmpdir)).addr, trans.addr, trans.len);
-		((TREF(gtm_linktmpdir)).addr)[trans.len] = '\0';
+		gtm_autorelink_shm_min = (2 <= gtm_autorelink_shm_min) ? ceil_log2_32bit(gtm_autorelink_shm_min) : 0;
+		TREF(relinkctl_shm_min_index) = gtm_autorelink_shm_min;
 	}
-	STAT_FILE((TREF(gtm_linktmpdir)).addr, &outbuf, stat_res);
-	if ((-1 == stat_res) || !S_ISDIR(outbuf.st_mode))
-	{	/* Either the directory doesn't exist or the entity is not a directory */
-		send_msg_csa(CSA_ARG(NULL) VARLSTCNT(4) ERR_INVLINKTMPDIR, 2, (TREF(gtm_linktmpdir)).len,
-			     (TREF(gtm_linktmpdir)).addr);
-		(TREF(gtm_linktmpdir)).len = SIZEOF(DEFAULT_GTM_TMP) - 1;
-		(TREF(gtm_linktmpdir)).addr = DEFAULT_GTM_TMP;
+	/* See if gtm_autorelink_keeprtn is set */
+	val.addr = GTM_AUTORELINK_KEEPRTN;
+	val.len = SIZEOF(GTM_AUTORELINK_KEEPRTN) - 1;
+	TREF(gtm_autorelink_keeprtn) = logical_truth_value(&val, FALSE, &is_defined);
+	if (!is_defined)
+		TREF(gtm_autorelink_keeprtn) = FALSE;
+	/* See if gtm_autorelink_ctlmax is set */
+	val.addr = GTM_AUTORELINK_CTLMAX;
+	val.len = SIZEOF(GTM_AUTORELINK_CTLMAX) - 1;
+	TREF(gtm_autorelink_ctlmax) = trans_numeric(&val, &is_defined, TRUE);
+	if (!is_defined)
+		TREF(gtm_autorelink_ctlmax) = RELINKCTL_DEFAULT_ENTRIES;
+	else if (TREF(gtm_autorelink_ctlmax) > RELINKCTL_MAX_ENTRIES)
+	{
+		send_msg_csa(CSA_ARG(NULL) VARLSTCNT(6) ERR_ARCTLMAXHIGH, 4, LEN_AND_LIT(GTM_AUTORELINK_CTLMAX),
+			     TREF(gtm_autorelink_ctlmax), RELINKCTL_MAX_ENTRIES);
+		TREF(gtm_autorelink_ctlmax) = RELINKCTL_MAX_ENTRIES;
+	} else if (TREF(gtm_autorelink_ctlmax) < RELINKCTL_MIN_ENTRIES)
+	{
+		send_msg_csa(CSA_ARG(NULL) VARLSTCNT(6) ERR_ARCTLMAXLOW, 4, LEN_AND_LIT(GTM_AUTORELINK_CTLMAX),
+			     TREF(gtm_autorelink_ctlmax), RELINKCTL_MIN_ENTRIES);
+		TREF(gtm_autorelink_ctlmax) = RELINKCTL_MIN_ENTRIES;
 	}
-#	endif
+#	endif /* AUTORELINK_SUPPORTED */
 #	ifdef DEBUG
 	/* DEBUG-only option to bypass 'easy' methods of things and always use gtmsecshr for IPC cleanups, wakeups, file removal,
 	 * etc. Basically use gtmsecshr for anything where it is an option - helps with testing gtmsecshr for proper operation.
@@ -392,6 +475,24 @@ void	gtm_env_init_sp(void)
 	TREF(gtm_test_fake_enospc) = logical_truth_value(&val, FALSE, &is_defined);
 	if (!is_defined)
 		TREF(gtm_test_fake_enospc) = FALSE;
+	/* DEBUG-only option to enable autorelink on all directories in $zroutines (except for shlib directories) */
+	val.addr = GTM_TEST_AUTORELINK_ALWAYS;
+	val.len = SIZEOF(GTM_TEST_AUTORELINK_ALWAYS) - 1;
+	TREF(gtm_test_autorelink_always) = logical_truth_value(&val, FALSE, &is_defined);
+	if (!is_defined)
+		TREF(gtm_test_autorelink_always) = FALSE;
+	/* DEBUG-only option to enable counter semaphore to be incremented by more than the default value of 1 */
+	val.addr = GTM_DB_COUNTER_SEM_INCR;
+	val.len = SIZEOF(GTM_DB_COUNTER_SEM_INCR) - 1;
+	gtm_db_counter_sem_incr = trans_numeric(&val, &is_defined, TRUE);
+	if (!is_defined)
+		gtm_db_counter_sem_incr = DEFAULT_DB_COUNTER_SEM_INCR;
+	/* DEBUG-only option to force the journal pool accounting out of sync every n transactions. */
+	val.addr = GTM_TEST_JNLPOOL_SYNC;
+	val.len = SIZEOF(GTM_TEST_JNLPOOL_SYNC) - 1;
+	TREF(gtm_test_jnlpool_sync) = trans_numeric(&val, &is_defined, TRUE);
+	if (!is_defined)
+		TREF(gtm_test_jnlpool_sync) = 0;
 #	endif
 #	ifdef GTMDBGFLAGS_ENABLED
 	val.addr = GTMDBGFLAGS;
@@ -409,4 +510,38 @@ void	gtm_env_init_sp(void)
 	val.addr = GTM_DMTERM;
 	val.len = SIZEOF(GTM_DMTERM) - 1;
 	dmterm_default = logical_truth_value(&val, FALSE, NULL);
+	/* Set values for gtm_utfcgr_strings and gtm_utfcgr_string_groups */
+	val.addr = GTM_UTFCGR_STRINGS;
+	val.len = SIZEOF(GTM_UTFCGR_STRINGS) - 1;
+	TREF(gtm_utfcgr_strings) = trans_numeric(&val, &is_defined, TRUE);
+	if (!is_defined)
+	{
+		assert(GTM_UTFCGR_STRINGS_DEFAULT <= GTM_UTFCGR_STRINGS_MAX);
+		TREF(gtm_utfcgr_strings) = GTM_UTFCGR_STRINGS_DEFAULT;
+	} else if (GTM_UTFCGR_STRINGS_MAX < TREF(gtm_utfcgr_strings))
+		TREF(gtm_utfcgr_strings) = GTM_UTFCGR_STRINGS_MAX;
+	val.addr = GTM_UTFCGR_STRING_GROUPS;
+	val.len = SIZEOF(GTM_UTFCGR_STRING_GROUPS) - 1;
+	TREF(gtm_utfcgr_string_groups) = trans_numeric(&val, &is_defined, TRUE);
+	if (!is_defined)
+		TREF(gtm_utfcgr_string_groups) = GTM_UTFCGR_STRING_GROUPS_DEFAULT;
+	/* If gtm_locale is defined, reset the locale for this process - but only for UTF8 mode */
+	if (is_gtm_chset_utf8)
+	{
+		val.addr = GTM_LOCALE;
+		val.len = SIZEOF(GTM_LOCALE) - 1;
+		if (SS_NORMAL == (status = TRANS_LOG_NAME(&val, &trans, buf, SIZEOF(buf), do_sendmsg_on_log2long)))
+		{
+			if ((0 < trans.len) && (SIZEOF(buf) > trans.len))
+			{	/* Something was specified - need to clear LC_ALL and set LC_CTYPE but need room in buf[]
+				 * for string-ending null.
+				 */
+				putenv("LC_ALL");			/* Clear LC_ALL before LC_CTYPE can take effect */
+				buf[trans.len] = '\0';
+				status = setenv("LC_CTYPE", buf, TRUE);
+				if (0 != status)
+					send_msg_csa(CSA_ARG(NULL) VARLSTCNT(5) ERR_INVLOCALE, 2, val.len, val.addr, status);
+			}
+		}
+	}
 }
