@@ -1,6 +1,7 @@
 /****************************************************************
  *								*
- *	Copyright 2001, 2014 Fidelity Information Services, Inc	*
+ * Copyright (c) 2001-2015 Fidelity National Information 	*
+ * Services, Inc. and/or its subsidiaries. All rights reserved.	*
  *								*
  *	This source code contains the intellectual property	*
  *	of its copyright holder(s), and is made available	*
@@ -11,23 +12,27 @@
 
 /* gtm_logicals.h - Environment variables used by GT.M. */
 /* within each group, the entries are in alpha order of the third column */
+/* Warning: A consideration is that reordering these values reorders the output in v53003/D9I10002703 as well */
 /* -------------------------- Common to UNIX and VMS --------------------------  */
 
 #define	GTM_DIST_LOG			"$gtm_dist"
 
 /* Database */
-#define	GTM_GBLDIR			"$gtmgbldir"
 #define	GTM_BLKUPGRADE_FLAG		"$gtm_blkupgrade_flag"
 #define	GTM_DBFILEXT_SYSLOG_DISABLE	"$gtm_dbfilext_syslog_disable"
 #define	GTM_ENV_XLATE			"$gtm_env_translate"
 #define	GTM_FULLBLOCKWRITES		"$gtm_fullblockwrites"
 #define	GTM_GDSCERT			"$gtm_gdscert"
 #define	GTM_GVDUPSETNOOP		"$gtm_gvdupsetnoop"
+#define	GTM_GBLDIR			"$gtmgbldir"
 #define	GTM_GVUNDEF_FATAL		"$gtm_gvundef_fatal"
+#define	GTM_POOLLIMIT			"$gtm_poollimit"
 #define	GTM_TP_ALLOCATION_CLUE		"$gtm_tp_allocation_clue"
 #define	GTM_TPNOTACIDTIME		"$gtm_tpnotacidtime"
 #define	GTM_TPRESTART_LOG_DELTA		"$gtm_tprestart_log_delta"
-#define	GTM_TPRESTART_LOG_LIMIT		"$gtm_tprestart_log_first"
+#define	GTM_TPRESTART_LOG_FIRST		"$gtm_tprestart_log_first"
+#define GTM_NONTPRESTART_LOG_DELTA	"$gtm_nontprestart_log_delta"
+#define	GTM_NONTPRESTART_LOG_FIRST	"$gtm_nontprestart_log_first"
 #define	GTM_ZMAXTPTIME			"$gtm_zmaxtptime"
 /* 	GTM_DIRTREE_COLLHDR_ALWAYS	"$gtm_dirtree_collhdr_always"	dbg-only use in gvcst_put hence no #define for it or
 									else the D9I10002703 subtest will need changes for this.
@@ -85,7 +90,6 @@
 
 /* Database */
 #define	GTM_TMP_ENV			"$gtm_tmp"
-#define	GTM_SHMFLAGS			"$gtm_shmatflags"
 #define	GTM_TRIGGER_ETRAP		"$gtm_trigger_etrap"
 #define	GTM_SNAPTMPDIR			"$gtm_snaptmpdir"
 #define	GTM_DB_STARTUP_MAX_WAIT		"$gtm_db_startup_max_wait"
@@ -113,6 +117,9 @@
 /* [Auto]Relink related */
 #define	GTM_LINK			"$gtm_link"
 #define GTM_LINKTMPDIR			"$gtm_linktmpdir"
+#define	GTM_AUTORELINK_SHM 		"$gtm_autorelink_shm"
+#define	GTM_AUTORELINK_KEEPRTN		"$gtm_autorelink_keeprtn"	/* do not let go of objects in rtnobj shm */
+#define	GTM_AUTORELINK_CTLMAX		"$gtm_autorelink_ctlmax"
 
 /* Miscellaneous */
 #define GTM_ERROR_ON_JNL_FILE_LOST	"$gtm_error_on_jnl_file_lost"
@@ -130,3 +137,7 @@
 #define GTM_MAX_STORALLOC		"$gtm_max_storalloc"
 #define GTM_IPV4_ONLY			"$gtm_ipv4_only"
 #define GTM_DMTERM			"$gtm_dmterm"
+#define	GTM_MUPJNL_PARALLEL		"$gtm_mupjnl_parallel"
+#define GTM_LOCALE			"$gtm_locale"
+#define GTM_UTFCGR_STRINGS		"$gtm_utfcgr_strings"
+#define GTM_UTFCGR_STRING_GROUPS	"$gtm_utfcgr_string_groups"
