@@ -1,6 +1,7 @@
 /****************************************************************
  *								*
- *	Copyright 2001, 2014 Fidelity Information Services, Inc	*
+ * Copyright (c) 2001-2016 Fidelity National Information	*
+ * Services, Inc. and/or its subsidiaries. All rights reserved.	*
  *								*
  *	This source code contains the intellectual property	*
  *	of its copyright holder(s), and is made available	*
@@ -54,6 +55,7 @@
 #include "repl_comm.h"
 #include "repl_instance.h"
 #include "ftok_sems.h"
+#include "ftok_sem_incrcnt.h"
 #include "gt_timer.h"		/* for LONG_SLEEP macro (hiber_start function prototype) */
 #include "init_secshr_addrs.h"
 #include "mutex.h"
@@ -114,13 +116,14 @@ int gtmsource()
 	char			print_msg[1024], tmpmsg[1024];
 	gd_region		*reg, *region_top;
 	sgmnt_addrs		*csa, *repl_csa;
-	boolean_t		all_files_open, isalive;
+	boolean_t		all_files_open, isalive, ftok_counter_halted;
 	pid_t			pid, ppid, procgp;
 	seq_num			read_jnl_seqno, jnl_seqno;
 	unix_db_info		*udi;
 	gtmsource_local_ptr_t	gtmsource_local;
 	boolean_t		this_side_std_null_coll;
 	int			null_fd, rc;
+	uint4			shutdowntime = 0;
 
 	memset((uchar_ptr_t)&jnlpool, 0, SIZEOF(jnlpool_addrs));
 	call_on_signal = gtmsource_sigstop;
@@ -128,23 +131,64 @@ int gtmsource()
 	if (-1 == gtmsource_get_opt())
 		rts_error_csa(CSA_ARG(NULL) VARLSTCNT(1) ERR_MUPCLIERR);
 	if (gtmsource_options.shut_down)
-	{	/* Wait till shutdown time nears even before going to "jnlpool_init". This is because the latter will return
-		 * with the ftok semaphore and access semaphore held and we do not want to be holding those locks (while
-		 * waiting for the user specified timeout to expire) as that will affect new GTM processes and/or other
-		 * MUPIP REPLIC commands that need these locks for their function.
-		 */
-		if (0 < gtmsource_options.shutdown_time)
+	{
+		if (gtmsource_options.zerobacklog)
 		{
-			repl_log(stdout, TRUE, TRUE, "Waiting for %d seconds before signalling shutdown\n",
-												gtmsource_options.shutdown_time);
+			gtmsource_options.shut_down = FALSE; /* for getbacklog need init but can't hold ftok for time */
+			jnlpool_init(GTMSOURCE, gtmsource_options.start, &is_jnlpool_creator);
+			gtmsource_options.shut_down = TRUE; /* restore actual value after the jnlpool_init */
+			if (0 < gtmsource_options.shutdown_time)
+			{
+				for (shutdowntime = gtmsource_options.shutdown_time; shutdowntime; shutdowntime--)
+				{
+					if (0 == gtmsource_checkforbacklog())
+					{
+						repl_log(stdout, TRUE, TRUE, "Shutting down with zero backlog\n");
+						break;
+					}
+					LONG_SLEEP(1);
+				}
+				if (0 == shutdowntime)
+				{
+					repl_log(stdout, TRUE, TRUE, "Shutting down with a backlog due to timeout\n");
+				}
+			}
+			jnlpool_detach(); /* reattach below so as to hold the semaphores */
+			if (0 != (save_errno = rel_sem(SOURCE, JNL_POOL_ACCESS_SEM)))
+				rts_error_csa(CSA_ARG(NULL) VARLSTCNT(7) ERR_JNLPOOLSETUP, 0,
+					ERR_TEXT, 2,
+					RTS_ERROR_LITERAL("Error from rel_sem in Source Server shutdown attempt"), save_errno);
+		} else if (0 < gtmsource_options.shutdown_time)
+		{
+			repl_log(stdout, TRUE, TRUE, "Waiting for %d second(s) before forcing shutdown\n",
+					gtmsource_options.shutdown_time);
 			LONG_SLEEP(gtmsource_options.shutdown_time);
 		} else
-			repl_log(stdout, TRUE, TRUE, "Signalling shutdown immediate\n");
+			repl_log(stdout, TRUE, TRUE, "Forcing immediate shutdown\n");
 	} else if (gtmsource_options.start)
 	{
 		repl_log(stdout, TRUE, TRUE, "Initiating START of source server for secondary instance [%s]\n",
 			gtmsource_options.secondary_instname);
 	}
+	if (gtmsource_options.activate && (ROOTPRIMARY_SPECIFIED == gtmsource_options.rootprimary))
+	{	/* MUPIP REPLIC -SOURCE -ACTIVATE -UPDOK has been specified. We need to open the gld and db regions now
+		 * in case this is a secondary -> primary transition. This is so we can later switch journal files in all
+		 * journaled regions when the transition actually happens inside "gtmsource_rootprimary_init". But since
+		 * we have not yet done a "jnlpool_init", we dont know if updates are disabled in it or not. Although we
+		 * need to do the gld/db open only if updates are currently disabled in the jnlpool, we do this always
+		 * because once we do a jnlpool_init, we will come back with the ftok on the jnlpool held and that has
+		 * issues with later db open since we will try to hold the db ftok as part of db open and the ftok logic
+		 * currently has assumptions that a process holds only one ftok at any point in time.
+		 */
+		assert(NULL == gd_header);
+		gvinit();
+		all_files_open = region_init(FALSE);
+		if (!all_files_open)
+		{
+			gtm_putmsg_csa(CSA_ARG(NULL) VARLSTCNT(1) ERR_NOTALLDBOPN);
+			gtmsource_exit(ABNORMAL_SHUTDOWN);
+		}
+	}
 	jnlpool_init(GTMSOURCE, gtmsource_options.start, &is_jnlpool_creator);
 	/* is_jnlpool_creator == TRUE ==> this process created the journal pool
 	 * is_jnlpool_creator == FALSE ==> journal pool already existed and this process simply attached to it.
@@ -183,6 +227,7 @@ int gtmsource()
 	/* Set "child_server_running" to FALSE before forking off child. Wait for it to be set to TRUE by the child. */
 	gtmsource_local = jnlpool.gtmsource_local;
 	gtmsource_local->child_server_running = FALSE;
+	udi = FILE_INFO(jnlpool.jnlpool_dummy_reg);
 	FORK(pid);
 	if (0 > pid)
 	{
@@ -208,7 +253,12 @@ int gtmsource()
 			if (0 != (save_errno = rel_sem(SOURCE, JNL_POOL_ACCESS_SEM)))
 				rts_error_csa(CSA_ARG(NULL) VARLSTCNT(7) ERR_JNLPOOLSETUP, 0,
 					ERR_TEXT, 2, RTS_ERROR_LITERAL("Error in rel_sem"), save_errno);
-			ftok_sem_release(jnlpool.jnlpool_dummy_reg, TRUE, TRUE);
+			/* If the child source server process got a ftok counter overflow, it would have recorded that in
+			 * jnlpool.repl_inst_filehdr->ftok_counter_halted. Decrement the ftok counter only if neither we nor the
+			 * child process got a counter overflow.
+			 */
+			ftok_sem_release(jnlpool.jnlpool_dummy_reg, udi->counter_ftok_incremented
+									&& !jnlpool.repl_inst_filehdr->ftok_counter_halted, TRUE);
 		} else
 		{	/* Child source server process errored out at startup and is no longer alive.
 			 * If we were the one who created the journal pool, let us clean it up.
@@ -239,7 +289,6 @@ int gtmsource()
 	 * to ensure they do not misrepresent the holder of those semaphores.
 	 */
 	ftok_sem_reg = NULL;
-	udi = FILE_INFO(jnlpool.jnlpool_dummy_reg);
 	assert(udi->grabbed_ftok_sem);
 	udi->grabbed_ftok_sem = FALSE;
 	assert(holds_sem[SOURCE][JNL_POOL_ACCESS_SEM]);
@@ -329,7 +378,7 @@ int gtmsource()
 	 * for the replication instance file. But the source server process (the child) that comes here would not have done
 	 * that. Do that while the parent is still holding on to the ftok semaphore waiting for our okay.
 	 */
-	if (!ftok_sem_incrcnt(jnlpool.jnlpool_dummy_reg))
+	if (!ftok_sem_incrcnt(jnlpool.jnlpool_dummy_reg, FILE_TYPE_REPLINST, &ftok_counter_halted))
 		rts_error_csa(CSA_ARG(NULL) VARLSTCNT(1) ERR_JNLPOOLSETUP);
 	/* Increment the source server count semaphore */
 	status = incr_sem(SOURCE, SRC_SERV_COUNT_SEM);
@@ -381,6 +430,7 @@ int gtmsource()
 		sgtm_putmsg(print_msg, VARLSTCNT(3) ERR_REPLINSTFREEZECOMMENT, 1, jnlpool.jnlpool_ctl->freeze_comment);
 		repl_log(gtmsource_log_fp, TRUE, TRUE, print_msg);
 	}
+	gtmsource_local->jnlfileonly = gtmsource_options.jnlfileonly;
 	do
 	{ 	/* If mode is passive, go to sleep. Wakeup every now and then and check to see if I have to become active. */
 		gtmsource_state = gtmsource_local->gtmsource_state = GTMSOURCE_START;
@@ -420,7 +470,7 @@ int gtmsource()
 		}
 		QWASSIGN(gtmsource_local->read_addr, jnlpool.jnlpool_ctl->write_addr);
 		gtmsource_local->read = jnlpool.jnlpool_ctl->write;
-		gtmsource_local->read_state = READ_POOL;
+		gtmsource_local->read_state = gtmsource_local->jnlfileonly ? READ_FILE : READ_POOL;
 		read_jnl_seqno = gtmsource_local->read_jnl_seqno;
 		assert(read_jnl_seqno <= jnlpool.jnlpool_ctl->jnl_seqno);
 		if (read_jnl_seqno < jnlpool.jnlpool_ctl->jnl_seqno)
