@@ -1,6 +1,7 @@
 /****************************************************************
  *								*
- *	Copyright 2001, 2014 Fidelity Information Services, Inc	*
+ * Copyright (c) 2001-2015 Fidelity National Information 	*
+ * Services, Inc. and/or its subsidiaries. All rights reserved.	*
  *								*
  *	This source code contains the intellectual property	*
  *	of its copyright holder(s), and is made available	*
@@ -34,6 +35,7 @@
 #include "gtm_utf8.h"
 #endif
 #include "gtmcrypt.h"
+#include "error.h"
 
 GBLREF	io_pair		io_curr_device;
 GBLREF	spdesc		stringpool;
@@ -69,9 +71,11 @@ void iorm_readfl_badchar(mval *vmvalptr, int datalen, int delimlen, unsigned cha
 	unsigned char   *delimend;
 	io_desc         *iod;
 	d_rm_struct	*rm_ptr;
+	boolean_t	ch_set;
 
 	assert(0 <= datalen);
 	iod = io_curr_device.in;
+	ESTABLISH_GTMIO_CH(&io_curr_device, ch_set);
 	rm_ptr = (d_rm_struct *)(iod->dev_sp);
 	assert(NULL != rm_ptr);
 	vmvalptr->str.len = datalen;
@@ -102,6 +106,7 @@ void iorm_readfl_badchar(mval *vmvalptr, int datalen, int delimlen, unsigned cha
 	len = SIZEOF(ONE_COMMA) - 1;
 	memcpy(iod->dollar.device, ONE_COMMA, len);
 	memcpy(&iod->dollar.device[len], BADCHAR_DEVICE_MSG, SIZEOF(BADCHAR_DEVICE_MSG));
+	REVERT_GTMIO_CH(&io_curr_device, ch_set);
 }
 #endif
 
@@ -117,7 +122,7 @@ int	iorm_readfl (mval *v, int4 width, int4 timeout) /* timeout in seconds */
 	int4		msec_timeout;	/* timeout in milliseconds */
 	int4		bytes2read, bytes_read, char_bytes_read, add_bytes, reclen;
 	int4		buff_len, mblen, char_count, bytes_count, tot_bytes_read, utf_tot_bytes_read;
-	int4		status, max_width, ltind, exp_width, from_bom;
+	int4		status, max_width, ltind, exp_width, from_bom, fol_bytes_read, feof_status;
 	wint_t		utf_code;
 	char		*errptr;
 	io_desc		*io_ptr;
@@ -149,7 +154,7 @@ int	iorm_readfl (mval *v, int4 width, int4 timeout) /* timeout in seconds */
 	int		fstat_res;
 	off_t		cur_position;
 	int		bom_size_toread;
-
+	boolean_t	ch_set;
 	DCL_THREADGBL_ACCESS;
 
 	SETUP_THREADGBL_ACCESS;
@@ -161,6 +166,7 @@ int	iorm_readfl (mval *v, int4 width, int4 timeout) /* timeout in seconds */
 	assert(stringpool.free <= stringpool.top);
 
 	io_ptr = io_curr_device.in;
+	ESTABLISH_RET_GTMIO_CH(&io_curr_device, -1, ch_set);
 	/* don't allow a read from a writeonly fifo */
 	if (((d_rm_struct *)io_ptr->dev_sp)->write_only)
 		rts_error_csa(CSA_ARG(NULL) VARLSTCNT(1) ERR_DEVICEWRITEONLY);
@@ -229,7 +235,7 @@ int	iorm_readfl (mval *v, int4 width, int4 timeout) /* timeout in seconds */
 	if (!rm_ptr->fifo && !rm_ptr->pipe && (2 < rm_ptr->fildes) && (RM_WRITE == rm_ptr->lastop))
 	{
 		/* need to do an lseek to get current location in file */
-		cur_position = lseek(rm_ptr->fildes, (off_t)0, SEEK_CUR);
+		cur_position = lseek(rm_ptr->fildes, 0, SEEK_CUR);
 		if ((off_t)-1 == cur_position)
 		{
 			rts_error_csa(CSA_ARG(NULL) VARLSTCNT(9) ERR_IOERROR, 7, RTS_ERROR_LITERAL("lseek"),
@@ -240,10 +246,10 @@ int	iorm_readfl (mval *v, int4 width, int4 timeout) /* timeout in seconds */
 		if (!rm_ptr->fixed && !utf_active)
 		{
 			/* move input stream */
-			if (-1 == fseek(rm_ptr->filstr, (long)rm_ptr->file_pos, SEEK_SET))
+			if ((off_t)-1 == fseeko(rm_ptr->filstr, rm_ptr->file_pos, SEEK_SET))
 			{
 				rts_error_csa(CSA_ARG(NULL) VARLSTCNT(9) ERR_IOERROR, 7,
-					      RTS_ERROR_LITERAL("fseek"),
+					      RTS_ERROR_LITERAL("fseeko"),
 					      RTS_ERROR_LITERAL("iorm_readfl()"), CALLFROM, errno);
 			}
 		}
@@ -278,15 +284,14 @@ int	iorm_readfl (mval *v, int4 width, int4 timeout) /* timeout in seconds */
 					bom_size_toread = 0;
 				if (0 < bom_size_toread)
 				{
-					if ((off_t)-1 == lseek(fildes, (off_t)0, SEEK_SET))
+					if ((off_t)-1 == lseek(fildes, 0, SEEK_SET))
 						rts_error_csa(CSA_ARG(NULL) VARLSTCNT(9) ERR_IOERROR, 7,
 							      RTS_ERROR_LITERAL("lseek"), RTS_ERROR_LITERAL(
 								      "Error setting file pointer to beginning of file"),
 							      CALLFROM, errno);
-
 					rm_ptr->bom_num_bytes = open_get_bom(io_ptr, bom_size_toread);
 					/* move back to previous file position */
-					if ((off_t)-1 == lseek(fildes, (off_t)rm_ptr->file_pos, SEEK_SET))
+					if ((off_t)-1 == lseek(fildes, rm_ptr->file_pos, SEEK_SET))
 						rts_error_csa(CSA_ARG(NULL) VARLSTCNT(9) ERR_IOERROR, 7,
 							      RTS_ERROR_LITERAL("lseek"), RTS_ERROR_LITERAL(
 								      "Error restoring file pointer"), CALLFROM, errno);
@@ -468,7 +473,8 @@ int	iorm_readfl (mval *v, int4 width, int4 timeout) /* timeout in seconds */
 					msec_timeout = -1;
                                         out_of_time = TRUE;
                                 } else
-					msec_timeout = (int4)(cur_time.at_sec * 1000 + cur_time.at_usec / 1000);
+					msec_timeout = (int4)(cur_time.at_sec * MILLISECS_IN_SEC +
+							      DIVIDE_ROUND_UP(cur_time.at_usec, MICROSECS_IN_MSEC));
 				if (rm_ptr->follow && !out_of_time && !msec_timeout)
 					msec_timeout = 1;
 				PIPE_DEBUG(PRINTF("piperfl: Taking timeout end time from read restart data - "
@@ -520,7 +526,8 @@ int	iorm_readfl (mval *v, int4 width, int4 timeout) /* timeout in seconds */
 							msec_timeout = -1;
 							out_of_time = TRUE;
 						} else
-							msec_timeout = (int4)(time_left.at_sec * 1000 + time_left.at_usec / 1000);
+							msec_timeout = (int4)(time_left.at_sec * MILLISECS_IN_SEC +
+									DIVIDE_ROUND_UP(time_left.at_usec, MICROSECS_IN_MSEC));
 						/* make sure it terminates with out_of_time */
 						if (!out_of_time && !msec_timeout)
 							msec_timeout = 1;
@@ -564,8 +571,9 @@ int	iorm_readfl (mval *v, int4 width, int4 timeout) /* timeout in seconds */
 								msec_timeout = -1;
 								out_of_time = TRUE;
 							} else
-								msec_timeout = (int4)(time_left.at_sec * 1000 +
-										      time_left.at_usec / 1000);
+								msec_timeout = (int4)(time_left.at_sec * MILLISECS_IN_SEC +
+										      DIVIDE_ROUND_UP(time_left.at_usec,
+												      MICROSECS_IN_MSEC));
 
 							/* make sure it terminates with out_of_time */
 							if (!out_of_time && !msec_timeout)
@@ -597,6 +605,7 @@ int	iorm_readfl (mval *v, int4 width, int4 timeout) /* timeout in seconds */
 							stringpool.free += tot_bytes_read; /* Don't step on our parade in
 											      the interrupt */
 							(TREF(pipefifo_interrupt))++;
+							REVERT_GTMIO_CH(&io_curr_device, ch_set);
 							outofband_action(FALSE);
 							assertpro(FALSE);	/* Should *never* return from outofband_action */
 							return FALSE;	/* For the compiler.. */
@@ -667,6 +676,7 @@ int	iorm_readfl (mval *v, int4 width, int4 timeout) /* timeout in seconds */
 					rm_ptr->mupintr = TRUE;
 					stringpool.free += tot_bytes_read;	/* Don't step on our parade in the interrupt */
 					(TREF(pipefifo_interrupt))++;
+					REVERT_GTMIO_CH(&io_curr_device, ch_set);
 					outofband_action(FALSE);
 					assertpro(FALSE);	/* Should *never* return from outofband_action */
 					return FALSE;	/* For the compiler.. */
@@ -687,7 +697,8 @@ int	iorm_readfl (mval *v, int4 width, int4 timeout) /* timeout in seconds */
 						msec_timeout = -1;
 						out_of_time = TRUE;
 					} else
-						msec_timeout = (int4)(time_left.at_sec * 1000 + time_left.at_usec / 1000);
+						msec_timeout = (int4)(time_left.at_sec * MILLISECS_IN_SEC +
+								      DIVIDE_ROUND_UP(time_left.at_usec, MICROSECS_IN_MSEC));
 					/* make sure it terminates with out_of_time */
 					if (!out_of_time && !msec_timeout)
 						msec_timeout = 1;
@@ -701,7 +712,8 @@ int	iorm_readfl (mval *v, int4 width, int4 timeout) /* timeout in seconds */
 			do
 			{
 				/* in follow mode a read will return an EOF if no more bytes are available*/
-				if (EOF != (status = getc(filstr)))
+				GETC(filstr, status);
+				if (EOF != status)
 				{
 					inchar = (unsigned char)status;
 					tot_bytes_read++;
@@ -720,10 +732,11 @@ int	iorm_readfl (mval *v, int4 width, int4 timeout) /* timeout in seconds */
 				} else
 				{
 					inchar = 0;
-					if (feof(filstr))
+					FEOF(filstr, feof_status);
+					if (feof_status)
 					{
 						status = 0;
-						clearerr(filstr);
+						CLEARERR(filstr);
 
 						if (rm_ptr->follow)
 						{
@@ -746,8 +759,9 @@ int	iorm_readfl (mval *v, int4 width, int4 timeout) /* timeout in seconds */
 									msec_timeout = -1;
 									out_of_time = TRUE;
 								} else
-									msec_timeout = (int4)(time_left.at_sec * 1000 +
-											      time_left.at_usec / 1000);
+									msec_timeout = (int4)(time_left.at_sec * MILLISECS_IN_SEC +
+											      DIVIDE_ROUND_UP(time_left.at_usec,
+													      MICROSECS_IN_MSEC));
 
 								/* make sure it terminates with out_of_time */
 								if (!out_of_time && !msec_timeout)
@@ -780,6 +794,7 @@ int	iorm_readfl (mval *v, int4 width, int4 timeout) /* timeout in seconds */
 								stringpool.free += tot_bytes_read; /* Don't step on our parade
 												      in the interrupt */
 								(TREF(pipefifo_interrupt))++;
+								REVERT_GTMIO_CH(&io_curr_device, ch_set);
 								outofband_action(FALSE);
 								/* Should *never* return from outofband_action */
 								assertpro(FALSE);
@@ -799,7 +814,8 @@ int	iorm_readfl (mval *v, int4 width, int4 timeout) /* timeout in seconds */
 			{
 				unsigned char tchar;
 				int tfcntl_res;
-				if (EOF != (status = getc(filstr)))
+				GETC(filstr, status);
+				if (EOF != status)
 				{
 					tchar = (unsigned char)status;
 					if (rm_ptr->input_encrypted)
@@ -808,11 +824,14 @@ int	iorm_readfl (mval *v, int4 width, int4 timeout) /* timeout in seconds */
 					/* force it to process below in case character read is a 0 */
 					if (!status)
 						status = 1;
-				}
-				else if (feof(filstr))
+				} else
 				{
-					status = 0;
-					clearerr(filstr);
+					FEOF(filstr, feof_status);
+					if (feof_status)
+					{
+						status = 0;
+						CLEARERR(filstr);
+					}
 				}
 
 				if (0 > status)
@@ -888,6 +907,7 @@ int	iorm_readfl (mval *v, int4 width, int4 timeout) /* timeout in seconds */
 					rm_ptr->mupintr = TRUE;
 					stringpool.free += tot_bytes_read;	/* Don't step on our parade in the interrupt */
 					(TREF(pipefifo_interrupt))++;
+					REVERT_GTMIO_CH(&io_curr_device, ch_set);
 					outofband_action(FALSE);
 					assertpro(FALSE);	/* Should *never* return from outofband_action */
 					return FALSE;	/* For the compiler.. */
@@ -906,17 +926,24 @@ int	iorm_readfl (mval *v, int4 width, int4 timeout) /* timeout in seconds */
 			if ((0 == buff_len) || zint_restart)
 			{	/* need to refill the buffer */
 				if (rm_ptr->follow)
+				{
 					buff_len = iorm_get_fol(io_ptr, &tot_bytes_read, &msec_timeout, timed, zint_restart,
 								&follow_timeout, end_time);
-				else
+					/* this will include the total bytes read including any stripped pad chars */
+					fol_bytes_read = rm_ptr->fol_bytes_read;
+				} else
+				{
 					buff_len = iorm_get(io_ptr, &blocked_in, rm_ptr->pipe, flags, &tot_bytes_read,
 						    timer_id, &msec_timeout, pipe_zero_timeout, zint_restart);
+					/* not using fol_bytes_read for non-follow mode */
+					fol_bytes_read = buff_len;
+				}
 				if (0 > buff_len)
 				{
 					bytes_count = 0;
 					if (errno == EINTR  &&  out_of_time)
 						buff_len = -2;
-				} else if (outofband && (buff_len < rm_ptr->recordsize))
+				} else if (outofband && (fol_bytes_read < rm_ptr->recordsize))
 				{
 					PIPE_DEBUG(PRINTF(" %d utf fixed outofband, buff_len: %d done_1st_read: %d\n", pid,
 							  buff_len, rm_ptr->done_1st_read); DEBUGPIPEFLUSH);
@@ -939,6 +966,7 @@ int	iorm_readfl (mval *v, int4 width, int4 timeout) /* timeout in seconds */
 					pipeintr->bytes_read = 0;
 					rm_ptr->mupintr = TRUE;
 					(TREF(pipefifo_interrupt))++;
+					REVERT_GTMIO_CH(&io_curr_device, ch_set);
 					outofband_action(FALSE);
 					assertpro(FALSE);	/* Should *never* return from outofband_action */
 					return FALSE;	/* For the compiler.. */
@@ -1080,7 +1108,8 @@ int	iorm_readfl (mval *v, int4 width, int4 timeout) /* timeout in seconds */
 							msec_timeout = -1;
 							out_of_time = TRUE;
 						} else
-							msec_timeout = (int4)(time_left.at_sec * 1000 + time_left.at_usec / 1000);
+							msec_timeout = (int4)(time_left.at_sec * MILLISECS_IN_SEC +
+									DIVIDE_ROUND_UP(time_left.at_usec, MICROSECS_IN_MSEC));
 
 						/* make sure it terminates with out_of_time */
 						if (!out_of_time && !msec_timeout)
@@ -1136,6 +1165,7 @@ int	iorm_readfl (mval *v, int4 width, int4 timeout) /* timeout in seconds */
 						rm_ptr->mupintr = TRUE;
 						(TREF(pipefifo_interrupt))++;
 						PIPE_DEBUG(PRINTF(" %d utf1.1 stream outofband\n", pid); DEBUGPIPEFLUSH);
+						REVERT_GTMIO_CH(&io_curr_device, ch_set);
 						outofband_action(FALSE);
 						assertpro(FALSE);	/* Should *never* return from outofband_action */
 						return FALSE;	/* For the compiler.. */
@@ -1211,8 +1241,10 @@ int	iorm_readfl (mval *v, int4 width, int4 timeout) /* timeout in seconds */
 										msec_timeout = -1;
 										out_of_time = TRUE;
 									} else
-										msec_timeout = (int4)(time_left.at_sec * 1000 +
-												      time_left.at_usec / 1000);
+										msec_timeout = (int4)(time_left.at_sec *
+												MILLISECS_IN_SEC +
+												DIVIDE_ROUND_UP(time_left.at_usec,
+												MICROSECS_IN_MSEC));
 
 									/* make sure it terminates with out_of_time */
 									if (!out_of_time && !msec_timeout)
@@ -1261,6 +1293,7 @@ int	iorm_readfl (mval *v, int4 width, int4 timeout) /* timeout in seconds */
 									/* Don't step on our parade in the interrupt */
 									stringpool.free += bytes_count;
 									(TREF(pipefifo_interrupt))++;
+									REVERT_GTMIO_CH(&io_curr_device, ch_set);
 									outofband_action(FALSE);
 									assertpro(FALSE);	/* Should *never* return from
 											   outofband_action */
@@ -1328,6 +1361,7 @@ int	iorm_readfl (mval *v, int4 width, int4 timeout) /* timeout in seconds */
 							/* Don't step on our parade in the interrupt */
 							stringpool.free += bytes_count;
 							(TREF(pipefifo_interrupt))++;
+							REVERT_GTMIO_CH(&io_curr_device, ch_set);
 							outofband_action(FALSE);
 							assertpro(FALSE);	/* Should *never* return from outofband_action */
 							return FALSE;	/* For the compiler.. */
@@ -1600,7 +1634,7 @@ int	iorm_readfl (mval *v, int4 width, int4 timeout) /* timeout in seconds */
 	PIPE_DEBUG(PRINTF(" %d notoutofband, %d %d\n", pid, status, line_term_seen); DEBUGPIPEFLUSH);
 	real_errno = errno;
 	if (TRUE == do_clearerr)
-		clearerr(filstr);
+		CLEARERR(filstr);
 	memcpy(io_ptr->dollar.device, "0", SIZEOF("0"));
 	io_ptr->dollar.za = 0;
 	/* On error, getc() returns EOF while read() returns -1. Both code paths converge here. Thankfully EOF is -1 on all
@@ -1655,6 +1689,7 @@ int	iorm_readfl (mval *v, int4 width, int4 timeout) /* timeout in seconds */
 		{
 			if (TRUE == io_ptr->dollar.zeof)
 				io_ptr->dollar.zeof = FALSE; /* no EOF in follow mode */
+			REVERT_GTMIO_CH(&io_curr_device, ch_set);
 			return FALSE;
 		}
 		/* on end of file set $za to 9 */
@@ -1720,5 +1755,6 @@ int	iorm_readfl (mval *v, int4 width, int4 timeout) /* timeout in seconds */
 	if (follow_timeout)
 		ret = FALSE;
 	assert (FALSE == rm_ptr->mupintr);
+	REVERT_GTMIO_CH(&io_curr_device, ch_set);
 	return (rm_ptr->pipe && out_of_time) ? FALSE : ret;
 }
