@@ -1,6 +1,7 @@
 /****************************************************************
  *                                                              *
- *    Copyright 2001, 2013 Fidelity Information Services, Inc   *
+ * Copyright (c) 2001-2015 Fidelity National Information	*
+ * Services, Inc. and/or its subsidiaries. All rights reserved.	*
  *                                                              *
  *    This source code contains the intellectual property       *
  *    of its copyright holder(s), and is made available         *
@@ -14,13 +15,14 @@
 
 #include <sys/types.h>
 
-#ifdef GTM64
-typedef	long		int8;		/* 8-byte signed integer */
-typedef unsigned long	uint8;	  	/* 8-byte unsigned integer */
-typedef unsigned long 	gtm_uint8;	/*these two datatypes are defined because */
-typedef		 long	gtm_int8;	/*int8 and uint8 are system defined in AIX_64*/
-#define INT8_NATIVE
+/* All current Unix platforms support {u,}int64_t, so set these up regardless of GTM64 setting. */
+#ifndef int8
+typedef	int64_t		int8;		/* 8-byte signed integer */
+typedef uint64_t	uint8;	  	/* 8-byte unsigned integer */
 #endif
+typedef uint64_t 	gtm_uint8;	/* these two datatypes are defined because */
+typedef	int64_t		gtm_int8;	/* int8 and uint8 are system defined in AIX_64 */
+#define INT8_NATIVE
 
 #ifdef __s390__
 typedef		 short int2;		/* 2-byte signed integer */
@@ -42,6 +44,7 @@ typedef uint2 mach_inst;
 #ifdef __sparc
 #define CACHELINE_SIZE	256
 #define USHBIN_SUPPORTED
+#define AUTORELINK_SUPPORTED
 #define LINKAGE_PSECT_BOUNDARY	8
 #define OFF_T_LONG
 #define INO_T_LONG
@@ -71,6 +74,9 @@ typedef uint4 mach_inst;
 #define MUTEX_MSEM_WAKE
 #define POSIX_MSEM
 #define USHBIN_SUPPORTED
+/* #define AUTORELINK_SUPPORTED -- not yet due to issues with replacing mapped object files - revisit if/when support objects in
+ * shared memory.
+ */
 #define OFF_T_LONG
 /* Make sure linkage Psect is aligned on appropriate boundary. */
 #ifdef __ia64
@@ -110,6 +116,8 @@ typedef unsigned short	in_port_t;
 #ifdef UNICODE_SUPPORTED_OBEYED
 #undef UNICODE_SUPPORTED
 #endif
+#define MUTEX_MSEM_WAKE
+#define POSIX_MSEM
 #define KEY_T_LONG			/* 8 bytes */
 #define SYS_ERRLIST_INCLUDE	<errno.h>
 #endif
@@ -126,7 +134,8 @@ typedef unsigned short	in_port_t;
 #  ifdef __linux__
 #    undef BIGENDIAN
 #    define USHBIN_SUPPORTED
-     /* Make sure linkage Psect is aligned on appropriate boundary. */
+#    define AUTORELINK_SUPPORTED /* If this ever comes back */
+     /* Make sure linkage Psect is aligned on appropriate boundary */
 #    define LINKAGE_PSECT_BOUNDARY  8
 typedef uint4 mach_inst;	/* machine instruction */
 #  elif defined(__hpux)
@@ -146,6 +155,7 @@ typedef char  mach_inst;	/* machine instruction */
 #ifdef __x86_64__
 #define CACHELINE_SIZE	64
 #define USHBIN_SUPPORTED
+#define AUTORELINK_SUPPORTED
 #define INO_T_LONG
 /*
 #define MUTEX_MSEM_WAKE
@@ -177,14 +187,18 @@ typedef char  mach_inst;	/* machine instruction */
 typedef struct
 {
 	unsigned short	mvtype;
-#ifdef	BIGENDIAN
-	unsigned	sgn	: 1;
-	unsigned	e	: 7;
-#else
-	unsigned	e	: 7;
-	unsigned	sgn	: 1;
-#endif
+#	ifdef	BIGENDIAN
+	unsigned char	sgn	: 1;
+	unsigned char	e	: 7;
+#	else
+	unsigned char	e	: 7;
+	unsigned char	sgn	: 1;
+#	endif
 	unsigned char	fnpc_indx;	/* Index to fnpc_work area this mval is using */
+#	ifdef UNICODE_SUPPORTED
+	unsigned int	utfcgr_indx;	/* Index to utfcgr_work area this mval is using */
+	NON_GTM64_ONLY(unsigned int filler2;) /* To 8 byte align mval on 32 bit platforms */
+#	endif
 	int4	m[2];
 	mstr	str;
 } mval;
@@ -194,6 +208,10 @@ typedef struct
 	unsigned short	mvtype;
 	unsigned char	sgne;
 	unsigned char	fnpc_indx;	/* Index to fnpc_work area this mval is using */
+#	ifdef UNICODE_SUPPORTED
+	unsigned int	utfcgr_indx;	/* Index to utfcgr_work area this mval is using */
+	NON_GTM64_ONLY(unsigned int filler2;) /* To 8 byte align mval on 32 bit platforms */
+#	endif
 	int4	m[2];
 	mstr	str;
 } mval_b;
