@@ -1,6 +1,7 @@
 /****************************************************************
  *								*
- *	Copyright 2001, 2013 Fidelity Information Services, Inc	*
+ * Copyright (c) 2001-2016 Fidelity National Information	*
+ * Services, Inc. and/or its subsidiaries. All rights reserved.	*
  *								*
  *	This source code contains the intellectual property	*
  *	of its copyright holder(s), and is made available	*
@@ -15,7 +16,6 @@
 #include "gtm_fcntl.h"
 #include "gtm_unistd.h"
 
-#include <sys/sem.h>
 #include <sys/shm.h>
 #include <sys/mman.h>
 #include <errno.h>
@@ -58,6 +58,7 @@
 #include "anticipatory_freeze.h"
 #include "heartbeat_timer.h"
 #include "mutex.h"
+#include "do_semop.h"
 
 GBLREF	jnlpool_addrs		jnlpool;
 GBLREF	jnlpool_ctl_ptr_t	jnlpool_ctl;
@@ -68,7 +69,9 @@ GBLREF	jnl_gbls_t		jgbl;
 GBLREF	mur_opt_struct		mur_options;
 GBLREF	mur_gbls_t		murgbl;
 GBLREF	boolean_t		holds_sem[NUM_SEM_SETS][NUM_SRC_SEMS];
+GBLREF	gd_addr			*gd_header;
 
+error_def(ERR_NOMORESEMCNT);
 error_def(ERR_MUJPOOLRNDWNFL);
 error_def(ERR_MUJPOOLRNDWNSUC);
 error_def(ERR_MURPOOLRNDWNFL);
@@ -89,20 +92,22 @@ boolean_t mu_rndwn_repl_instance(replpool_identifier *replpool_id, boolean_t imm
 					boolean_t *jnlpool_sem_created)
 {
 	boolean_t		jnlpool_stat = SS_NORMAL, recvpool_stat = SS_NORMAL, decr_cnt, sem_created = FALSE, ipc_rmvd;
+	boolean_t		clear_ftok_halted;
 	char			*instfilename;
 	unsigned char		ipcs_buff[MAX_IPCS_ID_BUF], *ipcs_ptr;
 	gd_region		*r_save;
 	repl_inst_hdr		repl_instance;
 	static	gd_region	*reg = NULL;
-	struct semid_ds		semstat;
 	struct shmid_ds		shmstat;
+	union semun		semarg;
 	unix_db_info		*udi;
 	int			save_errno, sem_id, shm_id, status;
 	sgmnt_addrs		*repl_csa;
-	boolean_t		was_crit;
+	boolean_t		was_crit, remove_sem, ftok_counter_halted = FALSE;
 	DCL_THREADGBL_ACCESS;
 
 	SETUP_THREADGBL_ACCESS;
+	assert(!jgbl.mur_rollback || !jgbl.mur_options_forward); /* ROLLBACK -FORWARD should not call this function */
 	if (NULL == reg)
 	{
 		r_save = gv_cur_region;
@@ -142,14 +147,17 @@ boolean_t mu_rndwn_repl_instance(replpool_identifier *replpool_id, boolean_t imm
 	memcpy((char *)reg->dyn.addr->fname, instfilename, reg->dyn.addr->fname_len + 1);
 	udi = FILE_INFO(reg);
 	udi->fn = (char *)reg->dyn.addr->fname;
+	ftok_sem_reg = NULL;	/* clean any residue from region work as we have now moved on to an instance */
 	/* Lock replication instance using ftok semaphore so that no other replication process can startup until we are done with
 	 * rundown
 	 */
-	if (!ftok_sem_get(reg, TRUE, REPLPOOL_ID, immediate))
+	if (!ftok_sem_get(reg, TRUE, REPLPOOL_ID, immediate, &ftok_counter_halted))
 		return FALSE;
 	ESTABLISH_RET(mu_rndwn_repl_instance_ch, FALSE);
 	repl_inst_read(instfilename, (off_t)0, (sm_uc_ptr_t)&repl_instance, SIZEOF(repl_inst_hdr));
 	assert(rndwn_both_pools || JNLPOOL_SEGMENT == replpool_id->pool_type || RECVPOOL_SEGMENT == replpool_id->pool_type);
+	if (!ftok_counter_halted && repl_instance.ftok_counter_halted && udi->counter_ftok_incremented)
+		udi->counter_ftok_incremented = FALSE;	/* See comment in mu_rndwn_file.c WHY this is necessary */
 	if (rndwn_both_pools || (JNLPOOL_SEGMENT == replpool_id->pool_type))
 	{	/* --------------------------
 		 * First rundown Journal pool
@@ -177,10 +185,11 @@ boolean_t mu_rndwn_repl_instance(replpool_identifier *replpool_id, boolean_t imm
 				ipcs_ptr = i2asc((uchar_ptr_t)ipcs_buff, shm_id);
 				*ipcs_ptr = '\0';
 				if (rndwn_both_pools && ((SS_NORMAL != jnlpool_stat) || ipc_rmvd))
-					gtm_putmsg(VARLSTCNT(6) (jnlpool_stat ? ERR_MUJPOOLRNDWNFL : ERR_MUJPOOLRNDWNSUC),
-						4, LEN_AND_STR(ipcs_buff), LEN_AND_STR(instfilename));
+					gtm_putmsg_csa(CSA_ARG(NULL) VARLSTCNT(6)
+							(jnlpool_stat ? ERR_MUJPOOLRNDWNFL : ERR_MUJPOOLRNDWNSUC),
+							4, LEN_AND_STR(ipcs_buff), LEN_AND_STR(instfilename));
 			}
-			assert(ipc_rmvd || (NULL != jnlpool_ctl));
+			assert(ipc_rmvd || (NULL != jnlpool_ctl) || !mur_options.rollback);
 			assert((NULL == jnlpool.jnlpool_ctl) || (SS_NORMAL == jnlpool_stat) || jgbl.onlnrlbk);
 			assert((INVALID_SHMID != repl_instance.jnlpool_shmid) || (0 == repl_instance.jnlpool_shmid_ctime));
 			assert((INVALID_SHMID == repl_instance.jnlpool_shmid) || (0 != repl_instance.jnlpool_shmid_ctime));
@@ -191,8 +200,11 @@ boolean_t mu_rndwn_repl_instance(replpool_identifier *replpool_id, boolean_t imm
 				 */
 				if (NULL == jnlpool_ctl)
 				{
-					if (((sem_created || (SS_NORMAL == jnlpool_stat))
-						&& (SS_NORMAL == mu_replpool_release_sem(&repl_instance, JNLPOOL_SEGMENT, TRUE))))
+					remove_sem = (sem_created || ((SS_NORMAL == jnlpool_stat) && ipc_rmvd));
+					if (!remove_sem)
+						add_to_semids_list(repl_instance.jnlpool_semid);
+					status = mu_replpool_release_sem(&repl_instance, JNLPOOL_SEGMENT, remove_sem);
+					if ((SS_NORMAL == status) && remove_sem)
 					{	/* Now that semaphores are removed, reset fields in file header */
 						if (!sem_created)
 						{	/* If sem_id was created by mu_replpool_grab_sem then do NOT report the
@@ -203,16 +215,64 @@ boolean_t mu_rndwn_repl_instance(replpool_identifier *replpool_id, boolean_t imm
 							 */
 							ipcs_ptr = i2asc((uchar_ptr_t)ipcs_buff, sem_id);
 							*ipcs_ptr = '\0';
-							gtm_putmsg(VARLSTCNT(9) ERR_MUJPOOLRNDWNSUC, 4, LEN_AND_STR(ipcs_buff),
-								LEN_AND_STR(instfilename), ERR_SEMREMOVED, 1, sem_id);
+							gtm_putmsg_csa(CSA_ARG(NULL) VARLSTCNT(9) ERR_MUJPOOLRNDWNSUC, 4,
+									LEN_AND_STR(ipcs_buff),
+									LEN_AND_STR(instfilename), ERR_SEMREMOVED, 1, sem_id);
 						}
-						repl_inst_jnlpool_reset();
+						/* If "ftok_counter_halted" is TRUE, and the recvpool shmid is already cleared in
+						 * the instance file, we can be sure there is no one interesting in this instance
+						 * file (because we just now run down the jnlpool shm as well). So clear the
+						 * halted flag.
+						 */
+						clear_ftok_halted = CLEAR_FTOK_HALTED_FALSE;
+						if (repl_instance.ftok_counter_halted
+							&& (INVALID_SHMID == repl_instance.recvpool_shmid))
+						{
+							/* ftok counter is not guaranteed to be at 1. So fix it that way a
+							 * "ftok_sem_release" done later WILL remove the ftok semaphore.
+							 */
+							semarg.val = 0;
+							if (-1 == semctl(udi->ftok_semid, DB_COUNTER_SEM, SETVAL, semarg))
+							{
+								save_errno = errno;
+								gtm_putmsg_csa(CSA_ARG(NULL) VARLSTCNT(8)
+									ERR_SYSCALL, 5, RTS_ERROR_LITERAL("semctl(SETVAL)"),
+									CALLFROM, save_errno);
+								/* In case not able to set counter to 1, proceed with rundown
+								 * without deleting the ftok semaphore (so keep
+								 * counter_ftok_incremented at FALSE)
+								 */
+							} else
+							{
+								save_errno = do_semop(udi->ftok_semid, DB_COUNTER_SEM,
+												DB_COUNTER_SEM_INCR, SEM_UNDO);
+								if (save_errno)
+								{
+									gtm_putmsg_csa(CSA_ARG(NULL) VARLSTCNT(8)
+										ERR_SYSCALL, 5, RTS_ERROR_LITERAL("do_semop()"),
+										CALLFROM, save_errno);
+									/* In case not able to set counter to 1, proceed with
+									 * rundown without deleting the ftok semaphore (so keep
+									 * counter_ftok_incremented at FALSE)
+									 */
+								} else
+								{
+									udi->counter_ftok_incremented = TRUE;
+									/* repl_instance.ftok_counter_halted will be set to
+									 * FALSE by the below call to repl_inst_jnlpool_reset.
+									 */
+									clear_ftok_halted = CLEAR_FTOK_HALTED_TRUE;
+								}
+							}
+						}
+						repl_inst_jnlpool_reset(clear_ftok_halted);
 					}
 				} else
-				{	/* Anticipatory Freeze scheme is turned ON. So, release just the JNL_POOL_ACCESS_SEM. The
-					 * semaphore will be released/removed in the caller (mupip_rundown)
+				{	/* This is MUPIP RUNDOWN -REG "*" and anticipatory freeze scheme is turned ON.
+					 * So, release just the JNL_POOL_ACCESS_SEM. The semaphore will be released/removed
+					 * in the caller (mupip_rundown).
 					 */
-					assert(ANTICIPATORY_FREEZE_AVAILABLE);
+					assert(INST_FREEZE_ON_ERROR_POLICY);
 					assertpro(SS_NORMAL == (status = rel_sem(SOURCE, JNL_POOL_ACCESS_SEM)));
 					assert(!holds_sem[SOURCE][JNL_POOL_ACCESS_SEM]);
 					/* Since we are not resetting the semaphore IDs in the file header, we need to write out
@@ -222,25 +282,13 @@ boolean_t mu_rndwn_repl_instance(replpool_identifier *replpool_id, boolean_t imm
 						repl_inst_write(instfilename, (off_t)0, (sm_uc_ptr_t)&repl_instance,
 									SIZEOF(repl_inst_hdr));
 				}
-				/* If semaphore is not created and the journal pool rundown failed (due to attached processes),
-				 * rundown process continues to holds the journal pool access control semaphore. This way, we hold
-				 * the semaphore on behalf of the source server (now no longer alive) to prevent mu_rndwn_sem_all
-				 * (invoked later) from cleaning up this orphaned semaphore (which causes REPLREQROLLBACK if the
-				 * source server is restarted). But, since the semaphore is not released (until the rundown process
-				 * dies), holds_sem[SOURCE][JNL_POOL_ACCESS_SEM] continues to remain TRUE. This causes asserts in
-				 * ftok_sem_get if mu_rndwn_repl_instance is invoked for a different journal/receive pool. To
-				 * workaround it, set holds_sem[SOURCE][JNL_POOL_ACCESS_SEM] to FALSE. This is an interim solution
-				 * until we record such semaphores in an ignore-list (or some such) and change mu_rndwn_sem_all to
-				 * skip the ones that are present in the ignore list.
-				 */
-				holds_sem[SOURCE][JNL_POOL_ACCESS_SEM] = FALSE;
 			}
 		} else if (rndwn_both_pools && (INVALID_SHMID != shm_id))
 		{
 			ipcs_ptr = i2asc((uchar_ptr_t)ipcs_buff, shm_id);
 			*ipcs_ptr = '\0';
 			if (rndwn_both_pools)
-				gtm_putmsg(VARLSTCNT(6) ERR_MUJPOOLRNDWNFL, 4, LEN_AND_STR(ipcs_buff),
+				gtm_putmsg_csa(CSA_ARG(NULL) VARLSTCNT(6) ERR_MUJPOOLRNDWNFL, 4, LEN_AND_STR(ipcs_buff),
 					LEN_AND_STR(instfilename));
 		}
 		*jnlpool_sem_created = sem_created;
@@ -273,8 +321,9 @@ boolean_t mu_rndwn_repl_instance(replpool_identifier *replpool_id, boolean_t imm
 				ipcs_ptr = i2asc((uchar_ptr_t)ipcs_buff, shm_id);
 				*ipcs_ptr = '\0';
 				if (rndwn_both_pools && ((SS_NORMAL != recvpool_stat) || ipc_rmvd))
-					gtm_putmsg(VARLSTCNT(6) (recvpool_stat ? ERR_MURPOOLRNDWNFL : ERR_MURPOOLRNDWNSUC),
-						4, LEN_AND_STR(ipcs_buff), LEN_AND_STR(instfilename));
+					gtm_putmsg_csa(CSA_ARG(NULL) VARLSTCNT(6)
+							(recvpool_stat ? ERR_MURPOOLRNDWNFL : ERR_MURPOOLRNDWNSUC),
+							4, LEN_AND_STR(ipcs_buff), LEN_AND_STR(instfilename));
 			}
 			assert((TRUE == ipc_rmvd) || (SS_NORMAL != recvpool_stat) || jgbl.onlnrlbk);
 			assert((INVALID_SHMID != repl_instance.recvpool_shmid) || (0 == repl_instance.recvpool_shmid_ctime));
@@ -284,8 +333,11 @@ boolean_t mu_rndwn_repl_instance(replpool_identifier *replpool_id, boolean_t imm
 			{	/* Invoked by MUPIP RUNDOWN in which case the semaphores needs to be removed. But, remove the
 				 * semaphore ONLY if we created it here OR the receive pool was successfully removed.
 				 */
-				if ((sem_created || (SS_NORMAL == recvpool_stat))
-					&& (SS_NORMAL == mu_replpool_release_sem(&repl_instance, RECVPOOL_SEGMENT, TRUE)))
+				remove_sem = sem_created || (SS_NORMAL == jnlpool_stat);
+				if (!remove_sem)
+					add_to_semids_list(repl_instance.jnlpool_semid);
+				status = mu_replpool_release_sem(&repl_instance, RECVPOOL_SEGMENT, remove_sem);
+				if ((SS_NORMAL == status) && remove_sem)
 				{	/* Now that semaphores are removed, reset fields in file header */
 					if (!sem_created)
 					{	/* if sem_id was "created" by mu_replpool_grab_sem then do NOT report the
@@ -296,14 +348,16 @@ boolean_t mu_rndwn_repl_instance(replpool_identifier *replpool_id, boolean_t imm
 						 */
 						ipcs_ptr = i2asc((uchar_ptr_t)ipcs_buff, sem_id);
 						*ipcs_ptr = '\0';
-						gtm_putmsg(VARLSTCNT(9) ERR_MURPOOLRNDWNSUC, 4, LEN_AND_STR(ipcs_buff),
-							LEN_AND_STR(instfilename), ERR_SEMREMOVED, 1, sem_id);
+						gtm_putmsg_csa(CSA_ARG(NULL) VARLSTCNT(9) ERR_MURPOOLRNDWNSUC, 4,
+								LEN_AND_STR(ipcs_buff), LEN_AND_STR(instfilename),
+								ERR_SEMREMOVED, 1, sem_id);
 					}
 					if (NULL != jnlpool_ctl)
 					{	/* Journal pool is not yet removed. So, grab lock before resetting semid/shmid
 						 * fields in the file header as the function expects the caller to hold crit
 						 * if the journal pool is available
 						 */
+						assert(INVALID_SHMID != repl_instance.jnlpool_shmid);
 						repl_csa = &FILE_INFO(jnlpool.jnlpool_dummy_reg)->s_addrs;
 						assert(!repl_csa->now_crit);
 						assert(!repl_csa->hold_onto_crit);
@@ -317,37 +371,68 @@ boolean_t mu_rndwn_repl_instance(replpool_identifier *replpool_id, boolean_t imm
 						if (!was_crit)
 							grab_lock(jnlpool.jnlpool_dummy_reg, TRUE, GRAB_LOCK_ONLY);
 					}
-					repl_inst_recvpool_reset();
+					/* If "ftok_counter_halted" is TRUE, and the jnlpool shmid is already cleared in
+					 * the instance file, we can be sure there is no one interesting in this instance file
+					 * (because we just now ran down the recvpool shm as well). So clear the halted flag.
+					 */
+					clear_ftok_halted = CLEAR_FTOK_HALTED_FALSE;
+					if (repl_instance.ftok_counter_halted && (INVALID_SHMID == repl_instance.jnlpool_shmid))
+					{
+						/* ftok counter is not guaranteed to be at 1. So fix it that way a
+						 * "ftok_sem_release" done later WILL remove the ftok semaphore.
+						 */
+						semarg.val = 0;
+						if (-1 == semctl(udi->ftok_semid, DB_COUNTER_SEM, SETVAL, semarg))
+						{
+							save_errno = errno;
+							gtm_putmsg_csa(CSA_ARG(NULL) VARLSTCNT(8)
+								ERR_SYSCALL, 5, RTS_ERROR_LITERAL("semctl(SETVAL)"),
+								CALLFROM, save_errno);
+							/* In case not able to set counter to 1, proceed with rundown without
+							 * deleting the ftok semaphore (so keep counter_ftok_incremented unchanged)
+							 */
+						} else
+						{
+							save_errno = do_semop(udi->ftok_semid, DB_COUNTER_SEM,
+											DB_COUNTER_SEM_INCR, SEM_UNDO);
+							if (save_errno)
+							{
+								gtm_putmsg_csa(CSA_ARG(NULL) VARLSTCNT(8)
+									ERR_SYSCALL, 5, RTS_ERROR_LITERAL("do_semop()"),
+									CALLFROM, save_errno);
+								/* In case not able to set counter to 1, proceed with
+								 * rundown without deleting the ftok semaphore (so keep
+								 * counter_ftok_incremented at FALSE)
+								 */
+							} else
+							{
+								udi->counter_ftok_incremented = TRUE;
+								/* repl_instance.ftok_counter_halted will be set to FALSE by the
+								 * below call to repl_inst_recvpool_reset.
+								 */
+								clear_ftok_halted = CLEAR_FTOK_HALTED_TRUE;
+							}
+						}
+					}
+					repl_inst_recvpool_reset(clear_ftok_halted);
 					if ((NULL != jnlpool_ctl) && !was_crit)
 						rel_lock(jnlpool.jnlpool_dummy_reg);
 				}
-				/* If semaphore is not created and the receive pool rundown failed (due to attached processes),
-				 * rundown process continues to holds the receive pool access control semaphore. This way, we hold
-				 * the semaphore on behalf of the receiver server (now no longer alive) to prevent mu_rndwn_sem_all
-				 * (invoked later) from cleaning up this orphaned semaphore (which causes REPLREQROLLBACK if the
-				 * receiver is restarted). But, since the semaphore is not released (until the rundown process
-				 * dies), holds_sem[RECV][RECV_POOL_ACCESS_SEM] continues to remain TRUE. This causes asserts in
-				 * ftok_sem_get if mu_rndwn_repl_instance is invoked for a different journal/receive pool. To
-				 * workaround it, set holds_sem[SOURCE][RECV_POOL_ACCESS_SEM] to FALSE. This is an interim solution
-				 * until we record such semaphores in an ignore-list (or some such) and change mu_rndwn_sem_all to
-				 * skip the ones that are present in the ignore list.
-				 */
-				assert((sem_created || (SS_NORMAL == recvpool_stat)) || holds_sem[RECV][RECV_POOL_ACCESS_SEM]);
-				DEBUG_ONLY(set_sem_set_recvr(sem_id));
+				assert(!holds_sem[RECV][RECV_POOL_ACCESS_SEM]);
 			}
 		} else if (rndwn_both_pools && (INVALID_SHMID != shm_id))
 		{
 			ipcs_ptr = i2asc((uchar_ptr_t)ipcs_buff, shm_id);
 			*ipcs_ptr = '\0';
 			if (rndwn_both_pools)
-				gtm_putmsg(VARLSTCNT(6) ERR_MURPOOLRNDWNFL, 4, LEN_AND_STR(ipcs_buff),
+				gtm_putmsg_csa(CSA_ARG(NULL) VARLSTCNT(6) ERR_MURPOOLRNDWNFL, 4, LEN_AND_STR(ipcs_buff),
 					LEN_AND_STR(instfilename));
 		}
 	}
-	assert(jgbl.onlnrlbk || ANTICIPATORY_FREEZE_AVAILABLE || (NULL == jnlpool.repl_inst_filehdr));
+	assert(jgbl.onlnrlbk || INST_FREEZE_ON_ERROR_POLICY || (NULL == jnlpool.repl_inst_filehdr));
 	if (mur_options.rollback && (SS_NORMAL == jnlpool_stat) && (SS_NORMAL == recvpool_stat))
 	{
-		assert(jgbl.onlnrlbk || ANTICIPATORY_FREEZE_AVAILABLE || ((INVALID_SHMID == repl_instance.jnlpool_shmid)
+		assert(jgbl.onlnrlbk || INST_FREEZE_ON_ERROR_POLICY || ((INVALID_SHMID == repl_instance.jnlpool_shmid)
 			&& (INVALID_SHMID == repl_instance.recvpool_shmid)));
 		/* Initialize jnlpool.repl_inst_filehdr as it is used later by gtmrecv_fetchresync() */
 		decr_cnt = FALSE;
@@ -375,10 +460,10 @@ boolean_t mu_rndwn_repl_instance(replpool_identifier *replpool_id, boolean_t imm
 		/* Flush changes to the replication instance file header to disk */
 		repl_inst_write(instfilename, (off_t)0, (sm_uc_ptr_t)&repl_instance, SIZEOF(repl_inst_hdr));
 	} else /* for MUPIP RUNDOWN, semid fields in the file header are reset and is written in mu_replpool_release_sem() above */
-		decr_cnt = (NULL == jnlpool_ctl); /* for anticipatory freeze, mupip_rundown releases the semaphore */
+		decr_cnt = (NULL == jnlpool_ctl); /* for anticipatory freeze, "mupip_rundown" releases the ftok semaphore */
 	REVERT;
-	/* Release replication instance ftok semaphore lock */
-	if (!ftok_sem_release(reg, decr_cnt, immediate)) /* Do not decrement the counter if ROLLBACK */
+	/* Release replication instance ftok semaphore lock. Do not decrement the counter if ROLLBACK */
+	if (!ftok_sem_release(reg, decr_cnt && udi->counter_ftok_incremented, immediate))
 		return FALSE;
 	return ((SS_NORMAL == jnlpool_stat) && (SS_NORMAL == recvpool_stat));
 }
@@ -389,7 +474,7 @@ CONDITION_HANDLER(mu_rndwn_repl_instance_ch)
 	sgmnt_addrs	*csa;
 	gd_region	*reg;
 
-	START_CH;
+	START_CH(TRUE);
 	reg = jnlpool.jnlpool_dummy_reg;
 	assert(NULL != reg);
 	if (NULL != reg)
@@ -397,7 +482,7 @@ CONDITION_HANDLER(mu_rndwn_repl_instance_ch)
 		udi = FILE_INFO(reg);
 		csa = &udi->s_addrs;
 		if (udi->grabbed_ftok_sem)
-			ftok_sem_release(reg, FALSE, TRUE);
+			ftok_sem_release(reg, udi->counter_ftok_incremented, TRUE);
 	}
 	NEXTCH;
 }
