@@ -1,6 +1,7 @@
 /****************************************************************
  *								*
- *	Copyright 2001, 2013 Fidelity Information Services, Inc	*
+ * Copyright (c) 2001-2016 Fidelity National Information	*
+ * Services, Inc. and/or its subsidiaries. All rights reserved.	*
  *								*
  *	This source code contains the intellectual property	*
  *	of its copyright holder(s), and is made available	*
@@ -40,7 +41,6 @@
 #include "gtmsource.h"
 #include "gtmimagename.h"
 #include "trans_log_name.h"
-#include "get_page_size.h"
 #include "filestruct.h"
 #include "jnl.h"
 #include "gdskill.h"
@@ -76,9 +76,9 @@
 #include "gtcm_gnp_pktdmp.h"
 #include "util.h"
 #include "getzdir.h"
-#include "gtm_env_init.h"	/* for gtm_env_init() prototype */
 #include "suspsigs_handler.h"
-#include "gtm_imagetype_init.h"
+#include "common_startup_init.h"
+#include "gtm_startup_chk.h"
 #include "gtm_threadgbl_init.h"
 #include "fork_init.h"
 #include "gt_timers_add_safe_hndlrs.h"
@@ -132,8 +132,6 @@ GBLDEF char			gtcm_gnp_server_log[MAX_FN_LEN + 1];
 /* the length is the orignal length */
 GBLDEF int			gtcm_gnp_log_path_len;
 
-OS_PAGE_SIZE_DECLARE
-
 static uint4			closewait;
 
 #define CM_SERV_WAIT_FOR_INPUT	100 /* ms */
@@ -159,7 +157,7 @@ static void gtcm_gnp_server_actions(void)
 	unsigned short		value;
 	char			reply;
 	connection_struct	*prev_curr_entry;
-	CMI_MUTEX_DECL;
+	CMI_MUTEX_DECL(cmi_mutex_rc);
 	DCL_THREADGBL_ACCESS;
 
 	SETUP_THREADGBL_ACCESS;
@@ -185,20 +183,20 @@ static void gtcm_gnp_server_actions(void)
 		}
 		if (blkdlist)
 			gtcml_chkreg();
-		CMI_MUTEX_BLOCK;
+		CMI_MUTEX_BLOCK(cmi_mutex_rc);
 		gtcm_remove_from_action_queue();
-		CMI_MUTEX_RESTORE;
+		CMI_MUTEX_RESTORE(cmi_mutex_rc);
 		if ((connection_struct *)INTERLOCK_FAIL == curr_entry)
-			rts_error(VARLSTCNT(1) CMERR_CMINTQUE);
+			rts_error_csa(CSA_ARG(NULL) VARLSTCNT(1) CMERR_CMINTQUE);
 		if ((connection_struct *)EMPTY_QUEUE != curr_entry)
 		{
 			if (1 == (curr_entry->int_cancel.laflag & 1))
 			{ /* valid interrupt cancel msg, handle in gtcm_urgread_ast */
-				CMI_MUTEX_BLOCK;
+				CMI_MUTEX_BLOCK(cmi_mutex_rc);
 				prev_curr_entry = curr_entry;
 				curr_entry = EMPTY_QUEUE;
 				gtcm_int_unpack(prev_curr_entry);
-				CMI_MUTEX_RESTORE;
+				CMI_MUTEX_RESTORE(cmi_mutex_rc);
 				continue;
 			}
 			switch (*curr_entry->clb_ptr->mbf)
@@ -294,7 +292,7 @@ static void gtcm_gnp_server_actions(void)
 					if (SS_NORMAL == status)
 					{
 						GET_LONG(status, curr_entry->clb_ptr->mbf);
-                                                rts_error(VARLSTCNT(3) ERR_BADGTMNETMSG, 1, status);
+                                                rts_error_csa(CSA_ARG(NULL) VARLSTCNT(3) ERR_BADGTMNETMSG, 1, status);
 					}
 					break;
 			}
@@ -313,9 +311,9 @@ static void gtcm_gnp_server_actions(void)
 					curr_entry = EMPTY_QUEUE;
 					if (1 == (prev_curr_entry->int_cancel.laflag & 1))
 					{  /* valid interrupt cancel msg, handle in gtcm_urgread_ast */
-						CMI_MUTEX_BLOCK;
+						CMI_MUTEX_BLOCK(cmi_mutex_rc);
 						gtcm_int_unpack(prev_curr_entry);
-						CMI_MUTEX_RESTORE;
+						CMI_MUTEX_RESTORE(cmi_mutex_rc);
 					} else if (CM_READ == reply)
 					{
 						prev_curr_entry->clb_ptr->ast = gtcm_read_ast;
@@ -369,8 +367,7 @@ int main(int argc, char **argv, char **envp)
 	cmi_descriptor		service_descr, log_path_descr;
 	unsigned short		service_len, log_path_len;
 	char			*ptr, service[512];
-	now_t			now;	/* for GET_CUR_TIME macro */
-	char			time_str[CTIME_BEFORE_NL + 2], *time_ptr; /* for GET_CUR_TIME macro */
+	char			time_str[CTIME_BEFORE_NL + 2];	/* for GET_CUR_TIME macro */
 	pid_t			pid;
 	struct sigaction	act;
 	DCL_THREADGBL_ACCESS;
@@ -378,17 +375,14 @@ int main(int argc, char **argv, char **envp)
         static boolean_t no_fork = FALSE;
 
 	GTM_THREADGBL_INIT;
-	set_blocksig();
-	gtm_imagetype_init(GTCM_GNP_SERVER_IMAGE);
-	gtm_wcswidth_fnptr = gtm_wcswidth;
-	gtm_env_init();	/* read in all environment variables */
+	common_startup_init(GTCM_GNP_SERVER_IMAGE);
 	gtmenvp = envp;
-	getjobnum();
 	err_init(stop_image_conditional_core);
 	assert(0 == offsetof(gv_key, top)); /* for integrity of CM_GET_GVCURRKEY */
 	assert(2 == offsetof(gv_key, end)); /* for integrity of CM_GET_GVCURRKEY */
 	assert(4 == offsetof(gv_key, prev)); /* for integrity of CM_GET_GVCURRKEY */
 	GTM_ICU_INIT_IF_NEEDED;	/* Note: should be invoked after err_init (since it may error out) and before CLI parsing */
+	gtm_chk_dist(argv[0]);
 	/* read comments in gtm.c for cli magic below */
 	cli_lex_setup(argc, argv);
 	if (1 < argc)
@@ -400,7 +394,7 @@ int main(int argc, char **argv, char **envp)
 	cli_lex_in_ptr->tp = cli_lex_in_ptr->in_str;
 	parse_ret = parse_cmd();
 	if (parse_ret && (EOF != parse_ret))
-		rts_error(VARLSTCNT(4) parse_ret, 2, LEN_AND_STR(cli_err_str));
+		rts_error_csa(CSA_ARG(NULL) VARLSTCNT(4) parse_ret, 2, LEN_AND_STR(cli_err_str));
 	service_len = (unsigned short)SIZEOF(service);
 	CMI_DESC_SET_POINTER(&service_descr, service);
 	service[0] = '\0';
@@ -424,7 +418,6 @@ int main(int argc, char **argv, char **envp)
 	gtcm_gnp_server_log[log_path_len] = '\0';
 	gtcm_open_cmerrlog();
         assert(0 == EMPTY_QUEUE);
-	get_page_size();
 	licensed = TRUE;
 	stp_init(STP_INITSIZE);
 	rts_stringpool = stringpool;
@@ -447,20 +440,21 @@ int main(int argc, char **argv, char **envp)
 	gtcm_connection = FALSE;
         if (!no_fork)
         {
-		FORK_CLEAN(pid);
+		FORK(pid);
                 if (0 > pid)
                 {
-			rts_error(VARLSTCNT(5) ERR_TEXT, 2, LEN_AND_LIT("Error forking gnp server into the background"), errno);
-                        exit(-1);
+			rts_error_csa(CSA_ARG(NULL) VARLSTCNT(5) ERR_TEXT, 2,
+					LEN_AND_LIT("Error forking gnp server into the background"), errno);
+                        EXIT(-1);
                 }
                 else if (0 < pid)
-                        exit(0);
+                        EXIT(0);
 		getjobnum();
                 (void) setpgrp();
         }
 	/* Write argv and the process id for ease of admin */
-	GET_CUR_TIME;
-	util_out_print("!AD : ", FALSE, CTIME_BEFORE_NL, time_ptr);
+	GET_CUR_TIME(time_str);
+	util_out_print("!AD : ", FALSE, CTIME_BEFORE_NL, time_str);
 	for (arg_index = 0; arg_index < argc; arg_index++)
 		util_out_print("!AZ ", FALSE, argv[arg_index]);
 	util_out_print("[pid : !UL]", TRUE, process_id);
@@ -472,9 +466,9 @@ int main(int argc, char **argv, char **envp)
 			  SIZEOF(connection_struct), CM_MSG_BUF_SIZE + CM_BUFFER_OVERHEAD);
 	if (CMI_ERROR(status))
 	{
-		gtm_putmsg(VARLSTCNT(7) ERR_NETFAIL, 0, ERR_TEXT, 2, LEN_AND_LIT("Network interface initialization failed"),
-				status);
-		exit(status);
+		gtm_putmsg_csa(CSA_ARG(NULL) VARLSTCNT(7) ERR_NETFAIL, 0,
+				ERR_TEXT, 2, LEN_AND_LIT("Network interface initialization failed"), status);
+		EXIT(status);
 	}
 	atexit(gtcm_exi_handler);
 	INVOKE_INIT_SECSHR_ADDRS;
@@ -496,5 +490,5 @@ int main(int argc, char **argv, char **envp)
 	{
 		gtcm_gnp_server_actions();
 	}
-	exit(SS_NORMAL);
+	return SS_NORMAL;
 }
